{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvCCRecrd.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvCCRecrd;

interface

Uses Sysutils;
type
  PByte = ^Byte;
  PWord = ^WORD;
  PDouble = ^Double;
  PInteger = ^Integer;
  PLongInt = ^LongInt;
  TJvCCSize = integer;

type
  EJvCCRecordMgr = class(Exception);

  TVisitDataProc = procedure(var buf);
  PVisitDataProc = ^TVisitDataProc;

  TJvCCRecordMgr = class(TObject)
  private
    FItemSize : integer;
    FInitItems : boolean; {optional to
     call ConstructItem, DestructItem}
  protected
    procedure InitRecordMgr(itemSize : integer;
       initItems:boolean); virtual;
  public
    constructor Create; virtual;
    function  CompareKey(const item; key:Pointer):integer;virtual;
    function  CompareItems(const p1, p2):integer;virtual; 
    procedure AssignItem(var dest;const src); virtual;
    procedure DestructItem(var item);virtual;
    procedure ConstructItem(var item);virtual;
    procedure SwapItems(var p1, p2); virtual;
    {properties}
    property ItemSize : integer read FItemSize;
    property InitItems : boolean read FInitItems;
  end;
  TJvCCRecordMgrClass = class of TJvCCRecordMgr;

type
  TJvCCItemPosition = record
    FData : Pointer;
    FPosition : LongInt;
  end;

  TJvCCItemPtr = Pointer;
  TJvCCItemSearch = (its_forward, its_backward);

TJvCCActionMethod = procedure(item:TJvCCItemPtr; data:Pointer) of object;
TJvCCActionProc = procedure(item:TJvCCItemPtr; data:Pointer);


  TJvCCBase = class(TObject)
  protected
    FCount : TJvCCSize; {number of items}
    FRecMgr : TJvCCRecordMgr;
    procedure SetCount(inCount : TJvCCSize); virtual;
  public
    constructor CreateMgr(reclass : TJvCCRecordMgrClass); virtual;
    destructor Destroy; override;

    procedure Clear; virtual;

    function LocateItem(const rec; var buf) : boolean; virtual;
    function LocateItemPtr(
      const rec;
      var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch) : boolean; virtual;

    function LocateKey(key:Pointer; var buf) : boolean; virtual;
    function LocateKeyPtr(
      key : Pointer;
      var startItem : TJvCCItemPtr;
      direction : TJvCCItemSearch) : boolean; virtual;

    procedure CopyItem(const pitem : TJvCCItemPtr; var buf);
    procedure CallActionMethod(method: TJvCCActionMethod; data:Pointer);
    procedure CallActionProcedure(proc: TJvCCActionProc; data:Pointer);

    function FirstItem : TJvCCItemPtr; virtual;
    function LastItem : TJvCCItemPtr; virtual;
    function NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; virtual;
    function PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr; virtual;
    function AppendItem(const buf): TJvCCItemPtr; virtual;
    function DeleteItem(const buf): boolean; virtual;
    procedure FreeItem(var pitem : TJvCCItemPtr);  virtual;
    procedure AppendFirst(const buf);virtual;
    procedure AppendLast(const buf); virtual;
    property Manager : TJvCCRecordMgr read FRecMgr;
    property Count : TJvCCSize read FCount write SetCount;
  end;


  TJvCCPtrRecMgr = class(TJvCCRecordMgr)
    constructor Create; override;
  end;

  TJvCCObjCompare = function(obj1, obj2 : TObject) : integer;
  TJvCCKeyObjCompare = function(obj : TObject;key : Pointer) : integer;
  TJvCCObjRecMgr = class(TJvCCRecordMgr)
  private
    FAutoDestroy : boolean;
  protected
    FKeyCompare : TJvCCKeyObjCompare;
    FObjCompare : TJvCCObjCompare;
  public
    constructor Create; override;
    procedure AssignItem(var dest;const src); override;
    procedure DestructItem(var item); override;
    function  CompareItems(const p1, p2):integer;override;
    function  CompareKey(const item; key:Pointer):integer;override;
    procedure ConstructItem(var item); override;
    procedure SwapItems(var p1, p2); override;
    property KeyCompare : TJvCCKeyObjCompare read FKeyCompare write FKeyCompare;
    property ObjCompare : TJvCCObjCompare read FObjCompare write FObjCompare;
    property AutoDestroy : boolean read FAutoDestroy write FAutoDestroy;
  end;
  TJvCCObjRecMgrClass = class of TJvCCObjRecMgr;

  TJvCCIntRecMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest;const src); override;
    procedure SwapItems(var p1, p2); override;
  end;

  TJvCCSmallIntMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest;const src); override;
    procedure SwapItems(var p1, p2); override;
  end;

  TJvCCDoubleMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest;const src); override;
    procedure SwapItems(var p1, p2); override;
  end;

  TJvCCSingleMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest;const src); override;
    procedure SwapItems(var p1, p2); override;
  end;

{$IFDEF USE_PSTRING}
  TJvCCString = PString;
{$ELSE}
  TJvCCString = string;
{$ENDIF}

  TJvCCStrRec = record
    nval : TJvCCString;
  end;
  PCCStrRec = ^TJvCCStrRec;

  TJvCCStrRecMgr = class(TJvCCRecordMgr)
  private
    FCaseSensitive : boolean;
  public
    constructor Create; override;
    procedure DestructItem(var item);override;
    procedure ConstructItem(var item);override;
    function  CompareItems(const p1, p2):integer;override;
    procedure SwapItems(var p1, p2); override;
    procedure AssignItem(var dest; const src);override;
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;

  TIntStrRec = record
    ival : integer;
    sval : TJvCCString;
  end;
  PIntStrRec = ^TIntStrRec;

  {TJvCCIntStrArray: In this class, the ival is used as an index for a string}
  TJvCCIntStrMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    procedure DestructItem(var item);override;
    procedure ConstructItem(var item);override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest; const src);override;
    procedure SwapItems(var p1, p2); override;
  end;

  TObjStrRec = record
    sval : TJvCCString;
    oval : TObject;
  end;

  PObjStrRec = ^TObjStrRec;
const
  kObjStrMax = MAXINT div sizeof(TObjStrRec) - 1;
type

  TObjStrBuf = array[0..kObjStrMax] of TObjStrRec;
  PObjStrBuf = ^TObjStrBuf;

  TJvCCStrObjMgr = class(TJvCCRecordMgr)
  private
    FCaseSensitive : boolean;
  public
    constructor Create; override;
    procedure DestructItem(var item);override;
    procedure ConstructItem(var item);override;
    function  CompareItems(const p1, p2):integer;override;
    procedure SwapItems(var p1, p2); override;
    procedure AssignItem(var dest; const src);override;
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;

  TJvCCStrStrRec = record
    nval : TJvCCString;
    sval : TJvCCString;
  end;
  PCCStrStrRec = ^TJvCCStrStrRec;

  TJvCCStrStrMgr = class(TJvCCRecordMgr)
  private
    FCaseSensitive : boolean;
  public
    constructor Create; override;
    procedure DestructItem(var item);override;
    procedure ConstructItem(var item);override;
    function  CompareItems(const p1, p2):integer;override;
    procedure SwapItems(var p1, p2); override;
    procedure AssignItem(var dest; const src);override;
    property CaseSensitive : boolean read FCaseSensitive write FCaseSensitive;
  end;

{string helper compatibility functions}
procedure InitCCStr(var rs : TJvCCString);
procedure KillCCStr(var rs : TJvCCString);
function GetCCStr(const rs : TJvCCString) : string;
procedure SetCCStr(var rs : TJvCCString; const sp : string);

{standard TObject compare function}
function ComparePtrValue(obj1, obj2 : TObject) : integer;

implementation

procedure ThrowNotImplemented(const s : string);
begin
  raise EJvCCRecordMgr.Create(
    Format('TJvCCBase class has not implemented %s',[s]));
end;

procedure ThrowNotImplemented2(const s : string);
begin
  raise EJvCCRecordMgr.Create(
    Format('TJvCCRecordMgr class has not implemented %s',[s]));
end;

constructor TJvCCBase.CreateMgr(reclass : TJvCCRecordMgrClass);
begin
  inherited Create;
  FRecMgr := reclass.Create;
end;

procedure  TJvCCBase.SetCount(inCount : TJvCCSize);
begin
  ThrowNotImplemented('SetCount');
end;

function TJvCCBase.AppendItem(const buf) : TJvCCItemPtr;
begin
  ThrowNotImplemented('AppendItem');
  Result := nil;
end;

function TJvCCBase.DeleteItem(const buf) : boolean;
begin
  Result := FALSE;
  ThrowNotImplemented('DeleteItem');
end;

procedure TJvCCBase.AppendFirst(const buf);
begin
  ThrowNotImplemented('AppendFirst');
end;

procedure TJvCCBase.AppendLast(const buf);
begin
  ThrowNotImplemented('AppendLast');
end;

function TJvCCBase.FirstItem:TJvCCItemPtr;
begin
  Result := nil;
end;

procedure TJvCCBase.FreeItem(var pitem : TJvCCItemPtr);
begin
  ThrowNotImplemented('FreeItem');
end;

function TJvCCBase.NextItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
begin
  Result := nil;
end;

function TJvCCBase.PriorItem(pitem : TJvCCItemPtr) : TJvCCItemPtr;
begin
  Result := nil;
end;

function TJvCCBase.LastItem:TJvCCItemPtr;
begin
  Result := nil;
end;

procedure TJvCCBase.CopyItem(const pitem : TJvCCItemPtr; var buf);
begin
  FRecMgr.AssignItem(buf, pitem^);
end;

procedure TJvCCBase.CallActionMethod(method: TJvCCActionMethod; data:Pointer);
var
  ip : TJvCCItemPtr;
  np : TJvCCItemPtr;
begin
  ip := FirstItem;
  while (ip <> nil) do
  begin
    np := NextItem(ip);
    method(ip,data);
    ip := np;
  end;
end;

procedure TJvCCBase.CallActionProcedure(proc: TJvCCActionProc; data:Pointer);
var
  ip : TJvCCItemPtr;
  np : TJvCCItemPtr;
begin
  ip := FirstItem;
  while (ip <> nil) do
  begin
    np := NextItem(ip);
    proc(ip,data);
    ip := np;
  end;
end;


function TJvCCBase.LocateItem(const rec; var buf) : boolean;
var
  p : TJvCCItemPtr;
begin
  p := nil;
  Result := LocateItemPtr(rec, p, its_forward);
  if (Result) then
  begin
    FRecMgr.AssignItem(buf, p^);
    Result := TRUE;
  end else
    Result := FALSE;
end;

function TJvCCBase.LocateItemPtr(
  const rec;
  var startItem : TJvCCItemPtr;
  direction : TJvCCItemSearch) : boolean;
var
  pitem : TJvCCItemPtr;
begin
  pitem := startItem;
  if (pitem = nil) then
  begin
    CASE direction of
      its_forward:
        pitem := FirstItem;
      its_backward:
        pitem := LastItem;
     ELSE
     begin
       Result := FALSE;
       exit;
     end;
    END;
  end;
  Result := FALSE;
  while(pitem<>nil) do
  begin
    if (FRecMgr.CompareItems(rec,pitem^) =0) then
    begin
      Result := TRUE;
      break;
    end;
    CASE direction of
      its_forward:
        pitem := NextItem(pitem);
      its_backward:
        pitem := PriorItem(pitem);
    END;
  end;
  if Result then
    startItem := pitem
  else
    startItem := nil;
end;


procedure TJvCCBase.Clear;
var
  ip : TJvCCItemPtr;
  np : TJvCCItemPtr;
begin
  ip := FirstItem;
  while (ip<>nil) do
  begin
    np := NextItem(ip);
    FreeItem(ip);
    ip := np;
  end;
end;


function TJvCCBase.LocateKey(key:Pointer; var buf) : boolean;
var
  p : TJvCCItemPtr;
begin
  p := nil;
  Result := LocateKeyPtr(key, p, its_forward);
  if (Result) then
  begin
    FRecMgr.AssignItem(buf, p^);
    Result := TRUE;
  end else
    Result := FALSE;
end;

function TJvCCBase.LocateKeyPtr(
  key : Pointer;
  var startItem : TJvCCItemPtr;
  direction : TJvCCItemSearch) : boolean;
var
  pitem : TJvCCItemPtr;
begin
  pitem := startItem;
  if (pitem = nil) then
  begin
    CASE direction of
      its_forward:
        pitem := FirstItem;
      its_backward:
        pitem := LastItem;
     ELSE
     begin
       Result := FALSE;
       exit;
     end;
    END;
  end;
  Result := FALSE;
  while(pitem<>nil) do
  begin
    if (FRecMgr.CompareKey(pitem^,key) =0) then
    begin
      Result := TRUE;
      break;
    end;
    CASE direction of
      its_forward:
        pitem := NextItem(pitem);
      its_backward:
        pitem := PriorItem(pitem);
    END;
  end;
  if Result then
    startItem := pitem
  else
    startItem := nil;
end;

destructor TJvCCBase.Destroy;
begin
  Clear;
  FRecMgr.Free;
  inherited;
end;

procedure TJvCCRecordMgr.InitRecordMgr(
  itemSize : integer;
  initItems : Boolean);
begin
  inherited Create;
  FItemSize := itemsize;
  FInitItems := initItems;
end;

constructor TJvCCRecordMgr.Create;
begin
  inherited Create;
end;

procedure TJvCCRecordMgr.DestructItem(var item);
begin
end;

procedure TJvCCRecordMgr.ConstructItem(var item);
begin
  FillChar(item,FItemSize,0);
end;
 
procedure TJvCCRecordMgr.AssignItem(var dest; const src);
begin
  System.Move(src,dest,FItemSize);
end;

procedure TJvCCRecordMgr.SwapItems(var p1, p2);
begin
  ThrowNotImplemented2('SwapItems');
end;

function  TJvCCRecordMgr.CompareKey(const item;key:Pointer):integer;
begin
  Result := 0;
  ThrowNotImplemented2('CompareKey');
end;

function TJvCCRecordMgr.CompareItems(const p1, p2) : integer;
var
  b1 : PByte;
  b2 : PByte;
  i : integer;
begin
  b1 := PByte(@p1);
  b2 := PByte(@p2);
  for i := 1 to FItemSize do
  begin
    if (b2^ <> b1^) then
    begin
      Result := (b1^ - b2^);
      exit;
    end;
    Inc(b1);
    Inc(b2);
  end;
  Result := 0;
end;

constructor TJvCCIntRecMgr.Create;
begin
  inherited;
  InitRecordMgr(sizeof(integer),TRUE);
end;

function  TJvCCIntRecMgr.CompareItems(const p1, p2):integer;
begin
  Result := integer(p1) - integer(p2);
end;

procedure TJvCCIntRecMgr.SwapItems(var p1, p2);
var
  temp : integer;
begin
  temp := integer(p1);
  integer(p1) := integer(p2);
  integer(p2) := temp;
end;

procedure TJvCCIntRecMgr.AssignItem(var dest;const src);
begin
  integer(dest) := integer(src);
end;

constructor TJvCCSmallIntMgr.Create;
begin
  inherited;
  InitRecordMgr(sizeof(SmallInt),TRUE);
end;

function  TJvCCSmallIntMgr.CompareItems(const p1, p2):integer;
begin
  Result := SmallInt(p1) - SmallInt(p2);
end;

procedure TJvCCSmallIntMgr.SwapItems(var p1, p2);
var
  temp : SmallInt;
begin
  temp := SmallInt(p1);
  SmallInt(p1) := SmallInt(p2);
  SmallInt(p2) := temp;
end;

procedure TJvCCSmallIntMgr.AssignItem(var dest;const src);
begin
  SmallInt(dest) := SmallInt(src);
end;

constructor TJvCCPtrRecMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(Pointer),TRUE);
end;


constructor TJvCCObjRecMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TObject),TRUE);
  ObjCompare := ComparePtrValue;
end;

procedure TJvCCObjRecMgr.AssignItem(var dest;const src);
begin
  if FAutoDestroy then
    TObject(dest).Free;
  TObject(dest) := TObject(src);
end;


procedure TJvCCObjRecMgr.ConstructItem(var item);
begin
  TObject(item) := nil;
end;

procedure TJvCCObjRecMgr.DestructItem(var item);
begin
  if FAutoDestroy then
    TObject(item).Free;
  TObject(item) := nil;
end;

procedure TJvCCObjRecMgr.SwapItems(var p1, p2);
var
  temp : TObject;
begin
  temp := TObject(p1);
  TObject(p1) := TObject(p2);
  TObject(p2) := temp;
end;

function ComparePtrValue(obj1, obj2 : TObject) : integer;
begin
  Result := ORD(PChar(obj1) - PChar(obj2));
end;

function  TJvCCObjRecMgr.CompareKey(const item;key:Pointer):integer;
begin
  if not assigned(FKeyCompare) then
    raise EJvCCRecordMgr.create('Nil KeyCompare function');
  Result := FKeyCompare(TObject(item),key);
end;

function  TJvCCObjRecMgr.CompareItems(const p1, p2):integer;
begin
  if not assigned(FObjCompare) then
    raise EJvCCRecordMgr.create('Nil ObjCompare function');
  Result := FObjCompare(TObject(p1),TObject(p2));
end;

constructor TJvCCStrRecMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TJvCCStrRec),TRUE);
end;

procedure TJvCCStrRecMgr.SwapItems(var p1, p2);
var
  rec:TJvCCStrRec;
begin
  ConstructItem(rec);
  rec.nval :=  TJvCCStrRec(p1).nval;
  TJvCCStrRec(p1).nval := TJvCCStrRec(p2).nval;
  TJvCCStrRec(p2).nval := rec.nval;
  DestructItem(rec);
end;

procedure TJvCCStrRecMgr.DestructItem(var item);
var
  p : PCCStrRec;
begin
  p := PCCStrRec(@item);
  KillCCStr(p^.nval);
end;

procedure TJvCCStrRecMgr.AssignItem(var dest; const src);
begin
{$IFDEF USE_PSTRING}
  AssignStr(TJvCCStrRec(dest).nval,TJvCCStrRec(src).nval^ );
{$ELSE}
  TJvCCStrRec(dest).nval := TJvCCStrRec(src).nval;
{$ENDIF}
end;


procedure TJvCCStrRecMgr.ConstructItem(var item);
begin
{$IFDEF USE_PSTRING}
  TJvCCStrRec(item).nval := nil;
{$ELSE}
  Pointer(TJvCCStrRec(item).nval) := nil;
{$ENDIF}
end;


function  TJvCCStrRecMgr.CompareItems(const p1, p2):integer;
begin
{$IFDEF USE_PSTRING}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrRec(p1).nval^,TJvCCStrRec(p2).nval^)
  else
    Result := CompareText(TJvCCStrRec(p1).nval^,TJvCCStrRec(p2).nval^);
{$ELSE}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrRec(p1).nval,TJvCCStrRec(p2).nval)
  else
    Result := CompareText(TJvCCStrRec(p1).nval,TJvCCStrRec(p2).nval);
{$ENDIF}
end;

procedure TJvCCIntStrMgr.DestructItem(var item);
begin
  KillCCStr(TIntStrRec(item).sval);
end;


procedure TJvCCIntStrMgr.ConstructItem(var item);
begin
  TIntStrRec(item).ival := 0;
  InitCCStr(TIntStrRec(item).sval);
end;

procedure TJvCCIntStrMgr.AssignItem(var dest; const src);
begin
  TIntStrRec(dest).ival := TIntStrRec(src).ival;
{$IFDEF USE_PSTRING}
  AssignStr(TIntStrRec(dest).sval,(TIntStrRec(src).sval)^);
{$ELSE}
  TIntStrRec(dest).sval := TIntStrRec(src).sval;
{$ENDIF}
end;

procedure TJvCCIntStrMgr.SwapItems(var p1, p2);
var
  isr:TIntStrRec;
begin
  ConstructItem(isr);
  AssignItem(isr,p1);
  AssignItem(p1,p2);
  AssignItem(p2,isr);
  DestructItem(isr);
end;

constructor TJvCCIntStrMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TIntStrRec),TRUE);
end;


function  TJvCCIntStrMgr.CompareItems(const p1, p2):integer;
begin
  Result := TIntStrRec(p1).ival - TIntStrRec(p2).ival;
end;


procedure TJvCCStrObjMgr.DestructItem(var item);
begin
  KillCCStr(TObjStrRec(item).sval);
end;

procedure TJvCCStrObjMgr.ConstructItem(var item);
begin
  InitCCStr(TObjStrRec(item).sval);
  TObjStrRec(item).oval := nil;
end;

procedure TJvCCStrObjMgr.SwapItems(var p1, p2);
var
  osr : TObjStrRec;
begin
  ConstructItem(osr);
  AssignItem(osr,p1);
  AssignItem(p1,p2);
  AssignItem(p2,osr);
  DestructItem(osr);
end;

function  TJvCCStrObjMgr.CompareItems(const p1, p2):integer;
begin
{$IFDEF USE_PSTRING}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrStrRec(p1).nval^,TJvCCStrStrRec(p2).nval^)
  else
    Result := CompareText(TJvCCStrStrRec(p1).nval^,TJvCCStrStrRec(p2).nval^);
{$ELSE}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrStrRec(p1).nval,TJvCCStrStrRec(p2).nval)
  else
    Result := CompareText(TJvCCStrStrRec(p1).nval,TJvCCStrStrRec(p2).nval);
{$ENDIF}
end;

procedure TJvCCStrObjMgr.AssignItem(var dest; const src);
begin
  TObjStrRec(dest).oval := TObjStrRec(src).oval;
{$IFDEF USE_PSTRING}
  AssignStr(TObjStrRec(dest).sval,(TObjStrRec(src).sval)^);
{$ELSE}
  TObjStrRec(dest).sval := TObjStrRec(src).sval;
{$ENDIF}
end;

constructor TJvCCStrObjMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TObjStrRec),TRUE);
end;


constructor TJvCCStrStrMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TJvCCStrStrRec),TRUE);
end;

procedure TJvCCStrStrMgr.DestructItem(var item);
begin
  KillCCStr(TJvCCStrStrRec(item).nval);
  KillCCStr(TJvCCStrStrRec(item).sval);
end;

procedure TJvCCStrStrMgr.ConstructItem(var item);
begin
  InitCCStr(TJvCCStrStrRec(item).nval);
  InitCCStr(TJvCCStrStrRec(item).sval);
end;

procedure TJvCCStrStrMgr.SwapItems(var p1, p2);
var
  rec:TJvCCStrStrRec;
begin
  ConstructItem(rec);
  AssignItem(rec,p1);
  AssignItem(p1,p2);
  AssignItem(p2,rec);
  DestructItem(rec);
end;

procedure TJvCCStrStrMgr.AssignItem(var dest; const src);
begin
{$IFDEF USE_PSTRING}
  AssignStr(TJvCCStrStrRec(dest).nval,TJvCCStrStrRec(src).nval^);
  AssignStr(TJvCCStrStrRec(dest).sval,TJvCCStrStrRec(src).sval^);
{$ELSE}
  TJvCCStrStrRec(dest).nval := TJvCCStrStrRec(src).nval;
  TJvCCStrStrRec(dest).sval := TJvCCStrStrRec(src).sval;
{$ENDIF}
end;

function TJvCCStrStrMgr.CompareItems(const p1, p2):integer;
begin
{$IFDEF USE_PSTRING}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrStrRec(p1).nval^,TJvCCStrStrRec(p2).nval^)
  else
    Result := CompareText(TJvCCStrStrRec(p1).nval^,TJvCCStrStrRec(p2).nval^);
{$ELSE}
  if FCaseSensitive then
    Result := CompareStr(TJvCCStrStrRec(p1).nval,TJvCCStrStrRec(p2).nval)
  else
    Result := CompareText(TJvCCStrStrRec(p1).nval,TJvCCStrStrRec(p2).nval);
{$ENDIF}
end;

constructor TJvCCDoubleMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(double),FALSE);
end;

function  TJvCCDoubleMgr.CompareItems(const p1, p2):integer;
var
  diff : double;
begin
  diff := double(p1) - double(p2);{a sign function would be nice}
  if (diff > 0) then
    Result := 1
  else if (diff < 0) then
    Result := -1
  else
    Result := 0;
end;

procedure TJvCCDoubleMgr.AssignItem(var dest;const src);
begin
  double(dest) := double(src);
end;

procedure TJvCCDoubleMgr.SwapItems(var p1, p2);
var
  temp:double;
begin
  temp := double(p1);
  double(p1) := double(p2);
  double(p2) := temp;
end;

constructor TJvCCSingleMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(Single),FALSE);
end;

function  TJvCCSingleMgr.CompareItems(const p1, p2):integer;
var
  diff : Single;
begin
  diff := Single(p1) - Single(p2);{a sign function would be nice}
  if (diff > 0) then
    Result := 1
  else if (diff < 0) then
    Result := -1
  else
    Result := 0;
end;

procedure TJvCCSingleMgr.AssignItem(var dest;const src);
begin
  Single(dest) := Single(src);
end;

procedure TJvCCSingleMgr.SwapItems(var p1, p2);
var
  temp:Single;
begin
  temp := Single(p1);
  Single(p1) := Single(p2);
  Single(p2) := temp;
end;

procedure InitCCStr(var rs : TJvCCString);
begin
{$IFDEF USE_PSTRING}
  rs := nil;
{$ELSE}
  Initialize(rs);
{$ENDIF}
end;

procedure KillCCStr(var rs : TJvCCString);
begin
{$IFDEF USE_PSTRING}
  DisposeStr(rs);
{$ELSE}
  Finalize(rs);
{$ENDIF}
end;


procedure SetCCStr(var rs : TJvCCString; const sp : string);
begin
{$IFDEF USE_PSTRING}
  AssignStr(rs,sp);
{$ELSE}
  rs := sp;
{$ENDIF}
end;

function GetCCStr(const rs : TJvCCString) : string;
begin
{$IFDEF USE_PSTRING}
  Result := rs^;
{$ELSE}
  Result := rs;
{$ENDIF}
end;


end.




