unit JvDirList;

interface
Uses Classes,SysUtils,JvCCRecrd,JvCCArray,JvRnnLogTx;
{$I JCL.INC}
type
  EFileSpec = class(Exception);

  TJvRnnSortFCList = (sort_path, sort_name,
    sort_fromtime, sort_totime, sort_type,
    sort_fromsize, sort_fromlocked,
    sort_tosize, sort_tolocked);

  PFileSpec = ^TFileSpec;
  TFileSpec = packed record
    size : LongInt;
    time : TDateTime;
    attr : WORD;
    fsflags : WORD;
  end;

  PFileInfo = ^TFileInfo;
  TFileInfo = packed record
    filename : TJvCCString;
    spec : TFileSpec;
  end;

const
  kMaxFileInfoArray = (MAXINT div sizeof(TFileInfo))-1;
type
  TFileInfoArray = packed array[0..kMaxFileInfoArray] of TFileInfo;
  PFileInfoArray = ^TFileInfoArray;

  TJvRnnFileInfoMgr = class(TJvCCRecordMgr)
    constructor create; override;
    procedure ConstructItem(var item); override;
    procedure DestructItem(var item); override;
    procedure AssignItem(var dest;const src); override;
    procedure SwapItems(var p1, p2); override;
    function  CompareItems(const p1, p2 ):integer;override;
  end;
  
  TFileInfoList = class(TJvCCArray)
  private
    FArrayPtr : PFileInfoArray;
  public
    constructor Create;
    procedure AddFileInfo( const inRec : TFileInfo);
    function FindFileName(const inName : string) : integer;
  end;

  TSourceDir = class(TFileInfoList)
    FDirectory : String;
    FFilter : String;
    FParent : TSourceDir;
    FRoot : PString;
    FDirSpec : TFileSpec;
    FSubDirList : TList;
    FInvalid : boolean;
    FDoSubDirs : boolean;

    constructor Create(const inDir, inFilter : string;
      const inspec : TFileSpec;
      inparent : TSourceDir);
    destructor Destroy; override;
    procedure FreeSubDirs;
    function  GetSubDir(const inDir : string) : TSourceDir;
    function NewSubDir(const inDir, inFilter : string; const inspec : TFileSpec) : TSourceDir;
    procedure SetFilter(const inFilter:string);
    procedure SetAbsPath(const inDir : string);
    function GetAbsPath : string;
    function GetSubPath : string;
    function GetRootSourceDir : TSourceDir;
    function GetRootPath : string;
    procedure SetDoSubDirs(doSubDirs:boolean);
    procedure DeleteFiles(delsubdirfiles : boolean);
    procedure DeleteSubDirs;
    function FindInTree(const fileSubPath:string; var info : TFileInfo) : boolean;
    procedure Refresh;
    procedure RefreshSubDirs;
    procedure SetParent(pdir:TSourceDir);
    procedure AddSubDir(srcDir:TSourceDir);

    property Filter : string read FFilter write SetFilter;
    property Parent : TSourceDir read FParent write SetParent;
    property AbsPath : string read GetAbsPath write SetAbsPath;
    property SubPath : string read GetSubPath;
    property DoSubDirs : boolean read FDoSubDirs write SetDoSubDirs;
    property Invalid : boolean read FInvalid;
  end;

  TSrcTreeIterator = class(TObject)
  protected
    FSrcDir : TSourceDir;
    FCurDir : TSourceDir;
    FIXArray : TJvCCIntArray;
    FIXIX : integer;
    FFileix : integer;
  public
    constructor create(src : TSourceDir);
    destructor Destroy; override;
    function GetSubPath : string;
    function GetAbsPath : string;
    function GetCurFilename : string;
    function HasStarted : boolean;
    procedure DeleteCurrent;
    function FirstFile(var info:TFileInfo):boolean;
    function NextFile(var info:TFileInfo):boolean;
    property AbsPath : string read GetAbsPath;
    property SubPath : string read GetSubPath;
    property FileName : string read GetCurFilename;
  end;

type
  PCopyRec = ^TCopyRec;
  TCopyRec = packed record
    subpath : TJvCCString;
    v1 : TFileSpec;
    v2 : TFileSpec;
    crFlags : WORD;
  end;

const
  kMaxCopyRecArray = (MAXINT div sizeof(TCopyRec))-1;
  kA_AND_B = 1;
  kA_NOT_B = 2;
  kB_NOT_A = 4;
  kA_NEWER_B = 8;
  kB_NEWER_A = 16;
  kAB_SAMETIME = 32;
  kAB_DIFFERENT = 64;

  kALL_CONDITIONS = kA_AND_B + kA_NOT_B + kB_NOT_A +
    kA_NEWER_B + kB_NEWER_A + kAB_SAMETIME + kAB_DIFFERENT;

  kFS_Exists = 1;

type
  TCopyRecPath = (path_A, path_B);
  TCopyRecArray = packed array[0..kMaxCopyRecArray] of TCopyRec;
  PCopyRecArray = ^TCopyRecArray;
  TJvRnnDirTreeWalk = procedure(const crec : TCopyRec) of Object;

  TJvRnnFileCprMgr = class(TJvCCRecordMgr)
  protected
    FSortType : TJvRnnSortFCList;
  public
    constructor Create; override;
    procedure ConstructItem(var item);override;
    procedure DestructItem(var item);override;
    function  CompareItems(const p1, p2):integer;override;
  end;

  TCompareArray = class(TJvCCArray)
  private
    FArrayPtr : PCopyRecArray;
    FDirA : string;
    FDirB : string;
    FFilter : string;
    FConditions : integer;
    FRecurse : boolean;
  protected
    procedure SetSortType(stype : TJvRnnSortFCList);
    function GetSortType :  TJvRnnSortFCList;
  public
    constructor Create(const rootA,rootB : string);
    procedure AddCopyRec(const copyRec : TCopyRec);
    function  FindSubPathIndex(const insubpath:string) : integer;
    function  GetSpec(const insubpath:string; recpath:TCopyRecPath;
      var spec : TFileSpec) : boolean;
    procedure DoSort;
    procedure DoCompare;
    procedure SetDir_A(const inA : string);
    procedure SetDir_B(const inB : string);

    property ArrayPtr : PCopyRecArray read FArrayPtr;
    property SortType : TJvRnnSortFCList read GetSortType write SetSortType;

    property Dir_A : string read FDirA write SetDir_A;
    property Dir_B : string read FDirB write SetDir_B;
    property Recurse : boolean read FRecurse write FRecurse default TRUE;
    property Filter : string read FFilter write FFilter;
    property Conditions : integer read FConditions write FConditions;
  end;

function FI_IsDirectory(const fi:TFileInfo) : boolean;
procedure DeleteFileMask(const inPath : string);
procedure DeleteDirTree(const inPath : string);
procedure SetFileInfo(var info : TFileInfo; const inrec : TSearchRec);
function GetFileSpec(const infile : string; var spec : TFileSpec) : boolean;
function GetDirectoryAttr(const Name: string; var attrCode : integer): Boolean;  {derived from FileCtrl}
function IsVolume(const inPath : string) : boolean;

function CreateDirList(inparent: TSourceDir;
  const inName, filters : string;
  recurse : boolean) : TSourceDir;

procedure CompareWalkDirectories(const dirA, dirB, filter : string;
                                  conditions : integer;
                                  paction : TJvRnnDirTreeWalk;
                                  recurse : boolean
                                  );
function CompareFilePaths(
  const prec, qrec : TCopyRec; sortType : TJvRnnSortFCList) : integer;
function GetAttrString( attr : WORD) : string;
const
  faNormalFile = $00000080; { - for the normal files with no attribute set}
  faEverything =  faNormalFile + faAnyFile;
  faAvoidLikePlague = faReadOnly + faHidden + faSysFile + faVolumeID;
  kDirSeparator = '\';
{$IFDEF LDEBUG}
var
  logout : TJvRnnLogTextFile;
{$ENDIF}

implementation
Uses JvRnnStrUt, JvSplatFile, FileCtrl, WinTypes, WinProcs, Dialogs;

function FI_IsDirectory(const fi : TFileInfo) : boolean;
begin
  Result := ((fi.spec.attr and faDirectory) <> 0);
end;

function FI_IsChildDirectory(const fi : TFileInfo) : boolean;
begin
  Result := FI_IsDirectory(fi);
  if (Result) then
    Result := (GetCCStr(fi.filename) <> '.') AND (GetCCStr(fi.filename) <> '..');
end;
 (*
procedure Err_nodirectory(const s : string);
begin
  raise EFileSpec.Create('Directory not found '+s);
end; *)

function IsVolume(const inPath : string) : boolean;
begin
  Result := COMPARETEXT(ExtractFileDrive(inPath), inPath)=0;
end;

procedure DeleteDirTree(const inPath : string);
var
  dellist :	TSourceDir;
  inDir : string;
  inFilter : string;
  delfile : string;
  inspec : TFileSpec;
  iter : TSrcTreeIterator;
  info : TFileInfo;
begin
  inDir := inPath;
  RemoveEndChar(inDir,kDirSeparator);
  GetFileSpec(inDir, inspec);
  inFilter := '*.*';
  dellist := TSourceDir.Create(inDir, inFilter, inspec, nil);
  dellist.DoSubDirs := TRUE;
  iter := nil;
{$IFDEF USE_PSTRING}
  InitCCStr(info.filename);
{$ENDIF}
  try
    dellist.Refresh;
    iter := TSrcTreeIterator.Create(dellist);
    if (iter.FirstFile(info)) then
    repeat
      delfile := dellist.AbsPath + kDirSeparator + GetCCStr(info.filename);
      RubOutFile(delfile);
    until not iter.NextFile(info);
  finally
{$IFDEF USE_PSTRING}
    KillCCStr(info.filename);
{$ENDIF}
    iter.free;
    dellist.free;
  end;
end;

procedure DeleteFileMask(const inPath : string);
var
  dellist :	TSourceDir;
  inDir : string;
  inFilter : string;
  delfile : string;
  inspec : TFileSpec;
  iter : TSrcTreeIterator;
  info : TFileInfo;
begin
  inDir := ExtractFilePath(inPath);
  RemoveEndChar(inDir,kDirSeparator);
  GetFileSpec(inDir, inspec);
  inFilter := ExtractFileName(inPath);
  dellist := TSourceDir.Create(inDir, inFilter, inspec, nil);
  dellist.DoSubDirs := FALSE;
  iter := nil;
  {$IFDEF USE_PSTRING}
  InitCCStr(info.filename);
  {$ENDIF}
  try
    dellist.Refresh;
    iter := TSrcTreeIterator.Create(dellist);
    if (iter.FirstFile(info)) then
    repeat
      delfile := dellist.AbsPath + kDirSeparator + GetCCStr(info.filename);
      RubOutFile(delfile);
    until not iter.NextFile(info);
  finally
  {$IFDEF USE_PSTRING}
    KillCCStr(info.filename);
  {$ENDIF}
    iter.free;
    dellist.free;
  end;
end;

function NameFitsFilter(const inName,inFilter : string) : boolean;
var
  k : integer;
  ix : Integer;
  n : integer;
  nextNonWild : Char;
  function CharsMatch(x, y : Char):boolean;
  begin
    if (ORD(x) >= ORD('a')) AND (ORD(x) <= ORD('z'))  then
      x := CHAR(ORD(x) - ORD('a') + ORD('A'));
    if (ORD(y) >= ORD('a')) AND (ORD(y) <= ORD('z'))  then
      y := CHAR(ORD(y) - ORD('a') + ORD('A'));
    Result := (x = y);
  end;

  procedure GetNextNonWild;
  begin
    nextNonWild  := #0;
    k := ix+1;
    while (k <= Length(inFilter)) do
    begin
      if (inFilter[k] <> '*') AND (inFilter[k] <> '?') then
      begin
        nextNonWild := inFilter[k];
        break;
      end;
      Inc(k);
    end;
  end;

begin
  Result := FALSE;
  n := 0;
  ix := 1;
  nextNonWild := #0;
    
  while (TRUE) do
  begin
    Inc(n);
    if (n > Length(inName)) then
      break;
    if (ix > Length(inFilter)) then
      break;
    CASE inFilter[ix] of
      '*':
        begin {   }
          if ix >= Length(inFilter) then
          begin
            Result := TRUE;
            exit; {fits everything left}
          end;
          GetNextNonWild;
          if (nextNonWild <> #0) then
          begin
            if CharsMatch(nextNonWild,inName[n]) then
              ix := k+1
              else if (inName[n] = '.') then
                exit;
          end;
        end;
      '?':
        Inc(ix);
    ELSE
      begin
        if CharsMatch(inFilter[ix],inName[n]) then
          Inc(ix)
        else
          exit;
      end;
    END; {CASE}
  end;
  Result := not ((n <= Length(inName)) or (ix <= Length(inFilter)));
end;

function GetAttrString( attr : WORD) : string;
begin
  Result := '';
  if ((attr and faArchive) <> 0)
    then Result := Result + 'A';
  if ((attr and faReadOnly) <> 0)
    then Result := Result + 'R';
  if ((attr and faHidden) <> 0)
    then Result := Result + 'H';
  if ((attr and faSysFile) <> 0)
    then Result := Result + 'S';
  if ((attr and faVolumeID) <> 0)
    then Result := Result + 'V';
  if ((attr and faDirectory) <> 0)
    then Result := Result + 'D';
end;

procedure SetFileInfo(var info : TFileInfo; const inrec : TSearchRec);
var
  s : string;
begin
  WITH info do
  begin
    SetCCStr(filename,inrec.name);
    spec.size := inrec.size;
    try
      spec.time := FileDateToDateTime(inrec.time)
      {if (inrec.time > 0) then
      else begin
        s := Format('nil time for %s %s',[inrec.name, GetAttrString(inrec.attr)]);
        ShowMessage(s);
        spec.time := 0;
      end;}
    except
      on x1:Exception do
      begin
        s := Format('error time %d for %s %s',[inrec.time, inrec.name,GetAttrString(inrec.attr)]);
        ShowMessage(s);
      end;
    end;
    spec.attr := inrec.attr;
    if (inrec.time <> 0) OR (inrec.size > 0) OR (inrec.attr <> 0) then
      spec.fsflags := kFS_Exists
    else
      spec.fsflags := 0;
  end;
end;


function GetFileSpec(const infile : string; var spec : TFileSpec) : boolean;
var
  srec : TSearchRec;
  info : TFileInfo;
begin
  FillChar(srec,sizeof(srec),0);
  if (FindFirst(infile, faAnyFile, srec) = 0) then
  begin
  {$IFDEF USE_PSTRING}
    InitCCStr(info.filename);
    try
  {$ENDIF}
    SetFileInfo(info, srec);
  {$IFDEF USE_PSTRING}
    finally
    KillCCStr(info.filename);
    end;
  {$ENDIF}
    spec := info.spec;
    Result := TRUE;
  end else
    Result := FALSE;
  if Result then
    SysUtils.FindClose(srec);
end;


constructor TFileInfoList.Create;
begin
  inherited CreateMgr(TJvRnnFileInfoMgr);
  BlockHdl := @FArrayPtr;
  Count := 0;
  Sorted := TRUE;
end;

procedure TFileInfoList.AddFileInfo(const inRec : TFileInfo);
var
  ix : integer;
begin
  if not GetItemIndex(ix,inRec) then
  begin
    SELF.InsertInto(ix,inRec);
  end else
    raise EFileSpec.create('File exists already');
end;


constructor TJvRnnFileInfoMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TFileInfo),TRUE);
end;

procedure TJvRnnFileInfoMgr.ConstructItem(var item);
begin
{$IFDEF USE_PSTRING}
  InitCCStr(TFileInfo(item).filename);
  FillChar(TFileInfo(item).spec, sizeof(TFileSpec),0);
{$ELSE}
  Initialize(TFileInfo(item));
{$ENDIF}
end;

procedure TJvRnnFileInfoMgr.DestructItem(var item); 
begin
{$IFDEF USE_PSTRING}
  KillCCStr(TFileInfo(item).filename);
{$ELSE}
  Finalize(TFileInfo(item));
{$ENDIF}
end;

procedure TJvRnnFileInfoMgr.AssignItem(var dest;const src);
begin
{$IFDEF USE_PSTRING}
  AssignStr(TFileInfo(dest).filename,TFileInfo(src).filename^);
{$ELSE}
  TFileInfo(dest).filename :=  TFileInfo(src).filename;
{$ENDIF}
  TFileInfo(dest).spec := TFileInfo(src).spec;
end;

procedure TJvRnnFileInfoMgr.SwapItems(var p1, p2);
var
  temp : string;
  tempspec : TFileSpec;
begin
  temp := GetCCStr(TFileInfo(p1).filename);
{$IFDEF USE_PSTRING}
  AssignStr(TFileInfo(p1).filename,TFileInfo(p2).filename^);
{$ELSE}
  TFileInfo(p1).filename := TFileInfo(p2).filename;
{$ENDIF}
  SetCCStr(TFileInfo(p2).filename,temp);
  tempspec :=  TFileInfo(p1).spec;
  TFileInfo(p1).spec := TFileInfo(p2).spec;
  TFileInfo(p2).spec := tempspec;
end;


function  TJvRnnFileInfoMgr.CompareItems(const p1, p2):integer;
begin
  Result := COMPARETEXT(
    GetCCStr(TFileInfo(p1).filename),GetCCStr(TFileInfo(p2).filename));
end;

function TFileInfoList.FindFileName(const inName : string) : integer;
var
  i : integer;
  info : TFileInfo;
begin
{$IFDEF USE_PSTRING}
  InitCCStr(info.filename);
  try
{$ENDIF}
  SetCCStr(info.filename,inName);
  if not GetItemIndex(i, info) then
    Result := -1
  else
    Result := i;
{$IFDEF USE_PSTRING}
  finally
    KillCCStr(info.filename);
  end;
{$ENDIF}
end;

constructor TSourceDir.Create(const inDir, inFilter : string;
  const inspec : TFileSpec;
  inparent : TSourceDir);
begin
  inherited Create;
  FDirSpec := inspec;
  FSubDirList := nil;
  Parent := inparent;
  AbsPath := inDir;
  Filter := inFilter;
end;

function TSourceDir.GetRootSourceDir : TSourceDir;
var
  lastParent : TSourceDir;
begin
  lastParent := SELF;
  repeat
    Result := lastParent;
    lastParent := Result.FParent;
  until (lastParent = nil);
end;

destructor TSourceDir.Destroy;
begin
  FreeSubDirs;
  inherited Destroy;
end;

procedure TSourceDir.SetParent(pdir:TSourceDir);
begin
  FParent := parent;
  if FParent <> nil then
  begin
    FRoot := FParent.FRoot;
    pdir.AddSubDir(SELF);
  end
  else
    FRoot := @FDirectory;
end;

function  TSourceDir.GetSubDir(const inDir : string) : TSourceDir;
var
  ix : integer;
begin
  if FSubDirList <> nil then
  for ix := 0 to FSubDirList.Count-1 do
  begin
    Result := TSourceDir(FSubDirList[ix]);
    if COMPARETEXT(Result.AbsPath,inDir) = 0 then
      exit;
  end;
  Result := nil;
end;


procedure TSourceDir.FreeSubDirs;
var
  i : integer;
  child : TSourceDir;
begin
  if FSubDirList <> nil then
  begin
    for i := 0 to FSubDirList.Count-1 do
    begin
      child := TSourceDir(FSubDirList[i]);
      child.Free;
    end;
    FSubDirList.Free;
    FSubDirList := nil;
  end;
end;

function GetFilters(const inFilter : string; slist : TStrings) : integer;
var
  i : integer;
  p : integer;
  s : string;
begin
  p := 1;
  i := 1;
  while p <= Length(inFilter) do
  begin
    if (inFilter[p] = ';') or (p=Length(inFilter))then
    begin
      if p=Length(inFilter) then
        Inc(p);
      s := Copy(inFilter,i,p-i);
      slist.Add(s);
      i := p+1;
    end;
    Inc(p);
  end;
  Result := slist.Count;
end;


procedure TSourceDir.SetDoSubDirs(doSubDirs:boolean);
begin
  FDoSubDirs := doSubDirs;
  FInvalid := TRUE;
end;

procedure TSourceDir.SetFilter(const inFilter:string);
begin
  FFilter := inFilter;
  FInvalid := TRUE;
end;

function TSourceDir.GetAbsPath : string;
begin
  if (FParent <> nil) then begin
    Result := Format('%s'+kDirSeparator+'%s',[FRoot^ ,FDirectory]);
  end else
    Result := FDirectory;
end;

function TSourceDir.GetSubPath : string;
begin
  if (FParent <> nil) then
    Result := FDirectory
  else
    Result := '';
end;


procedure TSourceDir.DeleteSubDirs;
var
  saveFilter : string;
  sub : TSourceDir;
begin
  saveFilter := FFilter;
  SELF.DoSubDirs := TRUE;
  try
    FFilter := '*.*';
    SELF.Refresh;
    while FSubDirList.Count > 0 do
    begin
      sub := TSourceDir(FSubDirList[0]);
      sub.DeleteSubDirs;
      FSubDirList.Delete(0);
      sub.Free;
    end;
  finally
    FFilter := saveFilter;
  end;
end;

procedure TSourceDir.DeleteFiles(delsubdirfiles : boolean);
var
  delfile : string;
  iter : TSrcTreeIterator;
  info : TFileInfo;
begin
  SELF.DoSubDirs := delsubdirfiles;
  iter := nil;
  {$IFDEF USE_PSTRING}
  InitCCStr(info.filename);
  {$ENDIF}
  try
    SELF.Refresh;
    iter := TSrcTreeIterator.Create(SELF);
    if (iter.FirstFile(info)) then
    repeat
      delfile := SELF.AbsPath + kDirSeparator + GetCCStr(info.filename);
        RubOutFile(delfile);
    until not iter.NextFile(info);
  finally
  {$IFDEF USE_PSTRING}
    KillCCStr(info.filename);
  {$ENDIF}
    iter.free;
  end;
end;

function TSourceDir.GetRootPath : string;
begin
  Result := FRoot^;
end;

procedure TSourceDir.SetAbsPath(const inDir : string);
var
  ix : integer;
  ip : Integer;
begin
  if FParent <> nil then
  begin
    ix := CIPOS(FRoot^,inDir);
    if (ix <= 0) then
      raise EFileSpec.create(inDir+' is not a subdirectory of ' + FRoot^);
    ip := Length(FRoot^)+1;
    FDirectory := Copy(inDir,ip+1,Length(inDir)-ip);
  end else
    FDirectory := inDir;
  FInvalid := TRUE;
end;

procedure TSourceDir.AddSubDir(srcDir:TSourceDir);
begin
  if FSubDirList = nil then
    FSubDirList := TList.Create;
  FSubDirList.Add(srcDir);
end;

function  TSourceDir.NewSubDir(const inDir, inFilter : string; const inSpec : TFileSpec) : TSourceDir;
begin
  Result := TSourceDir.Create(inDir,inFilter,inSpec,SELF);
  AddSubDir(Result);
end;

function TSourceDir.FindInTree(const fileSubPath:string; var info : TFileInfo) : boolean;
var
  path : string;
  i : integer;
  subdir : TSourceDir;
begin
  Result := FALSE;
  SetCCStr(info.filename,ExtractFileName(fileSubPath));
  path := ExtractFilePath(fileSubPath);
  RemoveEndChar(path,kDirSeparator);
  if COMPARETEXT(path,SubPath) = 0 then
  begin
    Result := GetItemIndex(i,info);
    if Result then
      GetItem(i,info);
  end else if (FSubDirList <> nil) then
  begin
    for i := 0 to FSubDirList.Count-1 do
    begin
      subdir := TSourceDir(FSubDirList[i]);
      if (CIPOS(subdir.SubPath,path) > 0) then
      begin
        Result := subdir.FindInTree(fileSubPath,info);
        if Result then
          exit;
      end;
    end;
  end;
end;


procedure TSourceDir.RefreshSubDirs;
var
  i : integer;
  subdir : TSourceDir;
begin
  if FSubDirList <> nil then
  WITH FSubDirList do
  for i := 0 to count-1 do
  begin
    subdir := TSourceDir(FSubDirList[i]);
    subdir.Refresh;
  end;
end;

function GetDirectoryAttr(const Name: string; var attrCode : integer): Boolean;
{derived from FileCtrl}
{$IFNDEF WIN32}
const
  FILE_ATTRIBUTE_DIRECTORY = faDirectory;
{$ENDIF}
begin
  {$IFDEF WIN32}
  attrCode := GetFileAttributes(PChar(Name));
  Result := (attrCode <> -1) and (FILE_ATTRIBUTE_DIRECTORY and attrCode <> 0);
  {$ELSE}
  attrCode := FileGetAttr(Name);
  if (attrCode < 0) then
  begin
    Result := IsVolume(Name);
    if not Result then
      if DirectoryExists(Name) then
      begin
        attrCode := faDirectory;
        Result := TRUE;
      end;
  end
  else
    Result := ((attrCode and faDirectory) <> 0);
  {$ENDIF}
end;

function CreateDirList(inparent:TSourceDir; const inName, filters : string;
  recurse : boolean) : TSourceDir;
var
  spec : TFileSpec;
  attrcode : integer;
begin
  if GetDirectoryAttr(inName,attrcode) then
  begin
    if not IsVolume(inName) then
    begin
      GetFileSpec(inName, spec);
      if spec.time = 0 then
        ShowMessage('Create dir list time 0 for '+inName);
    end
    else
    begin
      FillChar(spec,sizeof(spec),#0);
      spec.attr := attrcode;
    end;
    Result := TSourceDir.Create(inName, filters,spec,inparent);
    try
      Result.DoSubDirs := recurse;
      Result.Refresh;
    except
      Result.free;
      raise;
    end;                                                      
  end else
  begin
    FillChar(spec,sizeof(spec),#0);
    spec.attr := faDirectory;
    Result := TSourceDir.Create(inName,filters,spec,inparent);
    {Err_nodirectory(inName);}
  end;
end;

procedure  TSourceDir.Refresh;
var
  path : string;
  fileSpec : TSearchRec;
  slist : TStringList;
  i : integer;
  info : TFileInfo;
  subdir : TSourceDir;
  attrtest : WORD;
  iter : TSrcTreeIterator;
  fitsFilter : boolean;

  procedure sub_search(const spath:string; sattr:WORD);
  var
    sdpath:string;
    mypath:string;
    gotafile : integer;
    checkmask : integer;
  const
    faAllFiles = faArchive and (not faDirectory);
  begin
    FillChar(fileSpec,sizeof(filespec),0);
    mypath := AbsPath;
    gotafile := FindFirst(spath,sattr,fileSpec);
    checkmask := ( not ($0040) ) and faEveryThing;
    {since $40 is not screened for it is masked out}
    if gotafile=0  then
    Repeat
      if ((fileSpec.attr and faAvoidLikePlague) = faAvoidLikePlague) then
        continue;
      try
        attrtest := (fileSpec.attr and checkmask);
        if (attrtest = $0000) then
          attrtest := faNormalFile; {pretend its faNormalFile}
        if (attrtest and sattr) = 0 then
          continue;
        sdpath := mypath + kDirSeparator + filespec.name;
        if FindFileName(fileSpec.name) >= 0 then
          continue; {?}
        {if info.spec.time = 0 then
          ShowMessage('zero time for '+sdpath);}
        SetFileInfo(info,filespec);
        AddFileInfo(info);
        if ((faDirectory and attrtest) <> 0) then
        begin {is a directory}
          if (FDoSubDirs) and FI_IsChildDirectory(info) then
          begin
            subDir := GetSubDir(sdpath);
            if subDir = nil then
            begin
              subDir := NewSubDir(sdpath,FFilter, info.spec);
              subDir.DoSubDirs := TRUE;
            end;
          end;
        end;
      except
        on  ex1 : Exception do
        begin
          ShowMessage(
            Format('Error %s : %s',[sdpath,ex1.message]));
          raise;
        end;
      end;
    Until FindNext(fileSpec) <> 0;
    if (gotafile=0)  then
      SysUtils.FindClose(fileSpec);
  end;
begin
  if (FDoSubDirs) then
    FreeSubDirs;
  Count := 0;
  iter := nil;
  slist := TStringList.Create;
  {$IFDEF USE_PSTRING}
    InitCCStr(info.filename);
  {$ENDIF}
  try
    GetFilters(FFilter,slist);
    path := AbsPath + kDirSeparator+'*.*';
    sub_search(path,faEverything);
    iter := TSrcTreeIterator.Create(SELF);
    if iter.FirstFile(info) then
    repeat
      fitsFilter := FI_IsDirectory(info);
      if not fitsFilter then
      for i := 0 to slist.Count-1 do
      begin
        if NameFitsFilter(GetCCStr(info.filename), slist.Strings[i]) then
        begin
          fitsFilter := TRUE;
          break;
        end;
      end;
      if not fitsFilter then
        iter.DeleteCurrent;
    until not iter.NextFile(info);
    if FDoSubDirs then
      RefreshSubDirs;
    FDoSubDirs := FALSE;
  finally
  {$IFDEF USE_PSTRING}
    KillCCStr(info.filename);
  {$ENDIF}
    iter.free;
    slist.free;
  end;
end;
{---------------------------------------------------------------}

constructor TJvRnnFileCprMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TCopyRec),TRUE);
end;

constructor TCompareArray.Create(const rootA,rootB : string);
begin
  inherited CreateMgr(TJvRnnFileCprMgr);
  BlockHdl := @FArrayPtr;
  FFilter := '*.*';
  FConditions := -1;
  Sorted := TRUE;
  FRecurse := TRUE;
  Count := 0;
  Dir_A := rootA;
  Dir_B := rootB;
end;


function  TJvRnnFileCprMgr.CompareItems(const p1, p2):integer;
begin
  Result := CompareFilePaths(TCopyRec(p1), TCopyRec(p2),FSortType);
end;

function CompareFilePaths(
  const prec, qrec : TCopyRec; sortType : TJvRnnSortFCList) : integer;

  function sub_sign(diff : TDateTime) : integer;
  begin
    diff := INT(diff * 24 * 60 * 60 * 1000);
    if (Abs(diff) > MAXINT) then
      Result := MAXINT
    else
      Result := TRUNC(Abs(diff));
    if diff < 0 then
      Result := -Result;
  end;
begin
  {return true if prec > qrec, ie need to be swapped}
  CASE sortType of
    sort_name :
      Result := COMPAREText(ExtractFileName(GetCCStr(prec.subpath)),
        ExtractFileName(GetCCStr(qrec.subpath)));
    sort_path :
      Result := COMPAREText(GetCCStr(prec.subpath), GetCCStr(qrec.subpath));
    sort_type :
      Result := COMPAREText(ExtractFileExt(GetCCStr(prec.subpath)),
        ExtractFileExt(GetCCStr(qrec.subpath)));
    sort_fromtime :
      Result := sub_sign(prec.v1.Time - qrec.v1.Time);
    sort_fromsize :
      Result := (prec.v1.Size - qrec.v1.Size);
    sort_fromlocked :
      Result := (prec.v1.Attr and faReadOnly) - (qrec.v1.Attr and faReadOnly);
    sort_totime :
      Result := sub_sign(prec.v2.Time - qrec.v2.Time);
    sort_tosize :
      Result := (prec.v2.Size - qrec.v2.Size);
    sort_tolocked :
      Result := (prec.v2.Attr and faReadOnly) - (qrec.v2.Attr and faReadOnly);
  ELSE
    Result := 0;
  END; {CASE}
end;

function TCompareArray.GetSortType : TJvRnnSortFCList;
begin
  Result := TJvRnnFileCprMgr(FRecMgr).FSortType;
end;

procedure TCompareArray.SetSortType(stype : TJvRnnSortFCList);
begin
  if  SortType <> stype then
  begin
    SortType := stype;
    DoSort;
  end;
end;

procedure TCompareArray.DoSort;
begin
  QuickSort;
end;


procedure TCompareArray.AddCopyRec(const copyRec : TCopyRec);
var
  ix : integer;
  pcr : PCopyRec;
begin
  if not GetItemIndex(ix,copyRec) then
  begin
    CreateItemsAt(ix,1);
    pcr := @ArrayPtr^[ix];
    SetCCStr(pcr^.subpath,GetCCStr(copyRec.subpath));
    pcr^.v1 := copyRec.v1;
    pcr^.v2 := copyRec.v2;
    pcr^.crFlags := copyRec.crFlags;
  end;
end;

function GetDirSubPath(const base, leaf : string) : string;
var
  i : integer;
begin
  i := Length(base)+1;
  if (leaf[i] = kDirSeparator) then
    Inc(i);
  Result := Copy(leaf,i,Length(leaf)-i+1);
end;


procedure CompareWalkDirectories(const dirA, dirB, filter : string;
                                  conditions : integer;
                                  paction : TJvRnnDirTreeWalk;
                                  recurse : boolean
                                  );
var
  rootA : TSourceDir;
  rootB : TSourceDir;

  procedure CompareWalkFiles(
    listA, listB : TSourceDir;
    filter : string;
    conditions : integer;
    paction : TJvRnnDirTreeWalk;
    recurse : boolean;
    swapped : boolean
    );  forward;

  procedure CompareWalkSubDir(parentA, parentB : TSourceDir;
                                    const dirA, dirB : string;
                                    filter : string;
                                    conditions : integer;
                                    paction : TJvRnnDirTreeWalk;
                                    recurse,swapped : boolean
                                    );
  var
    listA : TSourceDir;
    listB : TSourceDir;
  begin
    if not assigned(paction) then
      exit;
    listA := nil;
    listB := nil;
    try
      listA := CreateDirList(parentA, dirA, filter, FALSE);

      listB := CreateDirList(parentB, dirB, filter, FALSE);

      if (conditions and (not kB_NOT_A)) <> 0 then
        CompareWalkFiles(listA,listB, filter, conditions, paction,recurse,swapped);
      {kB_NOT_A is only handled by CompareWalkDirectories by a swap}
    finally
      listA.free;
      listB.free;
    end;
end;

  procedure CompareWalkFiles(
    listA, listB : TSourceDir;
    filter : string;
    conditions : integer;
    paction : TJvRnnDirTreeWalk;
    recurse : boolean;
    swapped : boolean
    );
  var
    infoA : TFileInfo;
    infoB : TFileInfo;
    itorA : TSrcTreeIterator;
    A_NOT_B, A_NEWER_B, B_NEWER_A, AB_SAMETIME, AB_DIFFER : boolean;
    crec : TCopyRec;
    subdir : string;
    cpdir : string;
    matchFound : boolean;
    dosubdir : boolean;
    dirct : integer;
    p1 : PFileSpec;
    p2 : PFileSpec;
    origRoot : TSourceDir;
  begin
    itorA := TSrcTreeIterator.Create(listA);
    FillChar(crec,sizeof(crec),0);
    {$IFDEF USE_PSTRING}
    InitCCStr(infoA.filename);
    InitCCStr(infoB.filename);
    InitCCStr(crec.subpath);
    {$ENDIF}
    try
      A_NOT_B := ((conditions AND kA_NOT_B)<>0);
      {B_NOT_A := ((conditions AND kB_NOT_A)<>0);}
      A_NEWER_B := ((conditions AND kA_NEWER_B)<>0);
      B_NEWER_A := ((conditions AND kB_NEWER_A)<>0);
      AB_SAMETIME := ((conditions AND kAB_SAMETIME)<>0);
      AB_DIFFER := ((conditions AND kAB_DIFFERENT)<>0);
      if swapped then
      begin
        p2 := @crec.v1;
        p1 := @crec.v2;
      end else
      begin
        p1 := @crec.v1;
        p2 := @crec.v2;
      end;
      dirct := 0;
      dosubdir := recurse and A_NOT_B;
      if (itorA.FirstFile(infoA)) then
      repeat
        if not FI_IsDirectory(infoA) then
        begin
          subdir := itorA.FCurDir.SubPath;
          if Length(subdir) > 0 then
            SetCCStr(crec.subpath,subdir + kDirSeparator + GetCCStr(infoA.filename))
          else
            SetCCStr(crec.subpath,GetCCStr(infoA.filename));
          crec.crFlags := 0;
          p1^ := infoA.spec;
          p1^.fsflags := kFS_Exists;
          matchFound := listB.FindInTree(GetCCStr(crec.subpath),infoB);
          if matchFound then
          begin
            p2^ := infoB.spec;
            p2^.fsflags := kFS_Exists;
          end else
          with p2^ do begin
            time := 0;
            size := 0;
            attr := 0;
            p2^.fsFlags := 0;
          end;
          {set to full subpath}
          if swapped then
            origRoot := rootB
          else
            origRoot := rootA;
          cpdir := GetDirSubPath(
            origRoot.AbsPath,
            itorA.FCurDir.AbsPath + kDirSeparator + GetCCStr(infoA.filename));
          SetCCStr(crec.subpath,cpdir);
          if A_NOT_B and (not matchfound) then
            TJvRnnDirTreeWalk(paction)(crec)
          else if (matchfound) then
          begin
            if A_NEWER_B and (p1^.time > p2^.time) then
              TJvRnnDirTreeWalk(paction)(crec)
            else if B_NEWER_A and (p1^.time < p2^.time) then
              TJvRnnDirTreeWalk(paction)(crec)
            else begin
              if (p1^.size <> p2^.size) and (AB_DIFFER) then
                TJvRnnDirTreeWalk(paction)(crec)
              else
              if AB_SAMETIME and (p1^.time = p2^.time) then
                TJvRnnDirTreeWalk(paction)(crec);
            end;
          end
        end
        else if dosubdir then
        begin
          if FI_IsChildDirectory(infoA) then
            Inc(dirct);
        end;
      until( not itorA.NextFile(infoA));
      if (dirct > 0) and (itorA.FirstFile(infoA)) then
      repeat
        if FI_IsChildDirectory(infoA) then
        begin
          subdir := listA.AbsPath + kDirSeparator + GetCCStr(infoA.filename);
          cpdir := listB.AbsPath + kDirSeparator + GetCCStr(infoA.filename);
          CompareWalkSubDir(listA, listB, subdir,cpdir,filter,conditions,paction,TRUE,swapped);
        end;
      until( not itorA.NextFile(infoA));
    finally
      {$IFDEF USE_PSTRING}
        KillCCStr(infoA.filename);
        KillCCStr(infoB.filename);
        KillCCStr(crec.subpath);
      {$ENDIF}
      itorA.Free;
    end;
  end;

begin
  if not assigned(paction) then
    exit;
  rootA := nil;
  rootB := nil;
  try
    rootA := CreateDirList(nil, dirA, filter, FALSE);
    rootB := CreateDirList(nil, dirB, filter, FALSE);
    if (conditions and (not kB_NOT_A)) <> 0 then
      CompareWalkFiles(rootA,rootB, filter, conditions, paction,recurse,FALSE);
    if (conditions AND kB_NOT_A)<>0 then
      CompareWalkFiles(rootB,rootA, filter, kA_NOT_B, paction,recurse,TRUE);
  finally
    rootA.free;
    rootB.free;
  end;
end;

procedure TCompareArray.DoCompare;
{It may be a good idea to empty this list before calling}
var
  oldSort  :  TJvRnnSortFCList;
begin
  oldSort := SELF.SortType;
  SELF.SortType := sort_path;
  try
    CompareWalkDirectories(FDirA,FDirB,Filter,Conditions,
      AddCopyRec,Recurse);
  finally
    SELF.SortType := oldSort;
  end;
end;

procedure TCompareArray.SetDir_A(const inA : string);
begin
  FDirA := inA;
  RemoveEndChar(FDirA,kDirSeparator);
end;

procedure TCompareArray.SetDir_B(const inB : string);
begin
  FDirB := inB;
  RemoveEndChar(FDirB,kDirSeparator);
end;


function TCompareArray.FindSubPathIndex(const insubpath:string) : integer;
var
  i : integer;
  copyRec : TCopyRec;
  pcopy : PCopyRec;
begin
  Result := -1;
  if SortType = sort_name then
  begin
   (* already sorted by name *)
    FRecMgr.ConstructItem(copyRec);
    try
      SetCCStr(copyRec.subpath,insubpath);
      if GetItemIndex(i, copyRec) then
        Result := i;
    finally
      FRecMgr.DestructItem(copyRec);
    end;
  end else
  begin
    pcopy := PCopyRec(BlockPtr);
    for i := 0 to Count-1 do
    begin
      if COMPARETEXT(GetCCStr(pcopy^.subpath) ,insubpath) = 0 then
      begin
        result := i;
        exit;
      end;
      Inc(pcopy,1);
    end;
  end;
end;

function  TCompareArray.GetSpec(const insubpath:string;
  recpath:TCopyRecPath;
  var spec : TFileSpec) : boolean;
var
  i : integer;
  pcopy : PCopyRec;
begin
  i := FindSubPathIndex(insubpath);
  if ( i < 0 ) then
    Result := FALSE
  else begin
    Result := TRUE;
    pcopy := @ArrayPtr^[i];
    if recpath=path_A then
      spec := pcopy^.v1
    else
     spec := pcopy^.v2;
  end;
end;

procedure TJvRnnFileCprMgr.DestructItem(var item);
begin
  KillCCStr(TCopyRec(item).subpath);
end;

procedure TJvRnnFileCprMgr.ConstructItem(var item);
begin
  FillChar(item,sizeof(TCopyRec),0);
  InitCCStr(TCopyRec(item).subpath);
end;


constructor TSrcTreeIterator.create(src : TSourceDir);
begin
  FSrcDir := src;
  FIXArray := TJvCCIntArray.create(1);
end;

destructor TSrcTreeIterator.destroy;
begin
  FIXArray.free;
  inherited destroy;
end;

function TSrcTreeIterator.GetAbsPath : string;
begin
  if FCurDir = nil then
    raise EFileSpec.Create('No current directory');
  Result := FCurDir.GetAbsPath;
end;

function TSrcTreeIterator.GetSubPath : string;
begin
  if FCurDir = nil then
    raise EFileSpec.Create('No current directory');
  Result := FCurDir.SubPath;
end;

function TSrcTreeIterator.HasStarted : boolean;
begin
  Result := (FCurDir <> nil);
end;

function  TSrcTreeIterator.GetCurFilename : string;
var
  pfi:PFileInfo;
begin
  if FCurDir = nil then
    raise EFileSpec.Create('No current directory');
  if (FFileix > 0) and (FFileIX <= FCurDir.Count) then
  begin
    pfi := PFileInfo(FCurDir.Ptr[FFileIX-1]);
    Result := FCurDir.AbsPath + kDirSeparator + GetCCStr(pfi^.filename);
  end;
end;

procedure TSrcTreeIterator.DeleteCurrent;
begin
  if FCurDir = nil then
    exit;
  if (FFileix > 0) and (FFileIX <= FCurDir.Count) then
  begin
    Dec(FFileIX);
    FCurDir.DeleteItemsAt(FFileIX,1);
  end;
end;


function TSrcTreeIterator.FirstFile(var info:TFileInfo):boolean;
begin
  FCurDir := FSrcDir;
  FFileix := 0;
  FIXArray.count := 1;
  FIXIX := 0;
  FixArray.PArray^[FIXIX] := 0;
  Result := NextFile(info);
end;

function TSrcTreeIterator.NextFile(var info:TFileInfo):boolean;
var
  subix : integer;
begin
  if (FCurDir = nil) then
  begin
    Result := FALSE;
    exit;
  end;
  if FFileix < FCurDir.count then
  begin
    FCurDir.GetItem(FFileIX,info);
    Inc(FFileix);
    Result := TRUE;
  end else
  begin
    subix := FIXArray.PArray^[FIXIX];
    if (FCurDir.FSubDirList <> nil) and (subix < FCurDir.FSubDirList.Count) then
    begin
      FCurDir := TSourceDir(FCurDir.FSubDirList[subix]);
      Inc(subix);
      FIXArray.PArray^[FIXIX] := subix;
      Inc(FIXIX);
      if (FIXIX >= FIXArray.count) then
        FIXArray.count := FIXIX + 1;
      FIXArray.PArray^[FIXIX] := 0;
      FFileix := 0;
    end else
    begin {exhausted this node}
      FCurDir := FCurDir.FParent;
      Dec(FIXIX);
      if (FCurDir = nil) OR (FIXIX < 0) then
      begin
        Result := FALSE;
        exit; {end iteration}
      end;
      FFileix := FCurDir.FSubDirList.count;
    end;
    Result := NextFile(info);
  end;
end;

end.




