{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnDBScript.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnDBScript;

interface
Uses Classes, SysUtils, JvRnnScCst,
     JvRnnGenDB, JvCCRecrd,JvCCArray, JvRnnPars, DB;

const
  TJvGenDBScript_VERSION = '1.00';
type

  TDBLinkClassRec = record
    FDBLinkName : string;
    FDBLinkClass : TJvRnnGenDBLinkClass;
  end;
  PDBLinkClassRec = ^TDBLinkClassRec;


  TDBLinkClassMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer; override;
    function CompareKey(const buf; key:Pointer):integer; override;
    procedure ConstructItem(var buf); override;
    procedure DestructItem(var buf); override;
    procedure SwapItems(var b1,b2); override;
  end;

  TJvGenDBScript = class(TJvRnnCustomScript)
  protected
    FDBLinks : TJvCCObjArray;
    FDSList : TDatasetList;
    FGenDB : TJvRnnGenDBLink;

    class function GetDBLinkClass(const name:string) : TJvRnnGenDBLinkClass;
    class function GetFieldDefList(dataset:TDataset; slist:TStrings):TJvCCObjArray;
    procedure ForgetDBLink(dblink:TJvRnnGenDBLink);

    procedure RemoveDSRef(item:TJvCCItemPtr; data:Pointer);
    procedure Notification(
      AComponent:TComponent; operation:TOperation); override;
    procedure ExecuteStd(cmd : TStrings); override;

    procedure CheckGenDB;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    class procedure RegisterDBLink(
      const name:string; linkclass:TJvRnnGenDBLinkClass);

    function NewDBLink(
      const linkname, linktype:string;
      params : TStrings): TJvRnnGenDBLink;
    function NewDataset(
      const dsname:string;
      ds_link:TJvRnnGenDBLink) : TJvRnnDatasetLink;
    function FindDatasetLink(const name : string):TJvRnnDatasetLink;
    procedure FreeDatasetLink(ds_link:TJvRnnDatasetLink);
    function EvaluateName(var ps : TParseState) : string; override;
    property DSList : TDatasetList read FDSList;
    property GenDB : TJvRnnGenDBLink read FGenDB write FGenDB;
    function GetAboutMe: string;
  published
    property AboutMe: string read GetAboutMe stored False;
    property CmdExtract;
    property CmdProcess;
    property CmdExecute;
    property OnExecuteStd;

    property OnLogMsg;
    property BeforeCmdExecute;
    property OnLogError;
    property OnLogStrings;
    property DBLinks : TJvCCObjArray read FDBLinks;
    property ParamChar;
    property PrefixChar;
    property DeleteComments;
    property Terminator;
    property StdTerminator;
    property ScriptStore;
    property ValueList;

  end;

const
  kDB_CREATE = 'DB_CREATE';
  kDB_CONNECT = 'DB_CONNECT';
  kDB_DISCONNECT = 'DB_DISCONNECT';
  kDB_FREE = 'DB_FREE';
  kDB_EXEC = 'DB_EXEC';
  kDB_SETDB = 'DB_SETDB';
  kDB_DSCREATE = 'DB_DSCREATE';
  kDB_DSFREE = 'DB_DSFREE';
  kDB_CREATE_SOURCE = 'DB_CREATE_SOURCE';
  kDB_BATCH_APPENDUPDATE = 'DB_BATCH_APPENDUPDATE';
  kDB_SET = 'DB_SET';
  kDB_PARAMS = 'DB_SETPARAMS';

  kDB_TRANS_BEGIN = 'DB_TRANS_BEGIN';
  kDB_TRANS_COMMIT = 'DB_TRANS_COMMIT';
  kDB_TRANS_COMMIT_RETAINING = 'DB_TRANS_COMMIT_RETAINING';
  kDB_TRANS_ROLLBACK = 'DB_TRANS_ROLLBACK';

  kFOR_EACH_ROW = 'FOR_EACH_ROW';
  kNEXT_EACH_ROW = 'NEXT_EACH_ROW';
  //kDB_PRINT = 'DB_PRINT';

  // functions
  kDB_EXISTS_TABLE = 'DB_EXISTS_TABLE';

  // key words
  KW_SOURCE = 'SOURCE';
  KW_SOURCE_FIELD  = 'SOURCE_FIELD';
  KW_ADD_FIELD = 'ADD_FIELD';
  KW_PRIMARY_KEY = 'PRIMARY_KEY';
  KW_TABLENAME = 'TABLENAME';
  KW_FIELDS = 'FIELDS';
  KW_KEYS = 'KEYS';
  KW_ALL_DATA = 'ALL_DATA';
  KW_ACTION = 'ACTION';
  // Error templates

  kNO_DATASET_NAME = 'No dataset name';
  kNO_DBLINK_NAME = 'No dblink name';
  kNO_DBLINK_CLASS_NAME = 'No dblink class name';
  kNO_TABLENAME = 'No table name';
  
  // For format
  kCLASS_NOT_KNOWN = 'class %s not known';
  kDBLINK_NOT_FOUND = 'DBLink %s not found';
  kDATASET_EXISTS = 'Dataset %s exists';
  kDATASET_NOT_FOUND = 'Dataset %s not found';
  kKEYWORD_EXPECTED = 'Keyword "%s" expected';
  kCHARACTER_EXPECTED = 'Character %s expected';

var
  gDBLinkClasses : TJvCCArray;


// installed functions
function DBExistsTable(sct: TJvRnnCustomScript; args : TStrings) : string;

implementation
uses JvRnnStrUt;


type
  TForEachRow = class(TJvRnnScrObj)
  protected
    FFieldNames : TStrings;
    FSavedVar : TStrings;
    FLoopBackLine : integer;
    FDSLink : TJvRnnDatasetLink;
    FFirstPass : boolean;
    FScriptor : TJvGenDBScript;

    function InitLoop(cmdInfo : PJvScriptCmdInfo) : boolean;
    function NextRow(var ss : TJvRnnScriptState) : boolean;
    procedure AfterPopState(var rst:TJvRnnScriptState); override;
    procedure RestoreVariables;
    destructor Destroy; override;
  end;

{

 @DB_CREATE <AliasName> <RegisteredLinkClass> ;
 @DB_CONNECT <AliasName> Params(paramlist);
 @DB_DISCONNECT <AliasName>;
 @DB_FREE <AliasName>;
 
 @DB_DSCREATE <AliasName> <SrcDataName> <sql>;
 @DB_DSFREE <SrcDataName>;

 @DB_EXEC <AliasName> <sql>;
 @DB_SETDB <AliasName>;

 @DB_APPEND <SrcData> <DestData>
      InsertNap("dest1=src1",..);

 @DB_SET <AliasName> SET <varname> = <fieldname> USING
   select ...  // single or first row fieldnames

 @DB_UPDATE <SrcData> <DestData>
      KeyMap(dest1=src1,..) UpdateMap("dest1=src1",..);
 @DB_DELETE <SrcData> <DestData>
      KeyMap("dest1=src1",..);

 @DB_FIRST <SrcData>;
 @DB_NEXT  <SrcData>;
 @DB_LAST <SrcData>;
 @DB_PRIOR  <SrcData>;

 @WHILE <Condition>;
 @ENDWHILE;

 @IF <Condition>
 @ELSE
 @ENDIF;

 Functions
 DBF_COUNT(<SrcData>) = ?variant ?integer ?string
 DBF_BOF(<SrcData>) = 'Y'|'N'   ?boolean
 DBF_EOF(<SrcData>) = 'Y'|'N'   ?boolean
 DBF_VALUE(<SrcData>, <SrcField>) ?Variant  ?string

}



procedure CreateDBLink(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure ConnectDBLink(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure DisconnectDBLink(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure FreeDBLink(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure OpenDBData(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure CloseDBData(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure ExecDBSQL(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure SetDBAlias(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure BatchCreate_F(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure BatchUpdate_F(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure SetFromDBQuery_F(cmdInfo : PJvScriptCmdInfo); far; forward;
procedure ForEachRow_F(cmdInfo:PJvScriptCmdInfo);  far; forward;
procedure NextEachRow_F(cmdInfo:PJvScriptCmdInfo); far; forward;
procedure SetDBParams(cmd:PJvScriptCmdInfo); far; forward;




procedure TForEachRow.AfterPopState(var rst:TJvRnnScriptState);
begin
  RestoreVariables;
end;


destructor TForEachRow.Destroy;
begin
  FFieldNames.Free;
  FSavedVar.Free;

  inherited Destroy;
end;

procedure TForEachRow.RestoreVariables;
var
  varray : TJvCCStrStrArray;
  i : integer;
  vix : integer;
begin
  if (not FFirstPass) and (FFieldNames <> nil) then
  begin
    varray := FScriptor.Variables;

    for i := 0 to FFieldNames.count-1 do
    begin
    {remove temporary variables with same name}
      vix := varray.IndexOf(FFieldNames[i]);
      if vix >= 0 then
        varray.DeleteItemsAt(vix,1);
    end;
  {restore variables which had same name}
    if FSavedVar <> nil then
      varray.AppendValueStrings(FSavedVar);
  end;
end;

function TForEachRow.NextRow(var ss : TJvRnnScriptState) : boolean;
var
  varray : TJvCCStrStrArray;
  i : integer;
  vix : integer;
begin
  with FDSLink do
  begin
    if FFirstPass then
      Dataset.First
    else
      Dataset.Next;
    if Dataset.EOF then
    begin
      Result := FALSE;
      exit;
    end;
  end;
  if (FFieldNames<> nil) then
  begin
    varray := FScriptor.Variables;
    for i := 0 to FFieldNames.count-1 do
    begin
    {SAVE any global variables with same name}
      if FFirstPass then
      begin
        vix := varray.IndexOf(FFieldNames[i]);
        if vix >= 0 then
        begin
          if (FSavedVar=nil) then
            FSavedVar := TStringList.create;
          FSavedVar.Values[FFieldNames[i]] := varray.GetIndexValue(vix);
        end;
      end;
      varray.SetVariable(FFieldNames[i],TField(FFieldNames.Objects[i]).DisplayText);
    end;
  end;
  ss.lineNum := FLoopBackLine;
  FFirstPass := FALSE;
  Result := TRUE;

end;

function TForEachRow.InitLoop(cmdInfo : PJvScriptCmdInfo):boolean;
const
  kForSyntax = '@FOR_ROWS [| * | name1,name2..| ] in ''Table''';
var
  p, r : PChar;
  token : string;
  xtoken : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  allfields : boolean;
  i : integer;
  anyField : TField;
  dtfield : TDateTimeField;
  ss : TJvRnnScriptState;
  dct : integer;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise EJvRnnScript.create(kForSyntax);
  end;
begin
  Result := FALSE;
  FScriptor := cmdInfo^.scriptor as TJvGenDBScript;
  p := cmdInfo^.cmdcstr;
  sub_next;

  if not (COMPARETEXT('in',token)=0) then
  begin
    FFieldNames := TStringList.Create;
    allfields := (COMPARETEXT('*',token)=0);
    if not allfields then
    begin
      while (COMPARETEXT(token,'in')<>0) do
      begin
        if token <> ',' then
          FFieldNames.Add(token);
        sub_next;
      end;
    end else
    begin
      sub_next;
    end;
    if not (COMPARETEXT(token,'in')=0) then
      raise EJvRnnScript.create(kForSyntax);
  end else
  begin
    allfields := FALSE;
  end;
  {get dataset link name}
  sub_next;


  FDSLink := FScriptor.FindDatasetLink(token);
  if (FDSLink=nil) then
    EJvRnnScript.create(Format(kDATASET_NOT_FOUND,[token]));
  if allfields then
  with FDSLink do
  begin
    for i := 0 to Dataset.Fieldcount-1 do
      FFieldNames.Add(Dataset.Fields[i].FieldName);
  end;

  for i := 0 to FFieldNames.count-1 do
  with FDSLink do
  begin
    anyField := Dataset.FieldByName(FFieldNames[i]);
    if (Length(FScriptor.DateFormat)>0)
      and (anyField is TDateTimeField) then
    begin
      dtField := TDateTimeField(anyField);
      CASE (dtField.DataType) of
      ftDate : dtField.DisplayFormat := FScriptor.DateFormat;
      ftTime :  dtField.DisplayFormat := FScriptor.TimeFormat;
      ftDateTime : dtField.DisplayFormat := FScriptor.DateTimeFormat;
      END;
    end;
    FFieldNames.Objects[i] := anyField;
  end;


  ss := FScriptor.GetScriptState;
  FLoopBackLine := ss.linenum;
  dct := 1;

  {interpret succeeding lines until find ENDFOR;
  problem is succeeding lines may contain
  text that contains 'ENDFOR'}
  
  with ss do
  Repeat
    if (lineNum >= script.count) then
      break;
    xtoken := script[lineNum];
    r := StrToPChar(xtoken);
    r := StrStartToken(r);

    if (r^ = '@') then
    begin
      Inc(r);
      if StrNextTokenType(r,token, ptype, tokentype) then
      begin
        if COMPARETEXT(token,kFOR_EACH_ROW)=0 then
          Inc(dct);
        if COMPARETEXT(token,kNEXT_EACH_ROW)=0 then
          Dec(dct);
      end;
    end;
    Inc(lineNum);
  Until(dct = 0);

  if (dct <> 0) then
    raise EJvRnnScript.Create(kNEXT_EACH_ROW + ' not found');

  FScriptor.SetScriptState(ss);{ready for next statement after end_for}
  FFirstPass := TRUE;
  if NextRow(ss) then
    begin
    ss.scriptlabel := kFOR_EACH_ROW;
    ss.refptr := SELF;
    FScriptor.PushScriptState(ss);
    Result := TRUE;
    end
end;
{-------------------------------------------------------------}


function TJvGenDBScript.GetAboutMe: string;
begin
  Result := 'Version: '+TJvGenDBScript_VERSION;
end;


class procedure TJvGenDBScript.RegisterDBLink(
      const name:string; linkclass:TJvRnnGenDBLinkClass);
var
  rec : TDBLinkClassRec;
begin
  rec.FDBLinkName := name;
  rec.FDBLinkClass := linkclass;
  gDBLinkClasses.AppendItem(rec);
end;

class function TJvGenDBScript.GetFieldDefList(
  dataset:TDataset;
  slist:TStrings):TJvCCObjArray;
var
  i : integer;
  s : string;
  fld : TFieldDef;
  defs : TFieldDefs;
begin
  Result := TJvCCObjArray.CreateMgr(TJvCCObjRecMgr);
  defs := dataset.FieldDefs;
  for i := 0 to slist.count-1 do
  begin
    s := slist[i];
    fld := defs.Find(s);
    Result.AppendItem(fld);
  end;
end;


class function TJvGenDBScript.GetDBLinkClass(
  const name:string) : TJvRnnGenDBLinkClass;
var
  ip : TJvCCItemPtr;
  prec : PDBLinkClassRec;
begin
  ip := nil;
  if gDBLinkClasses.LocateKeyPtr(@name, ip, its_forward) then
  begin
    prec := PDBLinkClassRec(ip);
    Result := prec^.FDBLinkClass;
  end else
    Result := nil;
end;

procedure TJvGenDBScript.RemoveDSRef(item:TJvCCItemPtr; data:Pointer);
var
  ds : TJvRnnDatasetLink;
begin
  ds := TJvRnnDatasetLink(item^);
  FDSList.DeleteItem(ds);
end;

constructor TDBLinkClassMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TDBLinkClassRec), TRUE);
end;

function TDBLinkClassMgr.CompareItems(const p1, p2):integer;
begin
  Result := CompareStr(TDBLinkClassRec(p1).FDBLinkName,
                TDBLinkClassRec(p2).FDBLinkName);
end;

function TDBLinkClassMgr.CompareKey(const buf; key:Pointer):integer;
begin
  Result := CompareStr(TDBLinkClassRec(buf).FDBLinkName,
                PString(key)^);
end;

procedure TDBLinkClassMgr.ConstructItem(var buf);
begin
  Initialize(TDBLinkClassRec(buf));
end;

procedure TDBLinkClassMgr.DestructItem(var buf);
begin
  Finalize(TDBLinkClassRec(buf));
end;

procedure TDBLinkClassMgr.SwapItems(var b1,b2);
var
  Stuff : TDBLinkClassRec;
begin
  Stuff := TDBLinkClassRec(b1);
  TDBLinkClassRec(b1) := TDBLinkClassRec(b2);
  TDBLinkClassRec(b2) := Stuff;
end;


  {TJvCCObjCompare = function(obj1, obj2 : TObject) : integer;
  TJvCCKeyObjCompare = function(obj : TObject;key : Pointer) : integer;
  }

function DBLinkObjCompare(obj1,obj2: TObject) : integer;
begin
  Result := CompareText(
    TJvRnnGenDBLink(obj1).LinkName,
    TJvRnnGenDBLink(obj2).LinkName);
end;

function DBLinkKeyCompare(obj : TObject;key : Pointer) : integer;
begin
  Result := CompareText(
    TJvRnnGenDBLink(obj).LinkName,
    PString(key)^);
end;

procedure Throw(cmdInfo : PJvScriptCmdInfo; const s : string);
var
  xmsg : string;
begin
  xmsg := Format('Error: %s in %s',[s,cmdInfo^.cmdName]);
  raise EJvRnnScript.create(xmsg);
end;


function TJvGenDBScript.NewDBLink(
      const linkname, linktype:string;
      params : TStrings): TJvRnnGenDBLink;
var
  dbclass : TJvRnnGenDBLinkClass;
begin
  dbclass := GetDBLinkClass(linktype);
  if (dbclass = nil) then
    raise EJvRnnScript.Create(Format(kCLASS_NOT_KNOWN,[linktype]));
  Result := dbclass.Create(nil);
  Result.LinkName := linkname;
  Result.LinkType := linktype;
  Result.SetScriptor(SELF);
  DBLinks.AppendItem(Result);
  if (params<>nil) then
    Result.ConnectLink.Params := params;
end;


procedure CreateDBLink(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dbclassname : string;
  script : TJvGenDBScript;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_CLASS_NAME);
  dbclassname := ps.token;
  script := cmdInfo^.scriptor as TJvGenDBScript;
  script.NewDBLink(link_name, dbclassname, nil);
end;

{ @DB_CONNECT <AliasName> Params(paramlist);}

function Sub_GetDBLink(
  cmdInfo:PJvScriptCmdInfo; const link_name:string) : TJvRnnGenDBLink;
var
  script : TJvGenDBScript;
begin
  script := cmdInfo^.scriptor as TJvGenDBScript;
  Result := nil;
  if not (script.DBLinks.LocateKey(@link_name, Result)) then
    Throw(cmdInfo, Format(kDBLINK_NOT_FOUND,[link_name]));
end;

procedure ConnectDBLink(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  script : TJvGenDBScript;
  slist : TStrings;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  script := TJvGenDBScript(cmdInfo^.scriptor);
  if ps_nexttokentype(ps) then
  begin
    if not (CompareStr(ps.token,'Params')=0) then
      Throw(cmdInfo, 'Params expected');
    slist := script.ParseFunctionArgs(ps.p);
    if (slist <> nil) then
    try
      dblink.ConnectLink.Params := slist;
      dblink.Connected := TRUE;
    finally
      slist.free;
    end;
  end;
end;

procedure DisconnectDBLink(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.Connected := FALSE;
end;

procedure FreeDBLink(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  script : TJvGenDBScript;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.Connected := FALSE;
  script := TJvGenDBScript(cmdInfo^.scriptor);
  script.ForgetDBLink(dblink);
  dblink.free;
end;

procedure OpenDBData(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  ds : TJvRnnDatasetLink;
  ds_name : string;
  script : TJvGenDBScript;
  slist : TStrings;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dataset name');
  ds_name := ps.token;
  script := TJvGenDBScript(cmdInfo^.scriptor);
  ds := script.FindDatasetLink(ds_name);
  if (ds <> nil) then
    Throw(cmdInfo, Format('Dataset %s exists',[ds_name]));
  if not ps_skipwhite(ps) then
    Throw(cmdInfo, 'No SQL');
  slist := TStringList.Create;
  try
    slist.Text := StrPas(ps.p);
    ds := dblink.ReturnDataset(slist);
    ds.DSName := ds_name;
    dblink.DSList.AppendItem(ds);
    script := TJvGenDBScript(cmdInfo^.scriptor);
    script.DSList.AppendItem(ds);
  finally
    slist.free;
  end;
end;



procedure ForEachRow_F(cmdInfo:PJvScriptCmdInfo);
{@FOR [* | Fieldname1, Fieldname2, Fieldname#] in ':ALIAS:View.db'
	/* execute statements with variables set to field values for each row */
	delete from stqm where Fieldname1=$Fieldname1 and Fieldname2=$Fieldname2;
DONE}
var
  rowstep : TForEachRow;
begin
  rowstep := TForEachRow.Create;
  if not rowstep.InitLoop(cmdInfo) then
    rowstep.Free;
end;

procedure NextEachRow_F(cmdInfo:PJvScriptCmdInfo);
var
  scriptor : TJvGenDBScript;
  sspop : TJvRnnScriptState;
  rowstep : TForEachRow;
begin
  scriptor := cmdInfo^.scriptor as TJvGenDBScript;
  sspop :=  scriptor.GetScriptState;

  if sspop.scriptlabel = kFOR_EACH_ROW then
  begin
    rowstep :=  sspop.refptr as TForEachRow;
    if not rowstep.NextRow(sspop) then
      scriptor.PopScriptState
    else
      scriptor.SetScriptState(sspop);
  end
  else
    raise EJvRnnScript.create('END_FOR_ROWS without FOR_ROWS');
end;

procedure CloseDBData(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  ds : TJvRnnDatasetLink;
  ds_name : string;
  script : TJvGenDBScript;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dataset name');
  ds_name := ps.token;
  script := TJvGenDBScript(cmdInfo^.scriptor);
  ds := script.FindDatasetLink(ds_name);
  if (ds = nil) then
    Throw(cmdInfo, Format('Dataset %s not found',[ds_name]));
  script.FreeDatasetLink(ds);
end;

procedure ExecDBSQL(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  cmd : TStrings;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  if not ps_skipwhite(ps) then
    Throw(cmdInfo,'Empty command');
  cmd := TStringList.Create;
  try
    cmd.Text := ps.p;
    dblink.ExecuteStd(cmd);
  finally
    cmd.free;
  end;
end;

procedure SetDBAlias(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  script : TJvGenDBScript;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, 'No dblink name');
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  script := TJvGenDBScript(cmdInfo^.scriptor);
  script.GenDB := dblink;
end;

function DBExistsTable(sct: TJvRnnCustomScript; args : TStrings) : string;
var
  script:TJvGenDBScript;
begin
  script := sct as TJvGenDBScript;
  script.CheckGenDB;
  if (script.GenDB.TablesExist(args)) then
    Result := 'Y'
  else
    Result := 'N';
end;


procedure BatchCreate_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  script : TJvGenDBScript;
  keylist : TStrings;
  ds_name : string;
  ds : TJvRnnDatasetLink;
  new_table_name : string;
  def : TFieldDef;
  dsdefs : TFieldDefs;
  csql : string;
  ctflds : integer;
  i : integer;

  function IsKeyWord(const keyword : string):boolean;
  begin
    Result := (ps_nexttokentype(ps)) AND (CompareText(ps.token,keyword)=0);
  end;

  procedure CheckKeyWord(const keyword : string);
  begin
    if not IsKeyWord(keyword) then
      Throw(cmdInfo, Format(kKEYWORD_EXPECTED,[keyword]));
  end;

  function Matches(const keyword : string) : boolean;
  begin
    Result := (CompareText(ps.token,keyword)=0);
  end;

  procedure CheckLeftParen;
  begin
    if (not ps_nexttokentype(ps)) or (ps.token[1] <> '(') then
      Throw(cmdInfo, Format(kCHARACTER_EXPECTED,['(']));
  end;

begin
  ps_init(ps,cmdInfo^.cmdcstr);

  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  script := TJvGenDBScript(cmdInfo^.scriptor);
  CheckKeyWord(KW_SOURCE);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DATASET_NAME);
  ds := script.FindDatasetLink(ps.token);
  if (ds = nil) then
    Throw(cmdInfo, Format(kDATASET_NOT_FOUND,[ds_name]));
  CheckKeyWord(KW_TABLENAME);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_TABLENAME);
  new_table_name := ps.token;
  keylist := nil;
  dsdefs := ds.Dataset.FieldDefs;
  csql := Format('create table %s (', [new_table_name]);
  ctflds := 0;
  while (ps_nexttokentype(ps)) do
  begin
    if Matches(KW_SOURCE_FIELD) then
    begin
      try
        keylist :=  script.ParseFunctionArgs(ps.p);
        if (keylist=nil) or (keylist.count < 1) then
          Throw(cmdInfo, Format('1 arguments to %s expected',[KW_SOURCE_FIELD]));
        def := dsdefs.Find(keylist[0]);
        if (ctflds > 0) then
          csql := csql + ',';
        Inc(ctflds);
        csql :=  csql + dblink.GetSQLFieldType(def);
        if (keylist.count > 1) then
          csql := csql + Format(' %s',[keylist[1]]);
      finally
        keylist.free;
        keylist := nil;
      end;
    end else
    if Matches(KW_ADD_FIELD) then
    begin
      try
        keylist :=  script.ParseFunctionArgs(ps.p);
        if (keylist=nil) or (keylist.count < 1) then
          Throw(cmdInfo, Format('1 arguments to %s expected',[KW_SOURCE_FIELD]));
        if (ctflds > 0) then
          csql := csql + ',';
        csql := csql + Format('%s',[keylist[0]]);
        if (keylist.count > 1) then
          csql := csql + Format(' %s',[keylist[1]]);
      finally
        keylist.free;
        keylist := nil;
      end;
    end else
    if Matches(KW_PRIMARY_KEY) then
    begin
      try
        keylist :=  script.ParseFunctionArgs(ps.p);
        if (keylist=nil) or (keylist.count < 1) then
          Throw(cmdInfo, Format('1 arguments to %s expected',[KW_SOURCE_FIELD]));
        csql := csql + ', primary key (';
        for i := 0 to keylist.count-1 do
        begin
          if (i > 0) then
            csql := csql + ',';
          csql := csql + Format('%s',[keylist[i]]);
        end;
        csql := csql + ')';
      finally
        keylist.free;
        keylist := nil;
      end;
    end;
  end;
  csql := csql + ')';
  dblink.ExecSQLStr(csql);
end;

procedure BatchUpdate_F(cmdInfo : PJvScriptCmdInfo);
const
  KIX_FIELDS = 0;
  KIX_KEYS = 1;
  KIX_ACTION = 2;
  KIX_TABLENAME = 3;
  KIX_SOURCE = 4;
  maxkeyword = 4;

  keywords : array[0..maxkeyword] of string =
    (KW_FIELDS,KW_KEYS,KW_ACTION, KW_TABLENAME,KW_SOURCE);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
  script : TJvGenDBScript;
  slist : TStrings;
  ds_name : string;
  ds : TJvRnnDatasetLink;
  dest_table_name : string;
  table_fields : TJvCCObjArray;
  key_fields : TJvCCObjArray;
  haskeyword : array[0..4] of boolean;
  keywordix : integer;
  i : integer;
  ops : TDBLinkOpSet;
  s : string;

  function MatchKeyWord(const keyword : string; var keyix:integer) : boolean;
  var
    kix : integer;
  begin
    for kix := 0 to maxkeyword do
    begin
      if COMPARETEXT(keyword, keywords[kix]) = 0 then
      begin
        keyix := kix;
        Result := TRUE;
        exit;
      end;
    end;
    Result := FALSE;
  end;

  procedure CheckLeftParen;
  begin
    if (not ps_nexttokentype(ps)) or (ps.token[1] <> '(') then
      Throw(cmdInfo, Format(kCHARACTER_EXPECTED,['(']));
  end;

  procedure CheckDataset;
  begin
    if (ds = nil) then
      Throw(cmdInfo, 'SOURCE dataset not specified');
  end;

begin
  ps_init(ps,cmdInfo^.cmdcstr);

  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  script := TJvGenDBScript(cmdInfo^.scriptor);
  ops := [];
  for i := 0 to 4 do
    haskeyword[i] := FALSE;
  slist := nil;
  table_fields := nil;
  key_fields := nil;
  ds := nil;
  TRY
    while (ps_nexttokentype(ps)) do
    begin
      if (not MatchKeyWord(ps.token,keywordix)) then
        Throw(cmdInfo, Format(kKEYWORD_EXPECTED,[ps.token]));
      CASE keywordix of
      KIX_FIELDS :
        begin
          CheckLeftParen;
          slist := script.ParseFunctionArgs(ps.p);
          CheckDataset;
          table_fields := script.GetFieldDefList(ds.Dataset,slist);
        end;
      KIX_KEYS :
        begin
          CheckLeftParen;
          slist :=  script.ParseFunctionArgs(ps.p);
          CheckDataset;
          key_fields := script.GetFieldDefList(ds.Dataset,slist);
        end;
      KIX_ACTION :
        begin
          CheckLeftParen;
          slist :=  script.ParseFunctionArgs(ps.p);
          for i := 0 to slist.count-1 do
          begin
            s := slist[i];
            if COMPARETEXT(s, 'APPEND') = 0 then
              ops := ops + [lop_append]
            else
            if COMPARETEXT(s, 'UPDATE') = 0 then
              ops := ops + [lop_update]
            else
            if COMPARETEXT(s, 'DELETE') = 0 then
              ops := ops + [lop_delete]
            else
            if COMPARETEXT(s, 'INSERTNEW') = 0 then
              ops := ops + [lop_insertnew];
          end;
        end;
      KIX_TABLENAME :
        begin
          if (not ps_nexttokentype(ps)) then
            Throw(cmdInfo, 'Destination table name missing');
          dest_table_name := ps.token;
        end;
      KIX_SOURCE :
        begin
          if (not ps_nexttokentype(ps)) then
             Throw(cmdInfo, 'Source dataset name missing');
          ds := script.FindDatasetLink(ps.token);
          if (ds = nil) then
            Throw(cmdInfo, Format(kDATASET_NOT_FOUND,[ds_name]));
        end;
      END; {CASE}
      slist.free;
      slist := nil;
   end;

    dblink.AppendUpdateFromDataset(ds.Dataset, dest_table_name,
        table_fields,
        key_fields, ops);
  FINALLY
    slist.free;
    table_fields.free;
    key_fields.free;
  END;
end;


procedure TransBegin_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.BeginTransaction;
end;

procedure TransCommit_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.Commit;
end;

procedure TransCommitRetaining_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.CommitRetaining;
end;

procedure TransRollback_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);
  dblink.Rollback;
end;

procedure SetDBParams(cmd : PJvScriptCmdInfo);
var
  ps : TParseState;
  linkname : string;
  slist : TStrings;
  dblink : TJvRnnGenDBLink;
begin
  ps_init(ps,cmd^.cmdcstr);
  // read the database link name, its class type, and
  //   a line by line parameter list
  if (ps_nexttoken(ps)) then
    linkname := ps.token;
  slist := ParseLines(ps.p,FALSE);//si.ParseFunctionArgs(ps.p);
  if (slist <> nil) then
  try
    dblink := Sub_GetDBLink(cmd, linkname);
    dblink.ConnectLink.Params := slist;
  finally
    slist.free;
  end;
  (*
  try
    for i := 1 to cmdlist.count-1 do
    begin
      slist.Add(cmdlist[i]);
    end;
    si.NewDBLink(linkname, linktype, slist);
  finally
    slist.free;
  end; *)
end;

function GetSQLFieldValue(fld : TField) : string;
begin
  if fld.IsNull then
    Result := 'NULL'
  else
  CASE fld.DataType of
    ftDate :
      Result := GetSQLDateStr(fld.AsDateTime);
    ftDateTime :
      Result := GetSQLDateTimeStr(fld.AsDateTime);
  ELSE
    Result := fld.AsString;
  END;
end;

procedure SetFromDBQuery_F(cmdInfo : PJvScriptCmdInfo);
var
  ps : TParseState;
  link_name : string;
  varname : string;
  fldname : string;
  dblink : TJvRnnGenDBLink;
  sql : TStrings;
  savep : PChar;
  slist : TStrings;
  ds : TJvRnnDatasetLink;
  i : integer;
  dsfld : TField;
  script : TJvGenDBScript;
begin
  ps_init(ps,cmdInfo^.cmdcstr);
  if not ps_nexttoken(ps) then
    Throw(cmdInfo, kNO_DBLINK_NAME);
  link_name := ps.token;
  dblink := Sub_GetDBLink(cmdInfo, link_name);

  if (not ps_nexttokentype(ps)) or
    (CompareText('SET',ps.token)<>0) then
    raise EJvRnnScript.create('SET expected');
  // collect variable name = <fieldname> pairs
  // but store as <fieldname> = <var name>
  slist := TStringList.create;
  try
    while (true) do
    begin
      if (slist.count = 0) then
        if (not ps_nexttokentype(ps)) then
          raise EJvRnnScript.create('unexpected end of line');
      if CompareText(ps.token,'using') = 0 then
        break;
      varname := ps.token;
      if (not ps_nexttokentype(ps)) or (ps.toktype<>tok_equal) then
        raise EJvRnnScript.create('"=" expected');
      if (not ps_nexttokentype(ps)) then
        raise EJvRnnScript.create('field name expected');
      fldname := ps.token;
      if (not ps_nexttokentype(ps)) then  {',' or using}
        raise EJvRnnScript.create('unexpected end of line');
      slist.Values[fldname] := varname;
     end;
     savep := StrStartToken(ps.p);
     ps.p := savep;
     if (not ps_nexttokentype(ps))
       or (comparetext('select',ps.token)<>0) then  {',' or using}
       raise EJvRnnScript.create('"select" expected');
     sql := TStringList.Create;
     ds := nil;
     try
       sql.Text := StrPas(savep);
       ds := dblink.ReturnDataset(sql);
       ds.Dataset.first;
       script := TJvGenDBScript(cmdInfo^.scriptor);
       if (not ds.Dataset.eof) then
       begin
         for i := 0 to slist.count-1 do
         begin
           link_name := slist[i];
           fldname := JvRnnStrUt.GetStrName(link_name);
           varname := JvRnnStrUt.GetStrValue(link_name);
           dsfld := ds.Dataset.FieldByName(fldname);
           script.SetVariable(varname,GetSQLFieldValue(dsfld));
         end;
       end;
     finally
       ds.free;
       sql.Free;
     end;
  finally
    slist.free;
  end;
end;


constructor TJvGenDBScript.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FDBLinks := TJvCCObjArray.CreateMgr(TJvCCObjRecMgr);
  FDSList := TDatasetList.Create;
  FDBLinks.KeyCompare := DBLinkKeyCompare;
  FDBLinks.ObjCompare := DBLinkObjCompare;
  InstallCustomScriptOps(SELF);
  SELF.SetUserOp(kDB_CREATE,@CreateDBLink);
  SELF.SetUserOp(kDB_CONNECT,@ConnectDBLink);
  SELF.SetUserOp(kDB_PARAMS, @SetDBParams);
  SELF.SetUserOp(kDB_DISCONNECT,@DisconnectDBLink);
  SELF.SetUserOp(kDB_FREE,@FreeDBLink);
  SELF.SetUserOp(kDB_DSCREATE,@OpenDBData);
  SELF.SetUserOp(kDB_DSFREE,@CloseDBData);
  SELF.SetUserOp(kDB_EXEC,@ExecDBSQL);
  SELF.SetUserOp(kDB_SETDB,@SetDBAlias);
  SELF.SetUserOp(kDB_CREATE_SOURCE, @BatchCreate_F);
  SELF.SetUserOp(kDB_BATCH_APPENDUPDATE, @BatchUpdate_F);
  SELF.SetUserOp(kDB_SET, @SetFromDBQuery_F);
  SELF.SetUserOp(kDB_TRANS_BEGIN, @TransBegin_F);
  SELF.SetUserOp(kDB_TRANS_COMMIT, @TransCommit_F);
  SELF.SetUserOp(kDB_TRANS_COMMIT_RETAINING, @TransCommitRetaining_F);
  SELF.SetUserOp(kDB_TRANS_ROLLBACK, @TransRollback_F);
  SELF.SetUserOp(kFOR_EACH_ROW, @ForEachRow_F);
  SELF.SetUserOp(kNEXT_EACH_ROW, @NextEachRow_F);
  SELF.SetUserFunc('DB_EXISTS_TABLE',@DBExistsTable);
end;

destructor TJvGenDBScript.Destroy;
begin
  FDBLinks.Free;
  FDSList.Free;
  inherited;
end;

function TJvGenDBScript.FindDatasetLink(const name : string):TJvRnnDatasetLink;
begin
  if not DSList.LocateKey(@name, Result) then
    Result := nil;
end;

procedure TJvGenDBScript.FreeDatasetLink(ds_link : TJvRnnDatasetLink);
begin
  DSList.DeleteItem(ds_link);
  ds_link.free;
end;

procedure TJvGenDBScript.ForgetDBLink(dblink:TJvRnnGenDBLink);
begin
  if (dblink = GenDB) then
    GenDB := nil;
  dblink.DSList.CallActionMethod(RemoveDSRef,nil);
  FDBLinks.DeleteItem(dblink);
end;

procedure TJvGenDBScript.CheckGenDB;
begin
  if (GenDB = nil) then
     raise EJvRnnScript.Create('No current DB link');
end;

procedure TJvGenDBScript.Notification(
      AComponent:TComponent; operation:TOperation);
begin
  if operation = opRemove then
  begin
    if (AComponent is TJvRnnGenDBLink) then
      ForgetDBLink(TJvRnnGenDBLink(AComponent));
  end;
end;

function TJvGenDBScript.EvaluateName(var ps : TParseState) : string;
var
  ds : TJvRnnDatasetLink;
  vf : TField;
begin
  vf := nil;
  if (ps.p^ = '.') then
  begin
    ds := FindDatasetLink(ps.token);
    if (ds <> nil) then
    begin
      Inc(ps.p);
      if not ps_nexttokentype(ps) then
        raise EJvRnnScript.Create('No Fieldname');
      vf := ds.Dataset.FieldByName(ps.token);
      Result := vf.AsString;
    end;
  end;
  if (vf = nil) then
    Result := inherited EvaluateName(ps);
end;


function TJvGenDBScript.NewDataset(
  const dsname:string;
  ds_link:TJvRnnGenDBLink) : TJvRnnDatasetLink;
begin
  Result := SELF.FindDatasetLink(dsname);
  if (Result <> nil) then
  begin
    Result := nil;
    raise EJvRnnScript.create(
      Format('Dataset %s already exists',[dsname]));
  end;
  Result := ds_link.ReturnDataset(nil);
  if (Result<>nil) then
  try
    Result.DSName := dsname;
    ds_link.DSList.AppendItem(Result);
    SELF.DSList.AppendItem(Result);
  except
    FreeAndNil(Result);
    raise;
  end;
end;



procedure TJvGenDBScript.ExecuteStd(cmd : TStrings);
begin
  if (GenDB = nil) then
    raise EJvRnnScript.create('No current DB link');
  GenDB.ExecuteStd(cmd);
end;


initialization
  gDBLinkClasses := TJvCCArray.CreateMgr(TDBLinkClassMgr);
finalization
  gDBLinkClasses.Free;
end.
