{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnDB_IBX.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnDB_IBX;

interface
uses Classes,SysUtils,JvRnnGenDB,IBDatabase,IBQuery,IBSQL,IB,
    JvRnnDBScript,JvRnnScCst,DB;

type
  TJvRnnDBLinkIBX = class(TJvRnnGenDBLink)
  protected
    FDatabase : TIBDatabase;
    FSQL : TIBSQL;
    FTransaction : TIBTransaction;
    FIsDDL : boolean;
    FWasCommit : boolean;
    FCommitDDL : boolean;

    procedure CreateAndOpenDatabase(ps : PChar);
    procedure OpenDatabase(ps : PChar);
    procedure CreateScriptDatabase(const dbpath:string; dbparams:TStrings);
    procedure OpenScriptDatabase(const dbpath:string; dbparams:TStrings);
    procedure DisconnectScriptDatabase;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    class function DBLinkDescription : string; override;
    procedure SetConnected (islive : boolean); override;
    function  GetConnected : boolean; override;
    function ReturnDataset(slist:TStrings) : TJvRnnDatasetLink; override;
    function TablesExist(tablist:TStrings) : boolean; override;
    procedure ExecuteStd(cmd:TStrings);  override;
    function GetSQLFieldType(
      flddef : TFieldDef) : string; override;
    function  ParseSQL(cmd : TStrings): boolean;
    function RecordsExist(
      const select_sql:string;  xparams : variant) : boolean; override;
    procedure BeginTransaction; override;
    procedure Commit; override;
    procedure Rollback; override;
    procedure CommitRetaining; override;

    procedure UpdateRecords(
      const update_sql:string; xparams : variant); override;
    procedure InsertRecords(
      const insert_sql:string; xparams : variant); override;
   end;

  TJvRnnDatasetIBX = class(TJvRnnDatasetLink)
  protected
    procedure SetSQL(slist:TStrings); override;
    function GetSQL : TStrings; override;
  public
    constructor Create(dblink:TJvRnnGenDBLink); override;
    destructor Destroy; override;
  end;

procedure Register;

implementation
uses JvRnnStrUt,JvRnnPars,JvRnnStrLs;

const
  rdb_NILDataset = 0;
  kIB_sql_role_name = 'sql_role_name';
  kIB_password = 'password';
  kIB_user_name = 'user_name';
  kValueFormat = '%s=%s';
  kRoleName = 'RoleName';
  kEmptyStr = '';

procedure RDBError(i:Integer);
var
  s : string;
begin
  CASE i of
  rdb_NILDataset :
    s := 'NIL dataset';
  END;
  raise EJvRnnDB.Create(s);
end;

constructor TJvRnnDBLinkIBX.Create(AOwner:TComponent);
begin
  inherited;
  FDatabase := TIBDatabase.Create(SELF);
  FSQL := TIBSQL.Create(SELF);
  FSQL.Database := FDatabase;
  FTransaction := TIBTransaction.Create(SELF);
  FTransaction.AddDatabase(FDatabase);
  FDatabase.AddTransaction(FTransaction);
  FDatabase.DefaultTransaction := FTransaction;
  FSQL.Transaction := FTransaction;
end;

destructor TJvRnnDBLinkIBX.Destroy;
begin
  FDatabase.Free;
  inherited;
end;

procedure TJvRnnDBLinkIBX.SetConnected(islive : boolean);
var
  i : integer;
  slist : TStrings;
  temp : string;
  loginct : integer;
begin
  if (FDatabase.Connected = islive) then
    exit;
  if islive then
  begin
    slist := TStringList.Create;
    try
      loginct := 2;
      slist.Assign(FConnectLink.Params);
      i := slist.IndexOfName(kRoleName);
      if (i >= 0) then
      begin
        temp := GetStrValue(slist[i]);
        slist[i] := Format(kValueFormat,[kIB_sql_role_name,temp]);
      end;
      i := slist.IndexOfName(kUserName);
      if (i >= 0) then
      begin
        temp := GetStrValue(slist[i]);
        slist[i] := Format(kValueFormat,[kIB_user_name,temp]);
        Dec(loginct);
      end;
      i := slist.IndexOfName(kPassword);
      if (i >= 0) then
      begin
        temp := GetStrValue(slist[i]);
        slist[i] := Format(kValueFormat,[kIB_password,temp]);
        Dec(loginct);
      end;
      i := slist.IndexOfName(kDatabase);
      if (i >= 0) then
      begin
        FDatabase.DatabaseName := GetStrValue(slist[i]);
        slist.Delete(i);
      end;
      FDatabase.Params := slist;
      FDatabase.LoginPrompt := (loginct > 0);
    finally
      slist.free;
    end;
  end;
  FDatabase.Connected := islive;
end;

function TJvRnnDBLinkIBX.GetConnected : boolean;
begin
  Result := FDatabase.Connected;
end;

function TJvRnnDBLinkIBX.TablesExist(tablist:TStrings) : boolean;
var
  names : TStrings;
  i : integer;
begin
  names := TStringList.create;
  Result := FALSE;
  try
    FDatabase.GetTableNames(names,FALSE);
    for i := 0 to tablist.count-1 do
    begin
      Result := (names.IndexOf(tablist[i]) >= 0);
      if not Result then
        exit;
    end;
  finally
    names.free;
  end;
end;

function  TJvRnnDBLinkIBX.ParseSQL(cmd : TStrings): boolean;
var
  cs : PChar;
  p : PChar;
  token : string;
  ptype : TJvRnnParenType;
  toktype : TJvRnnTokenType;
  temp : boolean;
  dialect : integer;

  function TokenEquals(const invalue : string) : boolean;
  begin
    Result := (COMPARETEXT(token, invalue) = 0);
  end;

  function GotNextTokenType : boolean;
  begin
    Result := StrNextTokenType( p, token, ptype, toktype) ;
  end;

  function GotNextName : boolean;
  begin
    Result := GotNextTokenType and (toktype=tok_name);
  end;

  function GotNextInteger : boolean;
  begin
    Result := GotNextTokenType and (toktype=tok_integer);
  end;

begin
	 {check for SET TERM, CONNECT and CREATE DATABASE parameters}
  cs := StrListToCStr(cmd,' ');
  try
    Result := FALSE;
    FIsDDL := FALSE;
    FWasCommit := FALSE;

    p := cs;
    if GotNextName and (ptype=paren_none) then
    begin
      if TokenEquals('SET') then
      begin
        if GotNextName then
        begin
          if TokenEquals('SQL') then
          begin
            if GotNextName and TokenEquals('DIALECT') and GotNextInteger then
            begin
              dialect := StrToIntDef(token,3);
              FDatabase.SQLDialect := dialect;
              Result := TRUE;
            end else
              raise EJvRnnScript.create('Unsupported option '+StrPas(cs));
          end
          else if TokenEquals('TERM') then
          begin
            {expect ASCII printable character for new statement terminator}
            if StrNextTokenType( p, token, ptype, toktype)
              and (toktype=tok_bad) then
            begin
            {not a name or math op or relation op, must be a character?}
              Scriptor.StdTerminator := token[1];
              Result := TRUE;
            end
          end
          else if TokenEquals('AUTO')
            OR TokenEquals('AUTODDL') then
          begin
            {expect ON or OFF}
            temp := FALSE;
            if GotNextName then
            begin
              if TokenEquals('ON') then
              begin
                temp := TRUE;
                FCommitDDL := TRUE;
              end else if TokenEquals('OFF') then
              begin
                temp := TRUE;
                FCommitDDL := FALSE;
              end;
            end;
            if not temp then
              raise EJvRnnScript.create('ON or OFF expected');
            Result := TRUE;
          end
          else if TokenEquals('TRANSACTION') then
          begin
             BeginTransaction;
             Result := TRUE;
          end;
        end;
      end
      else if TokenEquals('CONNECT') then
      begin
        OpenDatabase(p);
        Result := TRUE;
      end
      else if TokenEquals('CREATE') then
      begin
        if GotNextName and TokenEquals('DATABASE') then
        begin
          CreateAndOpenDatabase(p);
          Result := TRUE;
        end else
          FIsDDL := TRUE;
      end
      else if TokenEquals('ALTER') OR TokenEquals('DROP') OR
        TokenEquals('GRANT') OR TokenEquals('REVOKE') then
        FIsDDL := TRUE
      else if TokenEquals('COMMIT') then
      begin
        FWasCommit := TRUE;
      end;
    end;
  finally
    StrDispose(cs);
  end;
end;

function TJvRnnDBLinkIBX.GetSQLFieldType(
      flddef : TFieldDef) : string;
      
  function sizeStr : string;
  begin
    Result := IntToStr(flddef.Size);
  end;

begin
  CASE flddef.DataType of
    ftString:
      Result := 'VARCHAR('+sizeStr+')';
    ftSmallInt:
      Result := 'SMALLINT';
    ftInteger:
      Result := 'INTEGER';
    ftFloat :
      Result := 'NUMERIC';
    ftDate, ftTime, ftDateTime :
      if (FDatabase.SQLDialect < 3) then
        Result := 'DATE'
      else
        Result := 'TIMESTAMP';
    ftBoolean :
      Result := 'BOOLEAN';
  END;{CASE}
  Result := flddef.Name + ' '+Result;
end;

class function TJvRnnDBLinkIBX.DBLinkDescription : string;
begin
  Result := 'Interbase Express Datalink';
end;

procedure TJvRnnDBLinkIBX.BeginTransaction;
begin
  If FTransaction.InTransaction then
    raise EJvRnnScript.create('Already in Transaction');
  FTransaction.Active := TRUE;
end;

procedure TJvRnnDBLinkIBX.Commit; 
begin
  If not FTransaction.InTransaction then
    raise EJvRnnScript.create('Not in Transaction');
  FTransaction.Commit;
end;

procedure TJvRnnDBLinkIBX.CommitRetaining;
begin
  If not FTransaction.InTransaction then
    raise EJvRnnScript.create('Not in Transaction');
  FTransaction.CommitRetaining;
end;


procedure TJvRnnDBLinkIBX.Rollback; 
begin
  If FTransaction.InTransaction then
    raise EJvRnnScript.create('Not in Transaction');
  FTransaction.Rollback;
end;


procedure TJvRnnDBLinkIBX.ExecuteStd(cmd : TStrings);
begin
  if ParseSQL(cmd) then
    exit;
    
  if FSQL.Open then
    FSQL.CheckClosed;
  if cmd.count=0 then exit;

  FSQL.Database := FDatabase;
  FSQL.SQL.Assign(cmd);
  try
    FTransaction.Active := TRUE;  {if not already active}
    FSQL.Prepare;
    FSQL.ExecQuery;
    FSQL.Close;
    if (FWasCommit) OR ((FIsDDL) and (FCommitDDL)) then
    begin
      if (FTransaction.InTransaction) then
        FTransaction.Commit;     {supports auto-commit DDL?}
    end;
  except
    On ex1:EIBError do
    begin
      Scriptor.LogMsg('Error Running ' + cmd[0] + ' : '+ex1.Message);
      raise;
    end;
  end;
end;

procedure TJvRnnDBLinkIBX.OpenDatabase(ps : PChar);
var
  dbparams : TJvRnnStrList;
  token : string;
  p : PChar;
  ptype : TJvRnnParenType ;
  toktype : TJvRnnTokenType;
  dbname : string;

  function NextToken : boolean;
  begin
    Result := StrNextTokenType( p, token, ptype, toktype) and (toktype<>tok_bad);
  end;

  procedure ValueExpected(const inOption : string);
  begin
    raise EJvRnnScript.create('Value expected for '+inOption);
  end;

  procedure CheckOption(const inOption:string);
    procedure  sub_getvalue;
    begin
      if not NextToken then
        ValueExpected(inOption);
      if toktype = tok_equal then
        if not NextToken then
          ValueExpected(inOption);
    end;

  begin
    if (COMPARETEXT(inOption,'USER')=0) then
    begin
      sub_getvalue;
      dbparams.Values['user_name'] := token;
    end else if (COMPARETEXT(inOption,'PASSWORD')=0) then
    begin
      sub_getvalue;
      dbparams.Values['password'] := token;
    end else
      raise EJvRnnScript.create('Unknown option '+inOption);
  end;

begin
  p := ps;
  if NextToken then
  begin
    dbname := token;
    {translate USER and PASSWORD options}
    dbparams := TJvRnnStrList.Create;
    try
      while NextToken do
        CheckOption(token);
      OpenScriptDatabase(dbname,dbparams);
    finally
      dbparams.free;
    end;
  end;
end;

procedure TJvRnnDBLinkIBX.CreateAndOpenDatabase(ps : PChar);
var
  p : PChar;
  token : string;
  ptype : TJvRnnParenType ;
  toktype : TJvRnnTokenType;
  dbname : string;
  dbparams : TJvRnnStrList;

  function NextToken : boolean;
  begin
    Result := (StrNextTokenType( p, token, ptype, toktype) and ((toktype=tok_name)or(toktype=tok_quotetext)));
  end;

begin
  {CREATE DATABASE "?DB_PATH" PAGE_SIZE 4096 USER "SYSDBA" PASSWORD "masterkey" }
  p := ps;
  if NextToken  then
  begin
    dbname := token;
    dbparams := TJvRnnStrList.create;
    try
      dbparams.Add(StrPas(p));
      CreateScriptDatabase(dbname,dbparams);
    finally
      dbparams.free;
    end;
  end;
end;

function  TJvRnnDBLinkIBX.ReturnDataset(slist:TStrings) : TJvRnnDatasetLink;
begin
  Result := TJvRnnDatasetIBX.Create(SELF);
  try
    if (slist <> nil) then
    begin
      Result.SQL := slist;
      if (Connected) then
        Result.Dataset.Open;
    end;
  except
    FreeAndNil(Result);
    raise;
  end;
end;

procedure  TJvRnnDBLinkIBX.DisconnectScriptDatabase;
begin
  if FDatabase.Connected then
    FDatabase.Connected := FALSE;
end;

procedure  TJvRnnDBLinkIBX.OpenScriptDatabase(const dbpath:string; dbparams:TStrings);
begin
  DisconnectScriptDatabase;
  FDatabase.DatabaseName := dbpath;
  FDatabase.Params.Assign(dbparams);
  FDatabase.Connected := TRUE;
end;

procedure  TJvRnnDBLinkIBX.CreateScriptDatabase(const dbpath:string; dbparams:TStrings);
begin
  DisconnectScriptDatabase;
  FDatabase.DatabaseName := dbpath;
  FDatabase.Params.Assign(dbparams);
  FDatabase.CreateDatabase;
end;

function TJvRnnDBLinkIBX.RecordsExist(
  const select_sql:string;  xparams : variant) : boolean;
var
  srec : TIBSQL;
  i : integer;
begin
  srec := TIBSQL.Create(nil);
  try
    srec.Database := SELF.FDatabase;
    srec.Transaction := SELF.FTransaction;
    srec.GotoFirstRecordOnExecute := TRUE;
    srec.SQL.Text := select_sql;
    srec.Prepare;
    with srec do
    for i := 0 to params.count-1 do
    begin
      params.Vars[i].AsVariant := xparams[i];
    end;
    srec.ExecQuery;
    Result := (not srec.EOF);
  finally
    srec.free;
  end;
end;

procedure TJvRnnDBLinkIBX.UpdateRecords(
  const update_sql:string; xparams : variant);
var
  srec : TIBSQL;
  i : integer;
begin
  srec := TIBSQL.Create(nil);
  try
    srec.Database := SELF.FDatabase;
    srec.Transaction := SELF.FTransaction;
    srec.GotoFirstRecordOnExecute := TRUE;
    srec.SQL.Text := update_sql;
    srec.Prepare;
    with srec do
    for i := 0 to params.count-1 do
    begin
      params.Vars[i].AsVariant := xparams[i];
    end;
    srec.ExecQuery;
  finally
    srec.free;
  end;
end;

procedure TJvRnnDBLinkIBX.InsertRecords(
  const insert_sql:string; xparams : variant);
var
  srec : TIBSQL;
  i : integer;
begin
  srec := TIBSQL.Create(nil);
  try
    srec.Database := SELF.FDatabase;
    srec.Transaction := SELF.FTransaction;
    srec.GotoFirstRecordOnExecute := TRUE;
    srec.SQL.Text := insert_sql;
    srec.Prepare;
    with srec do
    for i := 0 to params.count-1 do
    begin
      params.Vars[i].AsVariant := xparams[i];
    end;
    srec.ExecQuery;
  finally
    srec.free;
  end;
end;

constructor TJvRnnDatasetIBX.Create(dblink:TJvRnnGenDBLink);
var
  qry : TIBQuery;
  ibx : TJvRnnDBLinkIBX;
begin
  inherited Create(dblink);
  qry := TIBQuery.Create(dblink);
  ibx := TJvRnnDBLinkIBX(dblink);
  qry.Database := ibx.FDatabase;
  qry.Transaction := ibx.FTransaction;
  FDataset := qry;
end;

destructor TJvRnnDatasetIBX.Destroy;
begin
  FDataset.Free;
  FDataset := nil;
  inherited;
end;

function TJvRnnDatasetIBX.GetSQL : TStrings;
begin
  if FDataset = nil then
    RDBError(rdb_nilDataset);
  Result := TIBQuery(FDataset).SQL;
end;

procedure TJvRnnDatasetIBX.SetSQL(slist:TStrings);
begin
  if FDataset = nil then
    RDBError(rdb_nilDataset);
  TIBQuery(FDataset).SQL := slist;
end;




procedure SetupDB_IBX;
begin
  TJvGenDBScript.RegisterDBLink('IBX',TJvRnnDBLinkIBX);
end;

procedure Register;
begin
  RegisterComponents('RnnLib',[TJvRnnDBLinkIBX]);
end;

initialization
  SetupDB_IBX;
end.
