{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnPars.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnPars;

interface
Uses sysutils;

const
  kParenCt = 6;
  startParen:array[0..kParenCt] of Char = ('"','''','<','{','[','(',#0);
  endParen:array[0..kParenCt] of Char = ('"','''','>','}',']',')',#0);
  kLastQuoteIndex = 1;

  SetStartNameChars : set of Char = ['A'..'Z','a'..'z', '_'];
  SetNameChars : set of Char = ['0'..'9','A'..'Z','a'..'z', '_'];
  SetStartNumber : set of Char = ['-','+','0'..'9'];
  SetDigits : set of Char = ['0'..'9'];
  SetMathOps : set of Char = ['-','+','*','/'];
  SetRelOps : set of Char = ['<','>','='];

type
  TCharSet = set of Char;

  TJvRnnParenType = (paren_none, paren_dquote, paren_squote, paren_angle,
                   paren_brace, paren_square,paren_round);
  TJvRnnTokenType = (
     tok_none, tok_bad, tok_quotetext, tok_paren, tok_name, tok_varname,
     tok_symbol,tok_mathop, tok_integer, tok_real,
     tok_lessthan, tok_lesseq, tok_equal, tok_notequal,
     tok_greatereq, tok_greaterthan);
type
  EDuparse = class(Exception);

function StrNextToken( var sx : PChar;
  var token : string; var parentype: TJvRnnParenType) : boolean;
function  StrNextTokenType(var sx : PChar; var token : string;
  var parentype : TJvRnnParenType; var tokentype : TJvRnnTokenType) : boolean;

function StrEndLine( var sx : PChar; var line : string) : boolean;

function StrEndToken(sx:PChar):PChar;
function StrStartToken(sx:PChar) : PChar;
function FindCharIndex(sx : PChar;  inChar : Char):integer;
function GetAlphaToken(sx:PChar) : String;
Function GetAliasTableName(const s : string; var aliasname, tablename : string):boolean;
procedure AddCStrSeg(var w : string; sp, ep : PChar);
function RemoveComments(wasInComment : Boolean; var s : string) : Boolean;
procedure InitDuparse;

type
  TParseState = record
    p : PChar;
    token : string;
    toktype : TJvRnnTokenType;
    partype : TJvRnnParenType;
  end;

procedure ps_init(var prec:TParseState; sx:PChar);
function ps_skipwhite(var prec:TParseState):boolean;
function ps_nexttokentype(var prec:TParseState):boolean;
function ps_nexttoken(var prec:TParseState):boolean;
function ps_getline(var prec:TParseState):boolean;


var
  gSetStartNameChars : TCharSet;
  gSetNameChars : TCharSet;
  gSetStartNumber : TCharSet;
  gSetDigits : TCharSet;
  gSetMathOps : TCharSet;
  gSetRelOps : TCharSet;

implementation
Uses JvRnnStrUt;

function StripTwinQuoteChars(var p : PChar) : string; forward;


procedure InitDuparse;
begin
  gSetStartNameChars := SetStartNameChars;
  gSetNameChars := SetNameChars;
  gSetStartNumber := SetStartNumber;
  gSetDigits := SetDigits;
  gSetMathOps := SetMathOps;
  gSetRelOps := SetRelOps;
end;

procedure ps_init(var prec:TParseState; sx:PChar);
begin
  with prec do
  begin
    p := sx;
    token := '';
    toktype := tok_none;
    partype := paren_none;
  end;
end;

function ps_nexttoken(var prec:TParseState):boolean;
begin
  Result := StrNextToken(prec.p, prec.token, prec.partype);
end;

function ps_nexttokentype(var prec:TParseState):boolean;
begin
  Result := StrNextTokentype(
    prec.p, prec.token, prec.partype, prec.toktype);
end;

function ps_skipwhite(var prec:TParseState):boolean;
begin
  prec.p := StrStartToken(prec.p);
  Result := (prec.p <> #0);
end;

function ps_getline(var prec:TParseState):boolean;
begin
  Result := StrEndLine(prec.p, prec.token);
end;

procedure AddCStrSeg(var w : string; sp, ep : PChar);
{AddCStrSeg: add all characters from pointer sp to not including ep,
or if ep is null, add C string from sp}
var
  i : integer;
  k : integer;
begin
  if (ep <> nil) then
    i := (ep - sp)
  else
    i := StrLen(sp);
  k := Length(w);
  SetLength(w, k+i);
  System.Move(sp^,w[k+1],i);
end;

function StrEndTokenSet(sx:PChar; const validSet: TCharSet ):PChar;
begin
  while (sx^ <> #0) do
  begin
    if  not ((sx^) in validSet) then
      break;
    Inc(sx);
  end;
  Result := sx;
end;

function StrEndToken(sx:PChar):PChar;
begin
  while (sx^ <> #0) do begin
    if (ORD(sx^) <= ORD(CHAR($20))) or (sx^ = ',') then
      break;
    Inc(sx);
  end;
  Result := sx;
end;

function  StrEndNumber(sx:PChar; var tokentype : TJvRnnTokenType) : PChar;
var
  hasDP : boolean;
  hasScale : boolean;
  seqcount : integer;

  function IsDigit(dchar : Char) : boolean;
  begin
    Result := (dchar >= '0') AND (dchar <= '9');
  end;

begin
  hasDP := FALSE;
  hasScale := FALSE;
  seqcount := 0;
  while (sx^ <> #0) do
  begin
  if IsDigit(sx^) then
    begin
      Inc(sx);
      Inc(seqcount);
    end else if (sx^ = '.') then
    begin
      if hasScale or hasDP then
        break;
      hasDP := TRUE;
      Inc(sx);
      seqcount := 0;
    end else
    if (sx^='E') OR (sx^='e') then
    begin
      if hasScale then
        break;
      Inc(sx);
      if (sx^='+') OR (sx^='-') then
        Inc(sx);
      seqcount := 0;
    end else
      break;
  end;
  if (seqcount = 0) then
    tokentype := tok_bad
  else if (hasScale or hasDP) then
    tokentype := tok_real
  else
    tokentype := tok_integer;
  Result := sx;
end;

function  StrEndRelOp(sx:PChar; var tokentype : TJvRnnTokenType) : PChar;
begin
  {detect '=', '<>', '>', '<', '<=', '>='}
  Result := sx+1;
  CASE sx^ of
    '<':
      begin
        CASE Result^ of
          '>' :
            begin
            Inc(Result);
            tokentype := tok_notequal;
            end;
          '=' :
            begin
            Inc(Result);
            tokentype := tok_lesseq;
            end;
          ELSE
            tokentype := tok_lessthan;
        END;
      end;
    '>':
      begin
        CASE Result^ of
          '=' :
            begin
            Inc(Result);
            tokentype := tok_greatereq;
            end;
          ELSE
            tokentype := tok_greaterthan;
        END;
      end;
    '=':
      tokentype := tok_equal;
  END;
end;

function StrEndLine( var sx : PChar; var line : string) : boolean;
var
  p : PChar;
begin
  p := sx;
  while (p^ <> #0) and (p^ <> #13) and (p^ <> #10) do
    Inc(p);
  Result := (p > sx);
  SetLength(line,0);
  if Result then
    AddCStrSeg(line,sx,p);
  while (p^=#13) or (p^ = #10) do
    Inc(p);
  sx := p;
end;

{return index of inChar in parray, return -1 if not found}
function FindCharIndex(sx:PChar; inChar : Char):integer;
begin
  Result := 0;
  while( sx^ <> #0) do
  begin
    if ( sx^ = inChar) then
      exit;
    Inc(Result);
    Inc(sx);
  end;
  Result := -1;
end;

{
  StrNextTokenType
  returns a sequence of characters surrounded by quotes and quote type,
  after removing the quotes, or returns a specific token type,

  On return, sx points to next character after sequence
  token is returned sequence
  parentype is type of any quote (' ") removed
  '<' and '>' are relational operators, not parentheses
}

function  StrNextTokenType(var sx : PChar; var token : string;
  var parentype : TJvRnnParenType; var tokentype : TJvRnnTokenType) : boolean;
var
  i : integer;
  p : PChar;
  q : PChar;
begin
  Result := FALSE;
  parentype :=  paren_none;
  tokentype := tok_none;
  token := '';
  if (sx = nil) then
    exit;
  p := StrStartToken(sx);
  SetLength(token,0);
  if (p^ = #0) then
    exit;
  i := FindCharIndex(startparen, p^);
  if (i >= 0) and (i <= kLastQuoteIndex)  then
  {all paren types except '<' and '>'}
  begin
    tokentype := tok_quotetext;
    parentype := TJvRnnParenType(1+i);
    token := StripTwinQuoteChars(p);
    sx := p;
  end else
  begin
    if (i < 0) then
      i := FindCharIndex(endparen, p^);
    if ( i > kLastQuoteIndex) and
       (i <> (ORD(paren_angle)-1)) then
    begin
      q := p + 1;
      parentype := TJvRnnParenType(1+i);
      tokentype := tok_paren;
    end else
    begin
      parentype := paren_none;
      if (p^ in gSetStartNameChars) then
      begin
        q := StrEndTokenSet(p+1,gSetNameChars);
        tokentype := tok_name;
      end else
      if (p^ in gSetStartNumber) then
      begin
        q := StrEndNumber(p,tokentype);
      end else
      if (p^ in gSetMathOps) then
      begin
        q := p + 1;
        tokentype := tok_mathop;
      end else
      if (p^ in gSetRelOps) then
      begin
        q := StrEndRelOp(p,tokentype);
      end else
      begin
        q := p+1;
        tokentype := tok_bad;
      end;
    end;
    sx := q;
    AddCStrSeg(token,p,q);
  end;
  Result := (Length(token)>0);
end;

function StripTwinQuoteChars(var p : PChar) : string;
{on entry the first character is a quote which also terminates the string,
 but double occurence of quote character is turned to a single instance,
 and does not terminate the string,
 on return p points to first character after end quote}
var
  strip : Char;
  q : PChar;
begin
  Result := '';
  strip := p^;
  if strip = #0 then
    exit;
  Inc(p);
  if p^ = #0 then
  begin
    Result := strip; {one character only}
    exit;
  end;
  q := p;  {start of segment to be saved}
  while(true) do
  begin
    while (q^ <> #0) and (q^ <> strip) do
      Inc(q);
    if q^ = #0 then
      raise EDuparse.create('end quote missing');
    if q^ = strip then
    begin {save up to quote}
      AddCStrSeg(Result,p,q);
      Inc(q);
      p := q;
      if (q^ = strip) then
      begin
        Result := Result + strip;
        Inc(p);
      end else
        break;
      q := p;
    end;
  end;
end;
{
  StrNextToken
  Returns a sequence of characters seperated by whitespace,
  removes some kinds of parentheses.
  On return, sx points to next character after sequence,
  token is returned sequence,
  parentype is type of any parenthesis removed
}

function StrNextToken(var sx : PChar;
  var token : string; var parentype: TJvRnnParenType) : boolean;
var
  epos : integer;
  i : integer;
  p : PChar;
  q : PChar;
begin
  Result := FALSE;
  parentype := paren_none;
  token := '';
  if (sx = nil) then
    exit;
  p := StrStartToken(sx);
  if (p^ = #0) then
    exit;

  SetLength(token,0);
  i := FindCharIndex(startparen, p^);
  if (i >= 0) then
  begin
    parentype := TJvRnnParenType(1+i);
    if i <= kLastQuoteIndex then
      token := StripTwinQuoteChars(p)
    else
    begin
      Inc(p);
      q := p;
      epos := FindCharIndex(q, endparen[i]);
      if (epos < 0) then
        raise EDuparse.create('Matching '+endparen[i]+' missing');
      Inc(q,epos);
      AddCStrSeg(token,p,q); {add segment to before parenthesis}
      p := q+1; {exclude end parenthesis}
    end;
    sx := p;
  end else
  begin
    parenType := paren_none;
    q := StrEndToken(p);
    sx := q;
    AddCStrSeg(token,p,q);
  end;
  Result := (Length(token)>0);
end;

function StrStartToken(sx:PChar) : PChar;
begin
  while (sx^ <> #0) do
  begin
    if (ORD(sx^) > ORD(CHAR($20))) then
      break;
    Inc(sx);
  end;
  Result := sx;
end;

function GetAlphaToken(sx:PChar) : String;
var
  p : PChar;
  i : integer;  {return a continuous string of NameChars}
begin
  p := sx;
  i := 0;
  SetLength(Result,0);
  while (p^<>#0) do
  begin
    if not (p^ in gSetNameChars) then
      break;
    Inc(p);
    Inc(i);
  end;
  if (i > 0) then
    AddCStrSeg(Result,sx,p);
end;


Function GetTableName(s : PChar) : string;
var
  p : PChar;
begin
  Result := '';
  p := StrRScan(s,':');
  if (p = nil) then
    p := s
  else
    Inc(p);
  Result := StrPas(p);
end;

Function GetAliasName(s : PChar) : string;
var
  p1 : PChar;
  p2 : PChar;
begin
  Result := '';
  p1 := StrScan(s,':');
  p2 := StrRScan(s,':');
  if (p1 = nil) OR (p2 = nil) OR (p1>=p2) then
    exit;
  Inc(p1);
  while(p1 < p2) do
  begin
    Result := Result+ p1^;
    Inc(p1);
  end;
end;

Function GetAliasTableName(const s : string; var aliasname, tablename : string):boolean;
var
  p : PChar;
begin
  p := StrToPChar(s);
  aliasName := GetAliasName(p);
  if Length(aliasName) > 0 then
    tablename := GetTableName(p)
  else
    tablename := s;
  Result := Length(tablename) > 0;
end;

function RemoveComments(wasInComment : Boolean; var s : string) : Boolean;
var
  i : integer;
  cstart : integer;
  inquote : boolean;
  quotechar : Char;

  procedure sub_chopcomment(var cs : string; istart, iend : integer);
  begin
    if (istart > 1) then
    begin
      cs:= Copy(cs,1, istart-1) + Copy(cs, iend+1, Length(cs)-iend);
    end else
      cs := Copy(cs, iend+1, Length(cs)-1);
  end;
  {primative comment removal. Ignores /* starting inside balanced quotes}
begin
  i := 1;
  Result := wasInComment;
  if Result then
    cstart := 1
  else
    cstart := 0;

  inquote := FALSE;
  quotechar := ' ';
  while  i < Length(s) do
  begin
    if Result then
    begin  {look for end comment}
      if (s[i]='*') and (s[i+1]='/') then
      begin
        Inc(i);
        sub_chopcomment(s,cstart,i);
        i := cstart+1;
        Result := FALSE;
      end;
    end else
    begin {look for start comment}
      if not inquote then
      begin
        if (s[i]='/') and (s[i+1]='*') then
        begin
          Result := TRUE;
          cstart := i;
          Inc(i);
        end else
        if (s[i] = '''') or (s[i] = '"') then
        begin
          inquote := TRUE;
          quotechar := s[i];
        end;
      end else
      begin
        if inquote and (quotechar = s[i]) then
          inquote := FALSE;
      end;
    end;
    Inc(i);
  end;
  if Result then
    sub_chopcomment(s,cstart,Length(s));
end;

initialization
  InitDuparse;
end.







