{
-	RnnLib  Components
-
- RnnRegSc
-
- Freeware
- michrynn@ozemail.com.au
-
- Registry export file (*.reg) interpreter functions
- for TJvRnnCustomScript
}
unit JvRnnRegSc;

interface           
uses
  Windows, Registry, JvRnnScCst, Classes, SysUtils;
type
  ERegScript = class(Exception);
  PWriteRegDataProc = ^TWriteRegDataProc;
  TWriteRegDataProc = procedure(
     reg : TRegistry; const keyname, regvalue : string);

  PReadRegDataProc = ^TReadRegDataProc;
  TReadRegDataProc = procedure(
     reg : TRegistry; const keyname : string; var regvalue : string);

const
  kRnnSetRegistry = 'SET_REGISTRY';
  kRnnGetRegistry = 'GET_REGISTRY';
  kRnnRegistryBlock = 'REGISTRY_BLOCK';
  kEND_BLOCK = 'END_BLOCK';


resourcestring
  kHKEY_CLASSES_ROOT = 'HKEY_CLASSES_ROOT';
  kHKEY_CURRENT_USER = 'HKEY_CURRENT_USER';
  kHKEY_LOCAL_MACHINE = 'HKEY_LOCAL_MACHINE';
  kHKEY_USERS = 'HKEY_USERS';
  //kKEY_CURRENT_CONFIG = 'KEY_CURRENT_CONFIG';
  kHKEY_DYN_DATA = 'HKEY_DYN_DATA';



procedure RegisterRegConverter(
  const cvname:string; pr : PReadRegDataProc; pw : PWriteRegDataProc);

function ExtractRegRootPath(var root : HKEY; var path : string) : boolean;

procedure InstallRnnRegSc(scriptor:TJvRnnCustomScript);

resourcestring
  kSTRING = 'STRING';
  kDWORD = 'DWORD';
  kHEX = 'HEX';


implementation
Uses JvRnnPars, JvRnnStrUt, JvCCArray, JvCCRecrd;

type
  TJvRnnRegConverter = packed record
    FConvName : string[31];
    FConvRead : PReadRegDataProc;
    FConvWrite : PWriteRegDataProc;
  end;

const
  kMaxRegConv = (MAXINT div sizeof(TJvRnnRegConverter))-1;
type
  PRnnRegConvBuf = ^TJvRnnRegConvBuf;
  TJvRnnRegConvBuf = array[0..kMaxRegConv] of TJvRnnRegConverter;

  TJvRnnRegConvMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function  CompareItems(const p1, p2):integer;override;
    procedure AssignItem(var dest; const src); override;
  end;

  TJvRnnRegConvArray = class(TJvCCArray)
  private
    FArrayPtr : PRnnRegConvBuf;
    function GetRegConv(const cvname:string) : TJvRnnRegConverter;
  public
    constructor create;
    property ArrayPtr :  PRnnRegConvBuf read FArrayPtr;
  end;

var
  gRegConvArray : TJvRnnRegConvArray;

procedure RegWConvString(
  reg : TRegistry; const keyname, regvalue : string); far; forward;
procedure RegRConvString(
  reg : TRegistry; const keyname:string; var regvalue : string); far; forward;
procedure RegWConvWord(
  reg : TRegistry; const keyname, regvalue : string); far; forward;
procedure RegRConvWord(
  reg : TRegistry; const keyname:string; var regvalue : string); far; forward;
procedure RegWConvHex(
  reg : TRegistry; const keyname, regvalue : string); far; forward;
procedure RegRConvHex(
  reg : TRegistry; const keyname:string; var regvalue : string); far; forward;


procedure WriteRegFormat(
  reg : TRegistry;
  const keyname, regvalue, regformat : string); forward;

procedure SetRegistry_F(data:Pointer); far; forward;
{@SET_REGISTRY <Registry key name>  <registry value> [ as (string, dword,  ] }

procedure RegistryBlock_F(data:Pointer); far; forward;
{@SET_REGISTRY_BLOCK
[RegistryExportFormat1]
"Keyname1"=format1:value1
"Keyname2"=format2:value2
@END_REGISTRY_BLOCK;}
procedure EndBlock_F(data:Pointer); far; forward;
procedure GetRegistry_F(data:Pointer); far; forward;
{@GET_REGISTRY <variable name>  <Registry key name>
    [ as (string, dword,  ] }

const
  kErrKeyExpected = 'registry key expected';
  kErrValueExpected = 'registry value expected';
  kErrVariableExpected = 'variable name expected';
type
  TRegBlock = class(TJvRnnScrObj)
    FReg : TRegistry;
    FScriptor : TJvRnnCustomScript;
    FOldStd : TJvCmdProcessOp;
    FOldTerminator : Char;

    destructor Destroy; override;

    procedure ExecuteStd(Sender : TJvRnnAbstractScript; cmd : TStrings);
    procedure AfterPopState(var rst:TJvRnnScriptState); override;

  end;


destructor TRegBlock.Destroy;
begin
  FReg.Free;
  inherited Destroy;
end;

procedure TRegBlock.ExecuteStd(Sender : TJvRnnAbstractScript; cmd : TStrings);
{expecting a block of registry "INI" type formats. cmd should have one line only}
{look for [ ] delimiting a line to start a new registry key name}
var
  s : string;
  rname : string;
  rvalue : string;
  rformat : string;
  toktype : TJvRnnTokenType;
  token : string;
  ptype : TJvRnnParenType;
  p : PChar;
  keyroot : HKEY;

  procedure ThrowError(const inerr : string);
  begin
    raise EJvRnnScript.create('REGISTER_BLOCK : ' + inerr);
  end;

  function NextToken : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,toktype);
  end;

begin
  s := cmd[0];
  if length(s) = 0 then
    exit;

  if (s[1] = '[') then
  begin
    if (s[Length(s)] <> ']') then
      ThrowError('missing "]"');
    s := Copy(s,2,Length(s)-2);
    if (Length(s) > 0) then
    begin
      if not ExtractRegRootPath(keyroot, s) then
        keyroot := HKEY_CURRENT_USER;

      if SELF.FReg = nil then
        FReg := TRegistry.Create
      else begin
        FReg.LazyWrite := FALSE;
        FReg.CloseKey;
        FReg.LazyWrite := TRUE;
      end;
      FReg.RootKey := keyroot;
      FReg.OpenKey(s,TRUE);
    end else
    begin
      ThrowError('Empty Key');
    end;
  end else
  begin
    {decode the key name and its type : value}
    p := PChar(s);
    
    if NextToken then
      rname := token;

    if NextToken then
      if toktype <> tok_equal then
        ThrowError('"=" expected');
    if not NextToken then
      ThrowError('Value missing');
    if (ptype = paren_dquote) then
    begin
      rformat := kSTRING;
      rvalue := token;
    end else
    begin
      rformat := token;
      if (not NextToken) or (token[1] <> ':') then
          ThrowError('":" expected'); {must be a ':'}
      rvalue := StrPas(p);  {no de-quote if string ?}
    end;
    WriteRegFormat(FReg, rname, rvalue, rformat);
  end;

end;

procedure TRegBlock.AfterPopState(var rst:TJvRnnScriptState);
var
  sstate : TJvRnnScriptState;
begin
  FScriptor.Terminator := SELF.FOldTerminator;
  FScriptor.OnExecuteStd := SELF.FOldStd;
  
  sstate := FScriptor.GetScriptState;
  sstate.lineNum := rst.lineNum;
  sstate.linechar := rst.linechar;
  FScriptor.SetScriptState(sstate);
  
  inherited AfterPopState(rst);
end;

procedure InstallRnnRegSc(scriptor:TJvRnnCustomScript);
begin
  with scriptor do begin
    SetUserOp(kRnnSetRegistry,@SetRegistry_F);
    SetUserOp(kEND_BLOCK, @EndBlock_F);
    SetUserOp(kRnnRegistryBlock, @RegistryBlock_F);
    SetUserOp(kRnnGetRegistry,@GetRegistry_F);
  end;
end;

function ExtractRegRootPath(var root : HKEY; var path : string) : boolean;
var
   subpath : string;
   rootkey : string;
   ip : integer;
begin
{if the path contains a root specification, remove it from the path string
Return true if a root specifier was found}

  Result := FALSE;

  ip := Pos('\',path);
  if (ip > 0) then
  begin
    rootkey := Copy(path,1,ip-1);
    subpath := Copy(path,ip+1, Length(path)-ip);
  end else
    exit;

  if COMPARETEXT(kHKEY_CLASSES_ROOT,rootkey) = 0 then
  begin
    root := HKEY_CLASSES_ROOT;
    Result := TRUE;
  end else
  if COMPARETEXT(kHKEY_CURRENT_USER,rootkey) = 0 then
  begin
    root := HKEY_CURRENT_USER;
    Result := TRUE;
  end else
  if COMPARETEXT(kHKEY_LOCAL_MACHINE,rootkey) = 0 then
  begin
    root := HKEY_LOCAL_MACHINE;
    Result := TRUE;
  end else
  if COMPARETEXT(kHKEY_USERS,rootkey) = 0 then
  begin
    root := HKEY_USERS;
    Result := TRUE;
  end else
 (* if COMPARETEXT(kKEY_CURRENT_CONFIG,rootkey) = 0 then
  begin
    root := KEY_CURRENT_CONFIG;
    Result := TRUE;
  end else *)
  if COMPARETEXT(kHKEY_DYN_DATA,rootkey) = 0 then
  begin
    root := HKEY_DYN_DATA;
    Result := TRUE;
  end;
  if Result then
    path := subpath;
end;

procedure RegistryBlock_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  block : TRegBlock;
  state : TJvRnnScriptState;
begin
{create a TRegBlock object, which will change the terminator to CR,
 and replace the ExecuteStd event handler, until popped off by END_BLOCK}
  cmdInfo := PJvScriptCmdInfo(data);
  block := TRegBlock.Create;
  with block do
  begin
    FScriptor := cmdInfo^.scriptor;
    FOldStd := FScriptor.OnExecuteStd;
    FOldTerminator := FScriptor.StdTerminator;
    FScriptor.StdTerminator := #0;
    FScriptor.OnExecuteStd := block.ExecuteStd;
    state := FScriptor.GetScriptState; {get copy of script state}
    state.refptr := block;
    FScriptor.PushScriptState(state);
  end;
end;


procedure EndBlock_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  scriptor.PopScriptState;
end;

procedure WriteRegFormat(
  reg : TRegistry;
  const keyname, regvalue, regformat : string);
var
  rc : TJvRnnRegConverter;
begin
  rc := gRegConvArray.GetRegConv(regformat);
  TWriteRegDataProc(rc.FConvWrite)(reg,keyname,regvalue);
end;



procedure SetRegistry_F(data:Pointer);
var
  regkey:string;
  path : string;
  keyname : string;
  regvalue : string;
  cmdInfo : PJvScriptCmdInfo;
  xtoken : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  p : PChar;
  reg : TRegistry;
  datatype : string;
  useRoot : HKEY;
  rc : TJvRnnRegConverter;

  function sub_next(var token : string) : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,tokentype);
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);

  p := cmdInfo^.cmdcstr;

  if not sub_next(regkey) then
    raise EJvRnnScript.create(kErrKeyExpected);

  if not sub_next(regvalue) then
    raise EJvRnnScript.create(kErrValueExpected);

  if tokentype = tok_equal then
    if not sub_next(regvalue) then
      raise EJvRnnScript.create(kErrValueExpected);

  datatype := kSTRING;
  if sub_next(xtoken) then
  begin
    if COMPARETEXT(xtoken,'AS')=0 then
    begin
      if not sub_next(xtoken) then
        raise EJvRnnScript.create(' as <datatype> ');
      datatype := xtoken;
    end;
  end;
  path := ExtractFilePath(regkey);
  RemoveEndChar(path,'\');
  keyname := ExtractFileName(regkey);

  reg := TRegistry.Create;
  if ExtractRegRootPath(useRoot,path) then
    reg.RootKey := useRoot;

  try
    if reg.OpenKey(path,TRUE) then
    begin
      rc := gRegConvArray.GetRegConv(datatype);
      TWriteRegDataProc(rc.FConvWrite)(reg,keyname,regvalue);
    end;
  finally
    reg.free;
  end;
  end;

procedure GetRegistry_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  xtoken : string;
  varname : string;
  regkey : string;
  path : string;
  keyname : string;
  keyvalue : string;
  datatype : string;
  reg : TRegistry;
  ptype : TJvRnnParenType;
  toktype : TJvRnnTokenType;
  useRoot : HKEY;
  keyInfo : TRegDataInfo;
  rc : TJvRnnRegConverter;

  function sub_next(var token : string) : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,toktype);
  end;

  procedure ThrowError(const errorstr : string);
  begin
    raise EJvRnnScript.Create('GET_REGISTRY : '+errorstr);
  end;

begin
{attempt to read the entry and store it as a string in a variable}
  cmdInfo := PJvScriptCmdInfo(data);

  p := cmdInfo^.cmdcstr;
  if not sub_next(varname) then
    ThrowError( kErrVariableExpected );

  if not sub_next(regkey) then
    ThrowError(kErrKeyExpected);

  path := ExtractFilePath(regkey);
  RemoveEndChar(path,'\');
  keyname := ExtractFileName(regkey);

  datatype := kSTRING;
  if sub_next(xtoken) then
  begin
    if COMPARETEXT(xtoken,'AS')=0 then
    begin
      if not sub_next(xtoken) then
        ThrowError(' as <datatype> ');
      datatype := xtoken;
    end;
  end;

  reg := TRegistry.Create;
  TRY
    if ExtractRegRootPath(useRoot,path) then
      reg.RootKey := useRoot;
    if reg.OpenKey(path,FALSE) then
    begin
      if reg.GetDataInfo(keyname,keyInfo) then
      begin
        CASE keyInfo.RegData of
        rdString,
        rdExpandString :
          datatype := kSTRING;
        rdInteger :
          datatype := kDWORD;
        END; {case1}
      end;
      rc := gRegConvArray.GetRegConv(datatype);
      TReadRegDataProc(rc.FConvRead)(reg,keyname,keyvalue);
      cmdInfo^.scriptor.SetVariable(varname,keyvalue);
    end else
      raise ERegScript.create('Registry key not found: '+path);
  FINALLY
    reg.Free;
  END;
end;


procedure RegWConvString(
  reg : TRegistry; const keyname, regvalue : string);
begin
  reg.writestring(keyname,regvalue);
end;

procedure RegRConvString(
  reg : TRegistry; const keyname:string; var regvalue : string);
begin
  regvalue := reg.ReadString(keyname);
end;

procedure RegWConvWord(
  reg : TRegistry; const keyname, regvalue : string);
begin
  reg.writeInteger(keyname,HexStrToInt(regvalue));
end;

procedure RegRConvWord(
  reg : TRegistry; const keyname:string; var regvalue : string);
begin
  regvalue := IntToStr(reg.ReadInteger(keyname));
end;

procedure RegWConvHex(
  reg : TRegistry; const keyname, regvalue : string);
var
  memptr : Pointer;
  memsize : integer;
begin
  memptr := nil;
  memsize := 0;
  HexToBinary(PChar(regvalue),memptr,memsize);
  try
    if memsize > 0 then
      reg.WriteBinaryData(keyname,memptr^,memsize);
  finally
    if memsize > 0 then
      FreeMem(memptr, memsize);
  end;
end;

procedure RegRConvHex(reg : TRegistry; const keyname : string; var regvalue : string);
begin
  raise ERegScript.Create('Hex read not implemented');
end;

constructor TJvRnnRegConvArray.create;
begin
  inherited CreateMgr(TJvRnnRegConvMgr);
  BlockHdl := @FArrayPtr;
  Count := 0;
end;

function  TJvRnnRegConvMgr.CompareItems(const p1, p2):integer;
begin
  Result := COMPARETEXT(TJvRnnRegConverter(p1).FConvName,
    TJvRnnRegConverter(p2).FConvName);
end;

constructor TJvRnnRegConvMgr.create;
begin
  inherited create;
  InitRecordMgr(sizeof(TJvRnnRegConverter), TRUE);
end;

procedure TJvRnnRegConvMgr.AssignItem(var dest; const src);
begin
  TJvRnnRegConverter(dest).FConvName := TJvRnnRegConverter(src).FConvName;
  TJvRnnRegConverter(dest).FConvRead := TJvRnnRegConverter(src).FConvRead;
  TJvRnnRegConverter(dest).FConvWrite := TJvRnnRegConverter(src).FConvWrite;
end;


function TJvRnnRegConvArray.GetRegConv(const cvname:string): TJvRnnRegConverter;
var
  ix : integer;
  rc : TJvRnnRegConverter;
begin
  rc.FConvName := cvname;
  rc.FConvRead := nil;
  rc.FConvWrite := nil;
  if gRegConvArray.GetItemIndex(ix, rc) then
  begin {replace}
    gRegConvArray.GetItem(ix, Result);
  end else
  begin
    raise ERegScript.Create('No registry type '+cvname);
  end;
end;


procedure RegisterRegConverter(
  const cvname:string; pr : PReadRegDataProc; pw : PWriteRegDataProc);
var
  rc : TJvRnnRegConverter;
  ix : integer;
begin
  rc.FConvName := cvname;
  rc.FConvRead := pr;
  rc.FConvWrite := pw;
  if gRegConvArray.GetItemIndex(ix, rc) then
  begin {replace}
    gRegConvArray.SetItem(ix, rc);
  end else
  begin
    gRegConvArray.InsertInto(ix, rc);
  end;
end;

procedure InitRegConverters;
begin
  gRegConvArray:= TJvRnnRegConvArray.create;
  RegisterRegConverter(kSTRING,@RegRConvString,@RegWConvString);
  RegisterRegConverter(kDWORD,@RegRConvWord,@RegWConvWord);
  RegisterRegConverter(kHEX,@RegRConvHex,@RegWConvHex);
end;

initialization
  InitRegConverters;
finalization
  gRegConvArray.Free;
end.
