{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnScCst.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnScCst;

interface
Uses Classes,SysUtils,WinProcs,WinTypes,
  JvRnnStrLs, JvCCArray,JvRnnVarCp,JvRnnScrSt,
  JvCCRecrd, JvRnnPars;
const
  TJVRNNGENSCRIPT_VERSION = 401;

  kRnnDefaultDateFormat = 'dd-mmm-yyyy';
  kRnnDefaultTimeFormat = 'hh:nn:ss';
  kRnnDefaultDateTimeFormat = 'dd-mmm-yyyy hh:nn:ss';
type
  TJvRnnScrObj = class;
  TJvRnnScriptState = record
    script : TStrings;
    lineNum : integer;
    linechar : integer;
    scriptlabel : string[31];
    ownerid : string[31];	{owner of current iterator, if any}
    refptr  : TJvRnnScrObj;	{extra storage if any}
  end;
  PJvScriptState = ^TJvRnnScriptState;

  TJvRnnScrObj = class(TObject)
    procedure AfterPopState(var rst:TJvRnnScriptState); virtual;
  end;


  TJvRnnScriptStateMgr = class(TJvCCRecordMgr)
    constructor Create; override;
  end;

const
  kMaxScriptStateCt = (MAXINT div sizeof(TJvRnnScriptState)) - 1;

type
  TJvRnnAbstractScript = class;
  TJvRnnCustomScript = class;

  EJvRnnScript = class(Exception);

  TJvLogMsgEvent = procedure(Sender:TObject;
    const msg:string) of object;

  TJvCmdExtractOp = procedure( Sender : TJvRnnAbstractScript;
    var lineNum, linechar :integer; script,cmd : TStrings) of object;

  TJvCmdProcessOp = procedure(Sender : TJvRnnAbstractScript;
    cmd : TStrings) of object;

  PJvScriptCmdInfo = ^TJvScriptCmdInfo;
  TJvRegUserOp = procedure (cmdInfo:PJvScriptCmdInfo);
  PJvRegUserOp = ^TJvRegUserOp;

  TJvRegUserFunc = function (sct:TJvRnnCustomScript; args:TStrings) : string;
  PJvRegUserFunc = ^TJvRegUserFunc;

  TJvRnnScriptStateArray = array[0..kMaxScriptStateCt] of TJvRnnScriptState;
  PJvScriptStateArray = ^TJvRnnScriptStateArray;


  TJvRnnScrStack = class(TJvCCArray)
  private
    FArrayPtr : PJvScriptStateArray;
  public
    constructor Create(defsize : integer);
    procedure PushState(const ss : TJvRnnScriptState);
    function  PopState : TJvRnnScriptState;
    property  ArrayPtr : PJvScriptStateArray read FArrayPtr;
  end;

  TJvRnnScriptStep = class;


  TJvRnnAbstractScript = class(TComponent)
  private
    FCmdExtract : TJvCmdExtractOp;
    FCmdProcess : TJvCmdProcessOp;
    FBeforeCmdExecute : TJvCmdProcessOp;
    FCmdExecute : TJvCmdProcessOp;
    FExecuteStd : TJvCmdProcessOp;
    FLogMsgEvent : TJvLogMsgEvent;
    FLogErrorEvent : TJvLogMsgEvent;
    FLogStrings : TJvLogMsgEvent;
    {script processing}
  {$IFDEF EXP_LOCK}
    FLock: TRTLCriticalSection; {experimental}
  {$ENDIF}
  protected
    FStack :  TJvRnnScrStack;
    FStackIX : integer;             {-1 means empty stack}
    FScriptState : TJvRnnScriptState; {working top of stack}

    function  GetIsRunning : boolean;
    procedure DefaultCmdExtract(
       var lineNum, linechar :integer;
       script, cmd : TStrings); virtual; abstract;
    procedure DefaultCmdProcess(cmd:TStrings); virtual; abstract;
    procedure DefaultCmdExecute(cmd:TStrings); virtual; abstract;
    procedure ExecuteScriptFunction(cmdInfo:Pointer); virtual; abstract;
    procedure ResetStack;


  public
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy ; override;
    procedure LockScript;
    procedure UnlockScript;
    procedure LogMsg(const inMsg:string); virtual;
    procedure LogError(const inMsg:string); virtual;
    function  GetScriptState : TJvRnnScriptState;
    procedure SetScriptState (const state : TJvRnnScriptState);
    procedure PushScriptState(const state : TJvRnnScriptState);
    function  PopScriptState : TJvRnnScriptState;
    procedure Execute(startscript : TStrings); virtual;
    procedure ExecuteFile(const path : string);
    procedure ExecuteLine(const s : string);
  published
    property Running : boolean read GetIsRunning;
    property CmdExtract : TJvCmdExtractOp read FCmdExtract write FCmdExtract;
    property CmdProcess : TJvCmdProcessOp read FCmdProcess write FCmdProcess;
    property CmdExecute : TJvCmdProcessOp read FCmdExecute write FCmdExecute;
    property OnExecuteStd : TJvCmdProcessOp read FExecuteStd write FExecuteStd;
    property OnLogMsg : TJvLogMsgEvent read FLogMsgEvent write FLogMsgEvent;
    property BeforeCmdExecute : TJvCmdProcessOp read FBeforeCmdExecute write FBeforeCmdExecute;
    property OnLogError : TJvLogMsgEvent read FLogErrorEvent write FLogErrorEvent;
    property OnLogStrings : TJvLogMsgEvent read FLogStrings write FLogStrings;
  end;



  TJvScriptRunState = (srs_extract, srs_process, srs_execute, srs_finished, srs_error);

  TJvRnnScriptStep = class(TObject)
  private
    FRunState : TJvScriptRunState;
    FScript : TStrings; {starting script}
    FCmd : TStrings; {current command}
    FScriptor : TJvRnnAbstractScript;
    FNoExecute : boolean;
  public
    constructor create(scriptor:TJvRnnCustomScript; script : TStrings);
    destructor destroy;override;
    procedure ExecuteStep;
    procedure ExecuteNext;
    procedure ExecuteAll;
    procedure Reset;
    procedure Fetch;
    procedure Process;
    procedure Execute;
    function GetLineNum : integer;
    function GetLineChar : integer;
    function getStrings : TStrings;
    function IsFinished : boolean;
    property State : TJvScriptRunState read FRunState;
    property Line : integer read GetLineNum;
    property LineChar : integer read GetLineChar;
    property Command : TStrings read GetStrings;
    property NoExecute : boolean read FNoExecute write FNoExecute;
  end;

  TJvScriptCmdInfo = record
    scriptor : TJvRnnCustomScript;
    cmdstrings : TStrings; {current command}
    cmdname : PChar; {pointer to command name}
    cmdcstr : PChar; {command as null-terminated string}
  end;


  TJvRnnCustomScript = class(TJvRnnAbstractScript)
  private
    FParamChar : Char;
    FPrefixChar : Char;
    FTerminator : Char;
    FStdTerminator : Char;
  protected
    FVarComp : TJvRnnValueList;
    FVariables : TJvCCStrStrArray;
    FScriptStore : TJvRnnScriptStore;
    FUserOps : TStringList;
    FUserFuncs : TStringList;
    FExprRecursion : integer;
    FDeleteComments : boolean;

    FDateFormat : string;
    FTimeFormat : string;
    FDateTimeFormat : string;

    procedure Notification(
      AComponent: TComponent; Operation: TOperation); override;
    procedure SetVarComp(valList : TJvRnnValueList);
    {script}
    procedure DefaultCmdExtract(var lineNum, linechar :integer;
                      script,cmd : TStrings); override;

    procedure DefaultCmdProcess(cmd:TStrings); override;
    procedure DefaultCmdExecute(cmd:TStrings); override;
    procedure ExecuteScriptFunction(cmdInfo:Pointer); override;

    procedure ExecuteStd(cmd : TStrings); virtual;
    function  GetVariables : TJvCCStrStrArray;
    procedure CheckHasStore;
    function  GetScriptNoFail(const sname : string) : TStrings;
  public
    constructor Create(AOwner : TComponent); override;

    {script}
    procedure ExecuteName(const scriptName : string);
    procedure ExecuteNameParams(const scriptName : string; params : TStrings);
    procedure SubstituteParams(var s : string);
    function  EvaluateName(var ps : TParseState) : string; virtual;
    function  GetExpressionValue(var cp : PChar) : string; virtual;
    function  ParseFunctionArgs(var cp : PChar) : TStrings; virtual;

    {variables}
    procedure SetVariable(const varname, value:string);
    procedure KillVariable(const varname : string);
    function  GetValue(const varname : string) : string;

    {user operations and functions}
    procedure SetUserOp(const opname : string; opProc: PJvRegUserOp);
    procedure DeleteUserOp(const opname:string);
    function  GetUserOp(const opname:string) : PJvRegUserOp;

    procedure SetUserFunc(const opname : string; opProc: PJvRegUserFunc);
    procedure DeleteUserFunc(const opname:string);
    function  GetUserFunc(const opname:string) : PJvRegUserFunc;

    {properties}
    property DateFormat : string read FDateFormat write FDateFormat;
    property TimeFormat : string read FTimeFormat write FTimeFormat;
    property DateTimeFormat : string read FDateTimeFormat write FDateTimeFormat;

    property ParamChar : Char read FParamChar write FParamChar;
    property PrefixChar : Char read FPrefixChar write FPrefixChar;
    property Terminator : Char read FTerminator write FTerminator;
    property DeleteComments : Boolean
      read FDeleteComments write FDeleteComments default TRUE;
    property StdTerminator : Char read FStdTerminator write FStdTerminator;
    property Script[const sname : string] : TStrings read GetScriptNoFail;
    property Values[const varname : string] : string
      read GetValue write SetVariable;
    property ValueList : TJvRnnValueList read FVarComp write SetVarComp;
    property Variables : TJvCCStrStrArray read GetVariables;
    property ScriptStore : TJvRnnScriptStore read FScriptStore write FScriptStore;
  end;

  TJvRnnGenScript = class(TJvRnnCustomScript)
  published
    property CmdExtract;
    property CmdProcess;
    property CmdExecute;
    property OnExecuteStd;

    property OnLogMsg;
    property BeforeCmdExecute;
    property OnLogError;
    property OnLogStrings;

    property ParamChar;
    property PrefixChar;
    property DeleteComments;
    property Terminator;
    property StdTerminator;
    property ScriptStore;
    property ValueList;
  end;



procedure InitializeScriptState(var sst : TJvRnnScriptState);
procedure FinalizeScriptState(var sst : TJvRnnScriptState);
{--------------------- function evaluation ------------------------------------}

function ParseCommaList(var p : PChar; allowempty:boolean) :TStrings;
function ParseLines(var p : PChar; allowempty:boolean) : TStrings;
{---------------------- functions ---------------------------------------------}

function GetSysDate_F(sct: TJvRnnCustomScript; args : TStrings): string;
function GetSysTime_F(sct: TJvRnnCustomScript; args : TStrings): string;
function GetSysDateTime_F(sct: TJvRnnCustomScript; args : TStrings): string;
function Quote_F(sct: TJvRnnCustomScript; args : TStrings): string;
function Concat_F(sct: TJvRnnCustomScript; args : TStrings): string;
function CountArgs_F(sct: TJvRnnCustomScript; args : TStrings): string;
procedure InstallCustomScriptOps(scr : TJvRnnCustomScript);

const
  kFINALLY = 'FINALLY';
  kTRY =     'TRY';
  kENDTRY =  'ENDTRY';

implementation
Uses Forms,JvRnnStrUt,Dialogs;
{$IFNDEF WIN32}
{$R *.R16}
{$ENDIF}
const
  kMaxExprRecursion = 10;
  kExpressionTooComplex = 'Expression too complex';
{--------------------- script variables ---------------------------------------}

procedure SetParamChar_F(data:Pointer); far; forward;
{@SET_PARAM_CHAR <char> ;}

procedure KillVariable_F(data:Pointer); far; forward;
{@UNSET <Variable Name> <Value> ; }

procedure SetVariable_F(data:Pointer); far; forward;
{@SET <Variable Name> [=] <Value>|<Variable Name> | <Function name> ; }

procedure SetDateFormat_F(data:Pointer); far; forward;
{------------------------ execution control  ----------------------------------}

procedure IF_F(data:Pointer);far;forward;
{@IF <TRUE/FALSE> ; <statement>;[statement;] [ @ENDIF; }
procedure Else_F(data:Pointer); far; forward;
{@ELSE;}
procedure ENDIF_F(data:Pointer);far;forward;
{@ENDIF;}

procedure TRY_F(data:Pointer); far;forward;
{@TRY;}

procedure FINALLY_F(data:Pointer); far; forward;
{@FINALLY;}

procedure ENDTRY_F(data:Pointer); far; forward;
{@ENDTRY;}

procedure LOGMSG_F(data:Pointer);far;forward;
{@LOG_MSG <Message text>;}

procedure LOGERROR_F(data:Pointer);far;forward;
{@LOG_ERROR <Message text>;}

procedure Execute_F(data:Pointer); far; forward;
{@EXECUTE <script label>}

procedure Abort_F(data:Pointer); far; forward;
{@ABORT <text>;}



procedure InstallCustomScriptOps(scr : TJvRnnCustomScript);
begin
  with scr do
  begin
    SetUserOp('IF', @IF_F);
    SetUserOp('ELSE',@Else_F);
    SetUserOp('ENDIF', @ENDIF_F);
    SetUserOp(kTRY, @TRY_F);
    SetUserOp(kFINALLY, @FINALLY_F);
    SetUserOp(kENDTRY, @ENDTRY_F);

    SetUserOp('LOG_MSG',@LogMsg_F);
    SetUserOp('LOG_ERROR',@LogError_F);
    SetUserOp('EXECUTE', @Execute_F);
    SetUserOp('ABORT',@Abort_F);
    
    SetUserOp('SET', @SetVariable_F);
    SetUserOp('UNSET', @KillVariable_F);
    SetUserOp('SET_PARAM_CHAR',@SetParamChar_F);

    SetUserOp('SET_DATE_FORMAT',@SetDateFormat_F);
    SetUserOp('SET_TIME_FORMAT',@SetDateFormat_F);
    SetUserOp('SET_DATETIME_FORMAT',@SetDateFormat_F);

    SetUserFunc('SYSDATE',@GetSysDate_F);
    SetUserFunc('SYSTIME',@GetSysTime_F);
    SetUserFunc('SYSDATETIME',@GetSysDateTime_F);
    SetUserFunc('CONCAT',@Concat_F);
    SetUserFunc('QUOTE',@Quote_F);
    SetUserFunc('COUNT_ARG',@CountArgs_F);
  end;
end;
{built in functions for TRunSQLScript}


function TJvRnnCustomScript.GetExpressionValue(var cp:PChar) : string;
var
  ps : TParseState;
  func : PJvRegUserFunc;
  arglist : TStrings;
begin
  ps_init(ps,cp);
  if (not ps_nexttokentype(ps)) then
    raise ERnnScript.create('Expression expected');
  if ps.toktype = tok_quotetext then
  begin
    Result := ps.token;
    cp := ps.p;
    exit;
  end;
  Inc(FExprRecursion);
  If (FExprRecursion > kMaxExprRecursion) then
    raise ERnnScript.Create(kExpressionTooComplex);
  TRY
  if (ps.toktype = tok_name) then
  begin {token must be a variable,function or value}
    if (ps.p^ = '(') then
    begin  { a function is expected }
      func := GetUserFunc(ps.token);
      if func = nil then
        raise ERnnScript.create('Function '+ps.token+' not found');
      Inc(ps.p);
      arglist := ParseFunctionArgs(ps.p);
      try
        Result := TJvRegUserFunc(func)(SELF,arglist);
      finally
        arglist.free;
      end;
    end else
    begin
      Result := EvaluateName(ps);
    end
  end else
    Result := ps.token;
  cp := ps.p;
  FINALLY
    Dec(FExprRecursion);
  END;
end;

function ParseLines(var p : PChar; allowempty:boolean) : TStrings;
var
  s : string;
begin
  Result := nil;
  while (p^ <> #0) do
  begin
    StrEndLine(p, s);

    if (allowempty) or (Length(s) > 0) then
    begin
      if (Result = nil) then
        Result := TStringList.Create;
      Result.Add(s);
    end;
  end;
end;

function ParseCommaList(var p : PChar; allowempty:boolean) :TStrings;
var
  ct : integer;
  token : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  isComma : boolean;
  lastWasComma : boolean;

  function nextarg : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,tokentype);
  end;
  { a '(' has already been found, indicating a list of comma
    separated parameters }
begin
  Result := nil;
  ct := 0;
  isComma := FALSE;
  if p^ = #0 then
    exit;
  while (true) do
  begin
    if not nextarg then
      exit;
    if token[1] = ')' then
      exit;
    inc(ct);
    if ct=1 then
    begin
      Result := TStringList.Create;
    end;
    lastWasComma := isComma;
    isComma := (token[1] = ',');
    if (not isComma) OR (isComma and lastWasComma) then
    begin
      if (isComma) then
      begin
        if allowempty then
          Result.Add('')
        else
          raise ERnnScript.create('parameter expected')
      end
      else
        Result.Add(token);
    end;
  end;
end;

{ParseFunctionArgs:
  The first character must not be '(', but this eats up last ')'}
function TJvRnnCustomScript.ParseFunctionArgs(var cp : PChar) :TStrings;
var
  i : integer;
  n : integer;
  ct : integer;
  ps : TParseState;
  isComma : boolean;
  lastWasEquals : boolean;
  isEquals : boolean;
  lastWasComma : boolean;
  isBrackets : boolean;
  s : string;
  savep : PChar;
  tokenp : PChar;
  hasContent : boolean;

  function nextarg : boolean;
  begin
    savep := ps.p;
    Result := ps_nexttokentype(ps);
    hasContent := (Length(ps.token) > 0);
  end;

begin
  Result := nil;
  ct := 0;
  isComma := FALSE;
  isEquals := FALSE;
  if cp^ = #0 then
    exit;
  ps_init(ps,cp);
  tokenp := nil;
  Inc(FExprRecursion);
  If (FExprRecursion > kMaxExprRecursion) then
    raise ERnnScript.Create(kExpressionTooComplex);
  TRY
    while (true) do
    begin
      if (ps.p^ = '(') and (FExprRecursion=1) then
        Inc(ps.p);
      if (not nextarg) OR
         ((ps.partype=paren_round) and (ps.token[1] = ')')) then
      begin
        cp^ := ps.p^;
        exit;
      end;
      inc(ct);
      if ct=1 then
      begin
        Result := TStringList.Create;
      end;
      lastWasComma := isComma;
      lastWasEquals := isEquals;
      isBrackets := hasContent and (ps.token[1] = '(');
      if isBrackets then
      begin
        {recursion count?}
        if (tokenp <> nil) then
        begin
          {replace current token}
          i := Result.Count-1;
          s := Result[i];
          ps.token := GetExpressionValue(tokenp);
          n := Length(s);
          if (n > 0) and (s[n]='=') then
            s := s + ps.token
          else
            s := ps.token;
          Result[i] := ps.token;
          continue;
        end
        else
          raise ERnnScript.create('invalid expression');
      end;
      isComma := hasContent and (ps.token[1] = ',');
      isEquals := hasContent and (ps.token[1] = '=');
      if (not isComma) OR (isComma and lastWasComma) then
      begin
        if not isComma then
        begin
          if (isEquals or lastWasEquals) and (ct > 1) then
          begin
            i := Result.Count-1;
            s := Result[i];
            if lastWasEquals then
              s := s + ps.token
            else
              s := s + '=';
            Result[i] := s;
          end else
          begin
            tokenp := savep;
            (*if ps.toktype = tok_name then
            begin {try variable substitution}
              n := varray.IndexOf(ps.token);
              if (n >= 0) then
                ps.token := varray.GetIndexValue(n);
            end; *)
            Result.Add(ps.token);
          end;
        end
        else
          Result.Add('');
      end;
      cp := ps.p;
    end;
  FINALLY
    Dec(FExprRecursion);
  END;
end;

procedure ThrowNoVariables;
begin
  raise ERnnScript.create('No variables component');
end;

procedure InitializeScriptState(var sst : TJvRnnScriptState);
begin
  with sst do
  begin
    script := nil;
    lineNum := 0;
    linechar := 0;
    scriptlabel := '';
    ownerid := '';
    refptr := nil;
  end;
end;

procedure FinalizeScriptState(var sst : TJvRnnScriptState);
begin
  with sst do
  begin
    if refptr <> nil then
    begin
      refptr.free;
      refptr := nil;
    end;
  end;
end;

procedure TJvRnnScrObj.AfterPopState(var rst:TJvRnnScriptState);
begin
end;



constructor TJvRnnScriptStateMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TJvRnnScriptState),TRUE);
end;

constructor TJvRnnScrStack.Create(defsize : integer);
begin
  inherited CreateMgr(TJvRnnScriptStateMgr);
  BlockHdl := @FArrayPtr;
end;


procedure TJvRnnScrStack.PushState(const ss : TJvRnnScriptState);
var
  i : integer;
begin
  i := Count;
  SELF.CreateItemsAt(i,1);
  ArrayPtr^[i] := ss;
end;


function  TJvRnnScrStack.PopState : TJvRnnScriptState;
var
  i : integer;
begin
  i := Count-1;
  if ( i < 0) then
    raise ERnnScript.create('Pop of empty stack');
  Result := ArrayPtr^[i];
  Count := i;
end;

procedure TJvRnnAbstractScript.PushScriptState(const state : TJvRnnScriptState);
begin
  if (FStackIX >= 0) then
    FStack.PushState(FScriptState);
  Inc(FStackIX);
  FScriptState := state;
end;

function TJvRnnAbstractScript.PopScriptState:TJvRnnScriptState;
var
  scrobj : TJvRnnScrObj;
begin
  if FStackIX < 0 then
    raise ERnnScript.create('Pop of empty stack');

  Result := FScriptState; {get working top of stack}

  if (FStackIX > 0) then
    FScriptState := FStack.PopState; {new working top of stack}
  Dec(FStackIX);

  scrobj := Result.refptr as TJvRnnScrObj;

  if (scrobj <> nil) then
    scrobj.AfterPopState(Result);

  FinalizeScriptState(Result);
end;


procedure TJvRnnAbstractScript.LogMsg(const inMsg:string);
begin
  if Assigned(FLogMsgEvent) then
    FLogMsgEvent(SELF, inMsg);
end;


procedure TJvRnnAbstractScript.LogError(const inMsg:string);
begin
  if Assigned(FLogErrorEvent) then
    FLogErrorEvent(SELF, inMsg);
end;

constructor TJvRnnAbstractScript.Create(AOwner : TComponent);
begin
  {$IFDEF EXP_LOCK}
  InitializeCriticalSection(FLock);
  {$ENDIF}
  inherited Create(AOwner);
  FStackIX := -1;
  FStack := TJvRnnScrStack.Create(0);
end;

destructor  TJvRnnAbstractScript.Destroy;
begin
  {$IFDEF EXP_LOCK}
  DeleteCriticalSection(FLock);
  {$ENDIF}
  inherited Destroy;
end;

procedure TJvRnnAbstractScript.UnlockScript;
begin
  {$IFDEF EXP_LOCK}
  LeaveCriticalSection(FLock);
  {$ENDIF}
end;

procedure TJvRnnAbstractScript.LockScript;
begin
  {$IFDEF EXP_LOCK}
  EnterCriticalSection(FLock);
  {$ENDIF}
end;

procedure TJvRnnAbstractScript.SetScriptState (const state : TJvRnnScriptState);
begin
  FScriptState := state;
end;

procedure TJvRnnAbstractScript.ExecuteLine(const s : string);
var
  slist : TStrings;
begin
  slist := TStringList.Create;
  try
    slist.Add(s);
    Execute(slist);
  finally
    slist.free;
  end;
end;

function TJvRnnAbstractScript.GetScriptState : TJvRnnScriptState;
begin
  if FStackIX < 0 then
    raise ERnnScript.create('Stack empty');
  Result := FScriptState;
end;


procedure TJvRnnAbstractScript.ResetStack;
begin
  while FStackIX >= 0 do
    PopScriptState;
  FStack.Count := 0;
  FStackIX := -1;
end;

function  TJvRnnAbstractScript.GetIsRunning : boolean;
begin
  Result := (FStackIX > -1);
end;

procedure TJvRnnAbstractScript.ExecuteFile(const path : string);
var
  slist : TStrings;
begin
  slist := TStringList.Create;
  try
    slist.LoadFromFile(path);
    Execute(slist);
  finally
    slist.free;
  end;
end;

procedure TJvRnnAbstractScript.Execute(startscript : TStrings);
var
  cmd : TJvRnnStrList;
  ss : TJvRnnScriptState;
  stackct : integer;
  exceptct : integer;
  exmsg : string;

  procedure NexTJvCmdProcess; forward;

  procedure LoopLines;
  begin
    REPEAT
      if (exceptct > 1) OR
       ((exceptct > 0) and (FScriptState.ownerid <> kFINALLY)) then
      begin
        PopScriptState; 
      end
      else
      WITH FScriptState do
      TRY
        try
          while (lineNum < script.Count) do
            NexTJvCmdProcess;
        finally
          PopScriptState;
        end;
      EXCEPT
        on x:ERnnScript do
        begin
          Inc(exceptct);
          if (exceptct = 1) then
            exmsg := x.message;  // store initial exception msg
        end;
      END;
    UNTIL ((stackct=FStackIX)OR(FStackIX<0));
    if (exceptct > 0) then
      raise ERnnScript.Create(exmsg);
  end;

  procedure NexTJvCmdProcess;
  var
    i : integer;
  begin
    WITH FScriptState do
    if assigned(FCmdExtract) then
      FCmdExtract(SELF,linenum,linechar,script,cmd)
    else
      DefaultCmdExtract(lineNum,linechar,script,cmd);

    if assigned(FCmdProcess) then
      FCmdProcess(SELF,cmd)
    else
      DefaultCmdProcess(cmd);

    if assigned(FLogStrings) then
    begin
      for i := 0 to cmd.Count-1 do
        FLogStrings(SELF,cmd[i]);
      FLogStrings(SELF,'');
    end;
    if assigned(FBeforeCmdExecute) then
      FBeforeCmdExecute(SELF,cmd);
    if assigned(FCmdExecute) then
      FCmdExecute(SELF,cmd)
    else
      DefaultCmdExecute(cmd);
  end;


begin
  InitializeScriptState(ss);
  ss.ownerid := 'EXECUTE';
  ss.script := startscript;
  stackct := FStackIX;
  LockScript;
  PushScriptState(ss);
  UnlockScript;
  exceptct := 0;
  cmd := TJvRnnStrList.Create;
  try
    LoopLines;
  finally
    cmd.free;
  end;
end;


constructor TJvRnnScriptStep.create(scriptor:TJvRnnCustomScript; script : TStrings);
begin
  inherited create;
  FScriptor := scriptor;
  FCmd := TJvRnnStrList.create;
  FScript := script;
  Reset;
end;


destructor TJvRnnScriptStep.destroy;
begin
  FCmd.free;
  inherited Destroy;
end;

function TJvRnnScriptStep.GetLineNum : integer;
begin
  Result := FScriptor.FScriptState.lineNum;
end;

function TJvRnnScriptStep.GetLineChar : integer;
begin
  Result := FScriptor.FScriptState.linechar;
end;

function TJvRnnScriptStep.getStrings : TStrings;
begin
  Result := FCmd;
end;

procedure TJvRnnScriptStep.Reset;
var
  ss : TJvRnnScriptState;
begin
  FScriptor.ResetStack;
  InitializeScriptState(ss);
  ss.Script := FScript;
  FScriptor.PushScriptState(ss);
  FRunState := srs_extract;
end;

procedure TJvRnnScriptStep.Fetch;
var
   op:TJvCmdExtractOp;
   ilineNum, ilineChar : integer;
begin
  if IsFinished then
  begin
    FRunState := srs_finished;
    exit;
  end;
  op := FScriptor.CmdExtract;
  ilineNum := FScriptor.FScriptState.LineNum;
  ilineChar := FScriptor.FScriptState.LineChar;
  if assigned(op) then
    op(FScriptor,ilineNum,ilineChar,FScriptor.FScriptState.Script,FCmd)
  else
    FScriptor.DefaultCmdExtract(ilineNum,ilineChar,FScriptor.FScriptState.Script,FCmd);
  FScriptor.FScriptState.LineNum := ilineNum;
  FScriptor.FScriptState.LineChar := iLineChar;
  FRunState := srs_process;
end;

procedure TJvRnnScriptStep.ExecuteStep;
begin
  try
    CASE FRunState of
      srs_extract: Fetch;
      srs_process: Process;
      srs_execute: Execute;
    END;
  except
    on rsx:ERnnScript do
    begin
      MessageDlg('Run script error: '+rsx.message,mtError,[mbOK],0);
      FRunState := srs_error;
    end;
  end;
end;

procedure TJvRnnScriptStep.ExecuteNext;
begin
  REPEAT
    ExecuteStep;
  UNTIL ((FRunState = srs_extract)OR(FRunState >= srs_finished));
end;

procedure TJvRnnScriptStep.ExecuteAll;
begin
  REPEAT
    ExecuteStep;
  UNTIL (FRunState >= srs_finished);
end;


function TJvRnnScriptStep.IsFinished : boolean;
begin
  Result := FScriptor.Running;
  if Result then
  begin
    WITH FScriptor.FScriptState do
      Result := (LineNum >= Script.count);
  end;
end;

procedure TJvRnnScriptStep.Process;
var
   op:TJvCmdProcessOp;
begin
  op:=FScriptor.CmdProcess;
  if assigned(op) then
    op(FScriptor,FCmd)
  else
    FScriptor.DefaultCmdProcess(FCmd);
  FRunState := srs_execute;
end;

procedure TJvRnnScriptStep.Execute;
var
  op:TJvCmdProcessOp;
begin
  if (not FNoExecute) then
  begin
    op:=FScriptor.CmdExecute;
    if assigned(op) then
      op(FScriptor,FCmd)
    else
      FScriptor.DefaultCmdExecute(FCmd);
  end;
  FRunState := srs_extract;
end;

{------------------------------------------------------------------------------}
function TJvRnnCustomScript.GetVariables : TJvCCStrStrArray;
begin
  if FVariables = nil then
    ValueList := TJvRnnValueList.Create(SELF);
  Result := FVariables;
end;

procedure TJvRnnCustomScript.SetVariable(const varname, value:string);
begin
  Variables.SetVariable(varname,value);
end;

function TJvRnnCustomScript.EvaluateName(var ps : TParseState) : string;
var
  vlist : TJvCCStrStrArray;
  ix : integer;
begin
  vlist := ValueList.StrStrArray;
  ix := vlist.IndexOf(ps.token);
  if (ix >= 0) then
    Result := vlist.GetIndexValue(ix)
  else
    Result := ps.token;
end;

procedure TJvRnnCustomScript.KillVariable(const varname : string);
var
  ix : integer;
begin
  ix := Variables.IndexOf(varname);
  if (ix >= 0) then
    FVariables.DeleteItemsAt(ix,1);{else failure forgiven}
end;

function  TJvRnnCustomScript.GetValue(const varname : string) : string;
begin
  if FVariables = nil then
    ThrowNoVariables;
  Result := FVariables.Values[varname];
end;

constructor TJvRnnCustomScript.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FUserOps := TStringList.Create;
  FUserFuncs := TStringList.Create;
  FPrefixChar := '@';
  FParamChar := ':';
  FTerminator := ';';
  FStdTerminator := ';';
  FDateFormat := kRnnDefaultDateFormat;
  FTimeFormat := kRnnDefaultTimeFormat;
  FDateTimeFormat := kRnnDefaultDateTimeFormat;
  FDeleteComments := TRUE;
end;

procedure TJvRnnCustomScript.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Operation = opRemove then
  begin
    if AComponent = ValueList then
      ValueList := nil
    else if AComponent = ScriptStore then
      ScriptStore := nil;
  end;
  inherited Notification(AComponent, Operation);
end;

procedure TJvRnnCustomScript.SetVarComp(valList : TJvRnnValueList);
begin
  if (FVarComp <> nil) and
    (FVarComp.Owner = SELF) then
      FVarComp.Free;
  FVarComp := valList;
  if FVarComp <> nil then
    FVariables := valList.StrStrArray
  else
    FVariables := nil;
end;

procedure TJvRnnCustomScript.SetUserOp(const opname : string; opProc: PJvRegUserOp);
var
  ix : integer;
begin
  ix := FUserOps.IndexOf(opname);
  if (ix < 0) then
    FUserOps.AddObject(opname,TObject(opProc) )
  else
    FUserOps.Objects[ix] := TObject(opProc);
end;

procedure TJvRnnCustomScript.DeleteUserOp(const opname:string);
var
  ix : integer;
begin
  ix := FUserOps.IndexOf(opname);
  if (ix >= 0) then
    FUserOps.Delete(ix);
end;


procedure TJvRnnCustomScript.SetUserFunc(const opname : string; opProc: PJvRegUserFunc);
var
  ix : integer;
begin
  ix := FUserFuncs.IndexOf(opname);
  if (ix < 0) then
    FUserFuncs.AddObject(opname,TObject(opProc) )
  else
    FUserFuncs.Objects[ix] := TObject(opProc);
end;

procedure TJvRnnCustomScript.DeleteUserFunc(const opname:string);
var
  ix : integer;
begin
  ix := FUserFuncs.IndexOf(opname);
  if (ix >= 0) then
    FUserFuncs.Delete(ix);
end;


function  TJvRnnCustomScript.GetUserOp(const opname:string) : PJvRegUserOp;
var
	ix : integer;
begin
  ix := FUserOps.IndexOf(opname);
  if (ix >= 0) then
    Result := PJvRegUserOp(FUserOps.Objects[ix])
  else
    Result := nil;
end;


procedure TJvRnnCustomScript.SubstituteParams(var s : string);
var
  p : PChar;
  r : PChar;
  t : string;
  w : string;
{$IFDEF VER80}
  rbuf : array[0..255] of Char;
{$ENDIF}
  ix : integer;
  didSubstitute : Boolean;
begin
  w := '';
{$IFNDEF VER80}
  r := PChar(s);
{$ELSE}
  RnnStrPCopy(rbuf,s);
  r := rbuf;
{$ENDIF}
  p := StrScan(r,FParamChar);{SUBSTITUTE_CHAR must not be alphanum}
  didSubstitute := FALSE;
  if p<>nil then
  repeat	{point past SUBSTITUTE_CHAR}
    Inc(p);
    t := GetAlphaToken(p);
    if Length(t) > 0 then
      ix := FVariables.IndexOf(t)
    else
      ix := -1;
    if (ix >= 0) then
    begin {is a variable, substitute value}
      AddCStrSeg(w,r,p-1); {get string so far to p-1}
      Inc(p,Length(t)); {point past variable name}
      r := p;   {next search position}
      t := FVariables.GetIndexValue(ix);
      w := w + t;  {add value}
      didSubstitute := TRUE;
    end else
    begin {string up to SUBSTITUTE_CHAR is OK}
      AddCStrSeg(w,r,p);
      r := p; {next search position}
    end;
    p := StrScan(r,FParamChar);
  until (p=nil);
  if didSubstitute then
  begin
    AddCStrSeg(w,r,nil); {get remainder}
    s := w; {replace with substituted string}
  end;

end;

procedure TJvRnnCustomScript.DefaultCmdProcess(cmd:TStrings);
var
  p : PChar;
  r : PChar;
  t : string;
  w : string;
  rbuf : array[0..256] of Char;
  ix : integer;
  i : integer;
  didSubstitute : Boolean;
  va : TJvCCStrStrArray;
begin
  va := Variables;
  for i := 0 to cmd.Count-1 do
  begin
    w := '';
    RnnStrPCopy(rbuf,cmd[i]);
    r := rbuf;
    p := StrScan(r,FParamChar);{SUBSTITUTE_CHAR must not be alphanum}
    didSubstitute := FALSE;
    if p<>nil then
    repeat	{point past SUBSTITUTE_CHAR}
      Inc(p);
      t := GetAlphaToken(p);
      if Length(t) > 0 then
        ix := va.IndexOf(t)
      else
        ix := -1;
      if (ix >= 0) then
      begin {is a variable, substitute value}
        AddCStrSeg(w,r,p-1); {get string so far to p-1}
        Inc(p,Length(t)); {point past variable name}
        r := p;   {next search position}
        t := va.GetIndexValue(ix);
        w := w + t;  {add value}
        didSubstitute := TRUE;
      end else
      begin {string up to SUBSTITUTE_CHAR is OK}
        AddCStrSeg(w,r,p);
        r := p; {next search position}
      end;
      p := StrScan(r,FParamChar);
    until (p=nil);
    if didSubstitute then
    begin
      AddCStrSeg(w,r,nil); {get remainder}
      cmd[i] := w; {replace with substituted string}
    end;
  end;
end;



procedure TJvRnnCustomScript.DefaultCmdExtract(var lineNum,linechar:integer;
						script,cmd : TStrings);
var
  s : string;
  inComment : boolean;
  done: boolean;
  terminal : char;
begin
{extract the next command into cmd until a terminator,
remove comments if required}
  cmd.Clear;
  inComment := FALSE;
  done := FALSE;
  terminal := FStdTerminator;
  while (linenum < script.Count) and (not done) do
  begin
    s := script.Strings[linenum];
    Inc(lineNum);
    if Length(s)=0 then
      continue;
    {eliminate simple /*  */ comments}
    if FDeleteComments then
      inComment := RemoveComments(inComment, s);
    s := TRIM(s);
    if Length(s)=0 then
      continue;
    if (cmd.Count = 0) then
    begin
      if (s[1] = FPrefixChar) then
        terminal := FTerminator;
    end;
    if terminal = #0 then
      done := TRUE
    else if (s[Length(s)] = terminal) then
    begin
      SetLength(s,Length(s)-1);
      done := TRUE;
    end;
    cmd.Add(s);
  end;
end;

procedure TJvRnnCustomScript.DefaultCmdExecute(cmd:TStrings);
var
  p : PChar;
  rbuf : array[0..255] of char;
  r : PChar;
  cmdWord : string;
  parentype: TJvRnnParenType;
  cmdInfo : TJvScriptCmdInfo;
begin
  if cmd.Count = 0 then
    exit;
  RnnStrPCopy(rbuf,cmd[0]);
  p := rbuf;
  if not StrNextToken(p,cmdword,parentype) then
    exit;
  {handle first character perculiarities}
  TRY
    IF cmdword[1]=FPrefixChar then
    begin
      r := StrListToCStr(cmd,#10);
      try
        RnnStrPCopy(rbuf,cmdWord);
        WITH cmdInfo do
        begin
          scriptor := SELF;
          cmdcstr := r + Length(cmdword)+1;
          cmdname := rbuf+1; {exclude '@'}
          cmdstrings := cmd;
        end;
        ExecuteScriptFunction(@cmdInfo);
      finally
        StrDispose(r);
      end;
    end else
    begin
      if assigned(FExecuteStd) then
        FExecuteStd(SELF,cmd)
      else
        ExecuteStd(cmd);
    end;
  EXCEPT
    on ex2:ERnnScript do
    begin
      MessageDlg(ex2.Message+#13#10' Error in command: '+StrPas(rbuf),
                  mtError,[mbOK],0);
      raise;
    end;
  END;
end;

function  TJvRnnCustomScript.GetUserFunc(const opname:string) : PJvRegUserFunc;
var
	 ix : integer;
begin
	 ix := FUserFuncs.IndexOf(opname);
  if (ix >= 0) then
    Result := PJvRegUserFunc(FUserFuncs.Objects[ix])
  else
    Result := nil;
end;

procedure TJvRnnCustomScript.CheckHasStore;
begin
  if FScriptStore=nil then
    raise ERnnScript.create('No script store assigned');
end;

procedure TJvRnnCustomScript.ExecuteName(const scriptName : string);
var
  slist : TStrings;
begin
  CheckHasStore;
  slist := TStringList.create;
  try
    FScriptStore.GetScriptAsList(scriptName,slist);
    Execute(slist);
  finally
    slist.free;
  end;
end;


procedure TJvRnnCustomScript.ExecuteNameParams(const scriptName : string; params : TStrings);
var
  slist : TStrings;
begin
  CheckHasStore;
  slist := TStringList.create;
  try
    SELF.Variables.AppendValueStrings(params);
    FScriptStore.GetScriptAsList(scriptName,slist);

    Execute(slist);
  finally
    slist.free;
  end;
end;

procedure Abort_F(data:Pointer);
var
  msg : string;
  cmdInfo : TJvScriptCmdInfo;
  p : PChar;
begin
  cmdInfo := TJvScriptCmdInfo(data^);
  p := cmdInfo.cmdcstr;
  msg := StrPas(p);
  raise ERnnScript.Create(msg);
end;

procedure TJvRnnCustomScript.ExecuteScriptFunction(cmdInfo:Pointer);
var
  op : PJvRegUserOp;
begin
  op := GetUserOp(StrPas(TJvScriptCmdInfo(cmdInfo^).cmdname));
  if (op <> nil) then
    TJvRegUserOp(op)(cmdInfo);  {note op^ doesn't work}
end;

procedure TJvRnnCustomScript.ExecuteStd(cmd : TStrings);
begin
  {override}
end;


procedure KillVariable_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  token : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise ERnnScript.create('Character or string expected');
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  p := cmdInfo^.cmdcstr;
  tokentype := tok_none;

  Sub_Next;
  if tokentype = tok_name then
    cmdInfo^.scriptor.KillVariable(token);
end;



procedure SetParamChar_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  token : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise ERnnScript.create('Character or string expected');
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  p := cmdInfo^.cmdcstr;

  Sub_Next;
  if (tokentype=tok_bad) OR (tokentype = tok_name) then
    cmdInfo^.scriptor.ParamChar := token[1];
end;

procedure SetDateFormat_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  token : string;
  ptype : TJvRnnParenType;
  scriptor : TJvRnnCustomScript;

  procedure ThrowError;
  begin
    raise ERnnScript.create('Date/time format expected');
  end;

  procedure sub_next;
  begin
    if not StrNextToken(p,token,ptype) then
      ThrowError;
  end;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  p := cmdInfo^.cmdcstr;
  Sub_Next;
  p := cmdInfo^.cmdname;
  scriptor := cmdInfo^.scriptor;
  if StrPos(p,'DATETIME') <> nil then
  begin
    scriptor.DateTimeFormat := token;
  end else
  if StrPos(p,'DATE') <> nil then
  begin
    scriptor.DateFormat := token;
  end else
  if StrPos(p,'TIME') <> nil then
  begin
    scriptor.TimeFormat := token;
  end
  else
    ThrowError;
end;

procedure SetVariable_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  savep : PChar;
  token : string;
  varname : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  scriptor : TJvRnnCustomScript;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise ERnnScript.create('Character or string expected');
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  p := cmdInfo^.cmdcstr;
  tokentype := tok_none;
  Sub_Next;
  if tokentype = tok_name then
  begin
    varname := token;
    scriptor := cmdInfo^.scriptor;
    savep := p;
    Sub_Next;
    if not (tokentype = tok_equal) then
      p := savep;
    token := scriptor.GetExpressionValue(p);
    scriptor.SetVariable(varname,token);
  end;
end;


procedure LOGMSG_F(data:Pointer);
var
  msg : string;
  cmdInfo : PJvScriptCmdInfo;
  scriptor :  TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  msg := StrPas(cmdInfo^.cmdcstr);
  scriptor.LogMsg(msg);
end;

procedure LOGERROR_F(data:Pointer);
var
  msg : string;
  cmdInfo : PJvScriptCmdInfo;
  scriptor :  TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  msg := StrPas(cmdInfo^.cmdcstr);
  scriptor.LogError(msg);
end;


procedure TRY_F(data:Pointer);
var
  xtoken : string;
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
  tryct : integer;
  saveState : TJvRnnScriptState;
  newState : TJvRnnScriptState;
  finalstate : TJvRnnScriptState;
  ps : TParseState;

  procedure Error_IF(const s : string);
  begin
    raise ERnnScript.Create('IF <expression>=<expression>; ' + s);
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;

  saveState := scriptor.GetScriptState; {a copy of current script state}

  tryct := 1; 

  {look for matching ...FINALLY..ENDTRY  }
  InitializeScriptState(newState);
  InitializeScriptState(finalState);

  newState.script := saveState.script;  {1st statement of try}
  newState.linenum := saveState.linenum;
  newState.ownerid := kTRY;
  finalState.script := saveState.script;  {1st statement of final--except}
  finalState.linenum := saveState.linenum;

  with saveState do
  Repeat
    if (lineNum >= script.count) then
      Error_IF('Missing ENDTRY');
    xtoken := script[lineNum];
    if (Length(xtoken)=0) then
    begin
      Inc(lineNum);
      continue;
    end;
    ps_init(ps,PChar(xtoken));
    ps_skipwhite(ps);
    if (ps.p^ = '@') then
    begin
      Inc(ps.p);
      if ps_nexttokentype(ps) then
      begin
        if COMPARETEXT(ps.token,kTRY)=0 then begin
          Inc(tryct);
        end
        else if (tryct=1) AND
          ((COMPARETEXT(ps.token,kFINALLY)=0)
          OR(COMPARETEXT(ps.token,'EXCEPT')=0))then
        begin
          Inc(linenum);
          finalState.ownerid := ps.token;
          finalState.linenum := linenum; {this is where finally will start}
          finalState.script := saveState.script; {have a script as well}
        end
        else if COMPARETEXT(ps.token,kENDTRY)=0 then
        begin
          Dec(tryct);
          if (tryct) = 0 then
          begin
            Inc(linenum);  {ENDIF is not executed}
            break;
          end;
        end;
      end;
    end;
    Inc(lineNum);
  Until(tryct = 0);

  scriptor.SetScriptState(saveState);  {points to past ENDTRY}
  scriptor.PushScriptState(finalstate);
  scriptor.PushScriptState(newState); {execute the TRY}
end;

procedure FINALLY_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  scriptor.PopScriptState;
end;

procedure ENDTRY_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  scriptor.PopScriptState;
end;

procedure IF_F(data:Pointer);
var
  token : string;
  xtoken : string;
  val1 : extended;
  val2 : extended;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;

  true_value:boolean;
  do_else : boolean;

  dct : integer;
  saveState : TJvRnnScriptState;
  newState : TJvRnnScriptState;
  expr1, expr2 : string;
  p : PChar;
  inlist : TStrings;

  procedure Error_IF(const s : string);
  begin
    raise ERnnScript.Create('IF <expression>=<expression>; ' + s);
  end;


  function sub_next : boolean;
  begin
    Result := StrNextTokenType(p,token,ptype,tokentype);
  end;

begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  p := cmdInfo^.cmdcstr;

  expr1 := scriptor.GetExpressionValue(p);

  if (not sub_next) then
    Error_IF('operator expected');
  true_value := FALSE;
  if (ORD(tokentype) >= ORD(tok_lessthan)) AND
      (ORD(tokentype) <= ORD(tok_greaterthan)) then
  begin
    expr2 := scriptor.GetExpressionValue(p);
    if tokentype = tok_equal then
      true_value := (expr1 = expr2)
    else if tokentype = tok_notequal then
      true_value := not (expr1 = expr2)
    else
    TRY
      {number comparisons}
      val1 := StrToFloat(expr1);
      val2 := StrToFloat(expr2);
      CASE tokentype of
        tok_lessthan :
          true_value := (val1 < val2);
        tok_greaterthan :
          true_value := (val1 > val2);
        tok_lesseq :
          true_value := (val1 <= val2);
        tok_greatereq :
          true_value := (val1 >= val2);
      ELSE
        true_value := FALSE;
      END;
    EXCEPT
      scriptor.LogMsg(Format('Number conversion error %s,%s',[expr1,expr2]));
      raise;
    END;
  end
  else if (token = 'IN') then
  begin
    {get bracketed list of strings}
    if (not sub_next) OR (token[1] <> '(') then
      Error_If('"(" expected');
    inlist := scriptor.ParseFunctionArgs(p);
    if (inlist <> nil) then
    try
      true_value := (inlist.indexof(expr1) >= 0);
    finally
      inlist.free;
    end;
  end
  else if (token = 'LIKE') then
  begin
    {pattern matching}
  end;
  {if true_value, then will want to execute the
  succeeding lines till ENDIF,
  else just find the matching ENDIF, and
  finish}

  saveState := scriptor.GetScriptState; {a copy of current script state}

  dct := 1; {ENDIF - number to find}

  {gather up succeeding lines for a new script until find ENDIF;
  problem is succeeding lines may contain
  text that contains '@IF;  @ENDIF;'
  }
  InitializeScriptState(newState);

  if true_value then
  begin  {set new state to be within the if }
    newState.script := saveState.script;
    newState.linenum := saveState.linenum
  end;

  do_else := FALSE;
  with saveState do
  Repeat
    if (lineNum >= script.count) then
      Error_IF('Missing END_IF');
    xtoken := script[lineNum];
    p := StrToPChar(xtoken);
    {test if line starts with a @IF, and count the END_IF}
    p := StrStartToken(p);
    if (p^ = '@') then
    begin
      Inc(p);
      if sub_next then
      begin
        if COMPARETEXT(token,'IF')=0 then begin
          Inc(dct);
        end
        else if (dct=1) AND (COMPARETEXT(token,'ELSE')=0) then
        begin
          if (not true_value) then
          begin
            Inc(linenum);
            newstate.linenum := linenum; {this is where execution will start}
            newState.script := saveState.script; {have a script as well}
            do_else := TRUE;
          end;
        end
        else if COMPARETEXT(token,'ENDIF')=0 then
        begin
          Dec(dct);
          if (dct) = 0 then
          begin
            Inc(linenum);  {ENDIF is not executed}
            break;
          end;
        end;
      end;
    end;
    Inc(lineNum);
  Until(dct = 0);

  scriptor.SetScriptState(saveState);

  if true_value then begin
  newState.ownerid := 'IF';
  scriptor.PushScriptState(newState); {execute the IF}
  end else if do_else then begin
  newState.ownerid := 'ELSE';
  scriptor.PushScriptState(newState); {execute the ELSE}
  end;
end;


procedure ELSE_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
begin
	{if an else is reached, then pop execution state}
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  scriptor.PopScriptState;
end;

procedure ENDIF_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  scriptor : TJvRnnCustomScript;
begin
  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;
  scriptor.PopScriptState;
end;

function TJvRnnCustomScript.GetScriptNoFail(const sname : string): TStrings;
begin
  Result := ScriptStore.ScriptList.GetScriptStrings(sname);
  if Result = nil then
    raise ERnnScript.Create('Script '+sname+' not found');
end;

procedure Execute_F(data:Pointer);
var
  cmdInfo : PJvScriptCmdInfo;
  p : PChar;
  scriptor : TJvRnnCustomScript;
  token : string;
  ptype : TJvRnnParenType;
  tokentype : TJvRnnTokenType;
  newState : TJvRnnScriptState;

  procedure sub_next;
  begin
    if not StrNextTokenType(p,token,ptype,tokentype) then
      raise ERnnScript.create('No script label');
  end;

begin
  InitializeScriptState(newState);

  cmdInfo := PJvScriptCmdInfo(data);
  scriptor := cmdInfo^.scriptor;

  p := cmdInfo^.cmdcstr;

  sub_next;

  WITH newState do
  begin
    script := scriptor.Script[token];
    ownerid := 'EXECUTE';
    lineNum := 0;
    lineChar := 0;
    refptr := nil;
    scriptlabel := token;
  END;

  scriptor.PushScriptState(newState);
end;

function GetSysDate_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := FormatDateTime(sct.DateFormat,Date);
end;

function GetSysTime_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := FormatDateTime(sct.TimeFormat,Now);
end;

function GetSysDateTime_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := FormatDateTime(sct.DateTimeFormat,Now);
end;

function Quote_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := QuoteStr(StrListToString(args,' '));
end;

function Concat_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := StrListToString(args,#0);
end;


function CountArgs_F(sct: TJvRnnCustomScript; args : TStrings): string;
begin
  Result := IntToStr(args.Count);
end;

end.
