{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: RnnScrEd.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnScrEd;

interface

uses WinTypes, WinProcs, Classes, Graphics, Forms, Controls, Buttons,
  StdCtrls, ExtCtrls, JvRnnScrSt, Dialogs, Messages,
  JvCCArray,Menus,DsgnIntf
{$IFNDEF WIN32}
  ,D3Split
{$ENDIF}
;

type

  TRnnScriptListProperty = class(TPropertyEditor)
    procedure Edit ; override;
    function GetValue:string; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TfrmScriptListEdit = class(TForm)
    Panel5: TPanel;
    btnOpenFile: TSpeedButton;
    btnSaveAs: TSpeedButton;
    openTextFileDlg: TOpenDialog;
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    saveTextFileDlg: TSaveDialog;
    Label1: TLabel;
    Label2: TLabel;
    editName: TEdit;
    editType: TEdit;
    nameListMenu: TPopupMenu;
    NewScriptItem: TMenuItem;
    N1: TMenuItem;
    Delete1: TMenuItem;
    Panel4: TPanel;
    Panel2: TPanel;
    memo: TMemo;
    Panel1: TPanel;
    nameList: TListBox;
    Panel3: TPanel;
    Splitter1: TSplitter;
    dlgFont: TFontDialog;
    SpeedButton1: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);

    procedure editNameChange(Sender: TObject);
    procedure editTypeExit(Sender: TObject);
    procedure btnOpenFileClick(Sender: TObject);
    procedure Panel5Resize(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure btnSaveAsClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure editTypeChange(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure NewScriptItemClick(Sender: TObject);
    procedure nameListClick(Sender: TObject);
    procedure Delete1Click(Sender: TObject);
    procedure editNameExit(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
  private
    { Private declarations }
    FScriptList : TJvRnnScriptList;

    FChangingScript : Boolean;

    FModified : Boolean;
    FSaved : Boolean;
    FDisplayData : TJvCCArray;
    FOriginalList : TJvRnnScriptList;
    FEditedScript : TJvRnnScript;

    procedure RefreshScriptList;
    procedure SetScriptInfo(script:TJvRnnScript);
    function AddNewScript: TJvRnnScript;
     procedure DeleteDisplayIndex(ix:integer);
    procedure DeleteRyScript(script : TJvRnnScript);

    procedure SetEditScript(script:TJvRnnScript);
    procedure ReplaceScript;
    procedure RestoreScript;
    procedure RefreshDisplayData;
    procedure ClearDisplayData;
  public
    { Public declarations }

    procedure WMGetMaxMinInfo(var wmsg: TWMGetMinMaxInfo);  message WM_GETMINMAXINFO;

    property  EditedScript : TJvRnnScript read FEditedScript;

    procedure ApplyEditChanges;
    procedure SetScriptList(scriptList:TJvRnnScriptList);

    procedure SetDisplayIndex(ix:integer);
    property Modified : boolean read FModified;
    property ReturnList : TJvRnnScriptList read FOriginalList;
  end;

function EditScriptList(scriptList:TJvRnnScriptList):boolean; {true if modifed}

var
  frmScriptListEdit: TfrmScriptListEdit;


procedure Register;

implementation
Uses Sysutils,JvRnnStrms,JvCCRecrd;
{$R *.DFM}

type
  PScrDisplayRec = ^TScrDisplayRec;
  TScrDisplayRec = record
    rys : TJvRnnScript;
    selstart : integer;
    sellength : integer;
    dirty : boolean;
  end;

  TScrDisplayRecMgr = class(TJvCCRecordMgr)
    constructor Create; override;
    function    CompareItems(const p1, p2):integer;override;
  end;

  TScrDisplayRecArray = array [0..2000] of TScrDisplayRec;
  PScrDisplayRecArray = ^TScrDisplayRecArray;

  TScrDisplayArray = class(TJvCCArray)
  private
    FPtr : PScrDisplayRecArray;
  protected
  public
    constructor Create(inCount : integer);
    function  FindScript(ryscr:TJvRnnScript):integer;
    procedure AppendScript(ryscr:TJvRnnScript);
    function GetScript(ix:integer) : TJvRnnScript;
    property Ptr : PScrDisplayRecArray read FPtr;
  end;

type
  TJvRnnScriptProcess = procedure (script : TJvRnnScript) of object;

constructor TScrDisplayRecMgr.Create;
begin
  inherited Create;
  InitRecordMgr(sizeof(TScrDisplayRec),TRUE);
end;

function  TScrDisplayRecMgr.CompareItems(const p1, p2):integer;
begin
  Result := CompareText(TScrDisplayRec(p1).rys.Name,TScrDisplayRec(p2).rys.Name);
end;

function TScrDisplayArray.FindScript(ryscr:TJvRnnScript):integer;
var
  sdr:TScrDisplayRec;
begin
  sdr.rys := ryscr;
  if not GetItemIndex(Result,sdr) then
    Result := -1;
end;

procedure TScrDisplayArray.AppendScript(ryscr:TJvRnnScript);
var
  sdr:TScrDisplayRec;
  ix : integer;
begin
  sdr.rys := ryscr;
  if GetItemIndex(ix,sdr) then
    exit;
  CreateItemsAt(ix,1);
  Ptr^[ix].rys := ryscr;
end;

constructor TScrDisplayArray.Create(inCount : integer);
begin
  inherited CreateMgr(TScrDisplayRecMgr);
  BlockHdl := @FPtr;
  Count := inCount;
end;


function TScrDisplayArray.GetScript(ix:integer) : TJvRnnScript;
begin
  Result := Ptr^[ix].rys;
end;


function EditScriptList(scriptList:TJvRnnScriptList):boolean;
var
  ed :TfrmScriptListEdit;

begin
  ed := TfrmScriptListEdit.Create(Application);
  try
    ed.SetScriptList(scriptList);
    ed.ShowModal;
    Result := ed.Modified;
  finally
    ed.Free;
  end;
end;

procedure  TfrmScriptListEdit.WMGetMaxMinInfo(var wmsg: TWMGetMinMaxInfo);
begin
  with wmsg do
  begin
    with MinMaxInfo^.ptMinTrackSize do
    begin
      X := 500;
      Y := 200;
    end;
  Result := 0;
  end;
  inherited;
end;

procedure  TfrmScriptListEdit.RefreshScriptList;
var
  i : integer;
  script : TJvRnnScript;
  p : PScrDisplayRecArray;
  saveix : integer;
begin
  saveix := nameList.ItemIndex;
  nameList.Clear;
  p := FDisplayData.BlockPtr;
  for i := 0 to FDisplayData.Count-1 do
  begin
    script := p^[i].rys;
    nameList.Items.AddObject(script.Name, script);
  end;
  if (saveix < 0) OR (saveix>=FDisplayData.Count) then
    saveix := FDisplayData.Count-1;

  if saveix >= 0 then
  begin
    nameList.ItemIndex := saveix;
    SetEditScript(TScrDisplayArray(FDisplayData).GetScript(saveix));
  end else
  begin
    memo.clear;
    memo.enabled := FALSE;
  end;
  nameList.Refresh;
end;

procedure TfrmScriptListEdit.SetEditScript(script:TJvRnnScript);
var
  ix : integer;
  sdr : PScrDisplayRec;
begin
  if script <> nil then
  begin
    memo.Lines := script.Strx;
    ix := TScrDisplayArray(FDisplayData).FindScript(script);
    if ix >= 0 then
    begin
      sdr := @TScrDisplayArray(FDisplayData).Ptr^[ix];
      memo.enabled := TRUE;
      memo.selstart := sdr^.selstart;
      memo.sellength := sdr^.sellength;
    end;
  end;
  FEditedScript := script;
  SetScriptInfo(script);
end;

procedure TfrmScriptListEdit.ApplyEditChanges;
var
  ix : integer;
  sdr :PScrDisplayRec;
begin
  if FEditedScript = nil then
  begin
    memo.Lines.Clear;
    exit;
  end else
  begin
    FEditedScript.Strx := memo.Lines;
    ix := TScrDisplayArray(FDisplayData).FindScript(FEditedScript);
    if ix >= 0 then
    begin
      sdr := @TScrDisplayArray(FDisplayData).Ptr^[ix];
      sdr^.selstart := memo.selstart;
      sdr^.sellength := memo.sellength;
    end;
  end;
end;

procedure TfrmScriptListEdit.SetDisplayIndex(ix:integer);
var
  scr : TJvRnnScript;
begin
  if (ix < 0) OR (ix >= FDisplayData.Count) then
  begin
    ApplyEditChanges;
    exit;
  end;
  scr := TScrDisplayArray(FDisplayData).GetScript(ix);
  if scr<>EditedScript then
  begin
    ApplyEditChanges;
    SetEditScript(scr);
  end;
end;

procedure TfrmScriptListEdit.RefreshDisplayData;
var
  i : integer;
  scr : TJvRnnScript;
begin
  FDisplayData.Count:=0;
  for i := 0 to FScriptList.Count-1 do
  begin
    scr := FScriptList.Scripts[i];
    TScrDisplayArray(FDisplayData).AppendScript(scr);
  end;
end;

procedure TfrmScriptListEdit.SetScriptList(scriptList:TJvRnnScriptList);
begin
  if (FScriptList <> nil) then
    FScriptList.Free;
  FOriginalList := scriptList;
  FScriptList := TJvRnnScriptList.Create;
  FScriptList.Assign(FOriginalList);
  RefreshDisplayData;
  RefreshScriptList;
  if FDisplayData.Count > 0 then
  begin
    SetDisplayIndex(0);
  end;
end;


function TfrmScriptListEdit.AddNewScript : TJvRnnScript;
var
    newname : string;
    newix : integer;

    function makescriptname:string;
    begin
        Result := Format('Untitled#%d',[newix]);
    end;

begin
    newix := 1;

    newname := makescriptname;
    while FScriptList.GetScript(newname) <> nil do begin
        Inc(newix);
        newname := makescriptname;
    end;
    Result := TJvRnnScript.Create(newname,'TEXT',FScriptList);
    TScrDisplayArray(FDisplayData).AppendScript(Result);
    RefreshScriptList;
    newix := TScrDisplayArray(FDisplayData).FindScript(Result);
    if newix >= 0 then
        nameList.ItemIndex := newix;
end;


procedure TfrmScriptListEdit.FormCreate(Sender: TObject);
begin
     FDisplayData := TScrDisplayArray.Create(0);
     OKBtn.Caption := '';
     CancelBtn.Caption := '';
end;

procedure TfrmScriptListEdit.FormDestroy(Sender: TObject);
begin
    FDisplayData.Free;
end;



procedure TfrmScriptListEdit.DeleteRyScript(script : TJvRnnScript);
var
    dix : integer;
begin
    dix := TScrDisplayArray(FDisplayData).FindScript(script);

    if (dix >= 0) then
        FDisplayData.DeleteItemsAt(dix, 1);
    script.free;
end;



procedure TfrmScriptListEdit.DeleteDisplayIndex(ix:integer);
var
    scr : TJvRnnScript;
begin
    if (ix < 0) OR (ix >= FDisplayData.Count) then
        exit;
    scr := TScrDisplayArray(FDisplayData).GetScript(ix);
    DeleteRyScript(scr);
    memo.clear;
    RefreshScriptList;
end;

procedure TfrmScriptListEdit.SetScriptInfo(script:TJvRnnScript);
begin
    FChangingScript := TRUE;
    if (script <> nil) then begin
        editName.Text := script.Name;
        editType.Text := script.ScriptType;
    end else begin
        editName.Clear;
        editType.Clear;
    end;
    FChangingScript := FALSE;
end;



procedure TfrmScriptListEdit.editNameChange(Sender: TObject);
var
    script : TJvRnnScript;
begin
    if FChangingScript then exit;

    script := EditedScript;

    if script = nil then
        exit;

try
    script.Name := editName.Text;
    RefreshScriptList;
finally
    FModified := TRUE;
end;

end;


procedure TfrmScriptListEdit.editTypeExit(Sender: TObject);
var
    script : TJvRnnScript;
begin
    script := EditedScript;
    if script = nil then exit;

    script.scriptType := editType.Text;
    FModified := TRUE;
end;

procedure TfrmScriptListEdit.ClearDisplayData;
begin
    ApplyEditChanges;
    FEditedScript := nil;
    FDisplayData.Count := 0;
    nameList.Clear;
    memo.Clear;

end;

procedure TfrmScriptListEdit.btnOpenFileClick(Sender: TObject);
var
    ext : string;
    slist : TJvRnnScriptList; (* temporary slist *)
    instr : TStream;
begin
    ClearDisplayData;

    if not openTextFileDlg.Execute then  exit;

    ext := UPPERCASE(ExtractFileExt(openTextFileDlg.FileName));
    if (Length(ext)>0) and (ext[1]='.') then
        ext := Copy(ext,2,Length(ext)-1);

    instr := TJvRnnTextStream.CreateFromFile(openTextFileDlg.FileName,fmShareDenyWrite);
    slist := FScriptList;
try
    slist.Clear;
    slist.LoadFromStream(instr);
    RefreshDisplayData;
    RefreshScriptList;
    FModified := TRUE;
finally
    instr.free;
end;

end;




procedure TfrmScriptListEdit.Panel5Resize(Sender: TObject);
begin
      OKBtn.Left := Panel5.ClientWidth - (OKBtn.Width+10);
    CancelBtn.Left := OKBtn.Left - (CancelBtn.Width+10);
end;

procedure TfrmScriptListEdit.btnCloseClick(Sender: TObject);
begin
    if FDisplayData.Count = 0 then exit;
end;



procedure TfrmScriptListEdit.btnSaveAsClick(Sender: TObject);
var
    outstr : TStream;
begin
    ApplyEditChanges;

    if not saveTextFileDlg.Execute then
        exit;

    outstr := TJvRnnTextStream.CreateFromFile(saveTextFileDlg.FileName, fmCreate);
try
    FScriptList.SaveToStream(outstr);
finally
    outstr.Free;
end;
end;

procedure TfrmScriptListEdit.OKBtnClick(Sender: TObject);
begin
    {FOriginalList is used to hold current authorized version}
    if fModified then begin
        ApplyEditChanges;
        ReplaceScript;
    end;
end;

procedure TfrmScriptListEdit.ReplaceScript;
begin
    FOriginalList.TakeOverScripts(FScriptList);
    FScriptList.Free;
    FScriptList := nil;
    FSaved := TRUE;
    FModified := TRUE;
end;

procedure TfrmScriptListEdit.RestoreScript;
begin
    FScriptList.Free;{forget list}
    FScriptList := nil;
    FSaved := TRUE;
    FModified := FALSE;
end;


procedure TfrmScriptListEdit.editTypeChange(Sender: TObject);
begin
    FModified := TRUE;
end;

procedure TfrmScriptListEdit.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
    {FOriginalList is still valid}
    if (FModified) and (not FSaved) then begin
        if MessageDlg('Warning: Changes to component will be lost',
          mtWarning, [mbOK, mbCancel], 0) = mrOK then begin
            RestoreScript;
            CanClose := TRUE;
        end
        else begin
            ModalResult := mrNone;
            CanClose := FALSE;
        end;
    end else
        CanClose := TRUE;

end;

procedure TfrmScriptListEdit.NewScriptItemClick(Sender: TObject);
begin
    ApplyEditChanges;

    SetEditScript(AddNewScript);
end;

procedure TfrmScriptListEdit.nameListClick(Sender: TObject);
begin
    ApplyEditChanges;
    SetDisplayIndex(namelist.ItemIndex);
    ActiveControl := memo;
end;

procedure TfrmScriptListEdit.Delete1Click(Sender: TObject);
begin
{}
    FEditedScript := nil;
    DeleteDisplayIndex(nameList.ItemIndex);
end;

procedure TfrmScriptListEdit.editNameExit(Sender: TObject);
var
    script : TJvRnnScript;
begin
{}
   script := EditedScript;

    if script = nil then
        exit;

try
    script.Name := editName.Text;
    RefreshScriptList;
finally
    FModified := TRUE;
end;

end;

procedure TfrmScriptListEdit.SpeedButton1Click(Sender: TObject);
begin
    {}
    dlgFont.Font := memo.Font;
    if dlgFont.Execute then
        memo.Font := dlgFont.Font;
end;


procedure TRnnScriptListProperty.Edit;
var
    slist : TJvRnnScriptList;
begin
    slist := TJvRnnScriptList(GetOrdValue);
    if slist is TJvRnnScriptList then begin
        if EditScriptList(slist) then
            Modified;
    end;
end;

function TRnnScriptListProperty.GetValue:string;
begin
    Result := 'ScriptList';
end;

function TRnnScriptListProperty.GetAttributes: TPropertyAttributes;
{ Returns information used by the Object Inspector to show the approprate
   adornments (e.g., "..." button, drop-down button). }
begin
  GetAttributes := [paDialog];
end;  { GetAttributes }

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TJvRnnScriptList),nil,'',TRnnScriptListProperty);
end;

end.
