{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: JvRnnVarCp.PAS, released on 2000-12-01.

The Initial Developer of the Original Code is Michael Rynn.
Portions created by Michael Rynn are Copyright (C) 2000  Michael Rynn.
All Rights Reserved.

Contributor(s): ______________________________________.

Last Modified: 2000-12-01

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org

-----------------------------------------------------------------------------}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y+,Z1}

{$I JCL.INC}
unit JvRnnVarCp;

interface

Uses
  Classes,JvCCArray;

type
  TJvRnnValueList = class(TComponent)
  private
    FStrStr : TJvCCStrStrArray;
    procedure SetSSArray(invar : TJvCCStrStrArray);
    procedure ReadSSArray(Reader:TReader);
    procedure WriteSSArray(Writer:TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure SetVariable(const vname, vvalue : string);
    function GetVariable(const vname : string) : string;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    procedure Assign(source : TPersistent); override;

    property Values[const vname : string] : string
      read GetVariable write SetVariable;

  published
    property StrStrArray : TJvCCStrStrArray read FStrStr write SetSSArray;
  end;

implementation
{$IFNDEF WIN32}
{$R *.R16}
{$ENDIF}

procedure TJvRnnValueList.SetSSArray(invar : TJvCCStrStrArray);
begin
  FStrStr.Assign(invar);
end;

constructor TJvRnnValueList.Create(AOwner:TComponent);
begin
  FStrStr := TJvCCStrStrArray.Create(0);
  inherited Create(AOwner);
end;

destructor  TJvRnnValueList.Destroy;
begin
  FStrStr.Free;
  inherited Destroy;
end;

procedure TJvRnnValueList.SetVariable(const vname, vvalue : string);
begin
  FStrStr.SetVariable(vname,vvalue);
end;

function TJvRnnValueList.GetVariable(const vname : string) : string;
begin
  Result := FStrStr.GetValue(vname);
end;


procedure TJvRnnValueList.ReadSSArray(Reader:TReader);
var
  listcount : integer;
  sname : string;
  svalue : string;
  i : integer;
begin
  reader.ReadListBegin;
  listcount := reader.ReadInteger;
  FStrStr.Count := listcount;
	 if (listcount > 0) then
  begin
    i := 0;
    while not reader.EndOfList do
      begin
      sname := reader.ReadString;
      svalue := reader.ReadString;
      if (i < listcount) then
      begin
        FStrStr.SetStrStr(i,sname,svalue);
        Inc(i);
      end;
    end;
  end;
  reader.ReadListEnd;
end;


procedure TJvRnnValueList.DefineProperties(Filer: TFiler);
begin
  Inherited DefineProperties(Filer);
  with Filer do
  begin
    DefineProperty('Variables', ReadSSArray, WriteSSArray, TRUE);
  end;
end;

procedure TJvRnnValueList.Clear;
begin
  FStrStr.Clear;
end;

procedure TJvRnnValueList.Assign(source : TPersistent);
var
  src:TJvRnnValueList;
begin
  src := TJvRnnValueList(source);
  if src is TJvRnnValueList then
  begin
    FStrStr.Assign(src.StrStrArray);
  end else
    inherited Assign(source);
end;

procedure TJvRnnValueList.WriteSSArray(Writer:TWriter);
var
  sname : string;
  svalue : string;
  i : integer;
begin
  writer.WriteListBegin;
  writer.WriteInteger(FStrStr.Count);

	 if ( FStrStr.Count > 0) then
  begin
    for i := 0 to FStrStr.Count-1 do
    begin
      FStrStr.GetStrStr(i,sname,svalue);
      writer.WriteString(sname);
      writer.WriteString(svalue);
		  end;
  end;
	 writer.WriteListEnd;
end;

{_------------------_}



end.
