{******************************************************************************
{	RnnLib  Components  Release 1
{
{ RnnStrUt
{
{ Author : Michael Rynn
{
{ Freeware                    michrynn@ozemail.com.au
{
{ The TVarRec is an intriging handy thing hidden away in the VCL
{ This tries to make it more generally useful.
{
}

{$IFDEF WIN32}
  {$DEFINE LONGSTRINGS}
{$ENDIF}
unit JvVarRecUt;

interface
uses SysUtils;


const
  kMaxVarRec = MAXINT div sizeof(TVarRec) - 1;
type
  PVarRec = ^TVarRec;

  EVarRec = class(Exception);


procedure vrset_empty(var r:TVarRec);
procedure vrset_float(var r:TVarRec; v : Extended );
procedure vrset_integer(var r:TVarRec; v : Longint );
procedure vrset_boolean(var r:TVarRec; v : boolean );
procedure vrset_string(var r:TVarRec; const v : string );
procedure vrset_cstr(var r:TVarRec; v : PChar );
procedure vrset_object(var r:TVarRec; obj : TObject);
procedure vrset_ptr(var r:TVarRec; p : Pointer);
procedure vrset_data(var r:TVarRec; const buf; bsize:integer);
procedure vrforget_data(var r:TVarRec);

function vrget_float(const r:TVarRec) : Extended;
function vrget_boolean(const r:TVarRec) : Boolean;
function vrget_string(const r:TVarRec):string;
function vrget_cstr(const r:TVarRec):PChar;
function vrget_object(const r:TVarRec) : TObject;
function vrget_ptr(const r:TVarRec):Pointer;
function vrget_integer(const r:TVarRec): Longint;

function vr_readstring(const r:TVarRec) : string;

function vr_createptr(bufsize:integer) : Pointer;


implementation

procedure ThrowVarRecError;
begin
    raise EVarRec.create('Invalid TVarRec type');
end;

function vrget_integer(const r:TVarRec) : Longint;
begin
  if r.VType <> vtInteger then
    ThrowVarRecError;
  Result := r.VInteger;
end;

function vrget_float(const r:TVarRec) : Extended;
begin
  if r.VType <> vtExtended then
    ThrowVarRecError;
  Result := r.VExtended^;
end;

function vrget_boolean(const r:TVarRec) : Boolean;
begin
  if r.VType <> vtBoolean then
    ThrowVarRecError;
  Result := r.VBoolean;
end;

function vrget_string(const r:TVarRec):string;
begin
  if r.VType <> vtString then
    ThrowVarRecError;
  {$IFDEF LONGSTRINGS}
  Result := string(r.VAnsiString);
  {$ELSE}
  Result := r.VString^;
  {$ENDIF}
end;

function vrget_cstr(const r:TVarRec):PChar;
begin
  if r.VType <> vtPChar then
    ThrowVarRecError;
  Result := r.VPChar;
end;

function vrget_object(const r:TVarRec) : TObject;
begin
  if r.VType <> vtObject then
    ThrowVarRecError;
  Result := r.VObject;
end;

function vrget_ptr(const r:TVarRec):Pointer;
begin
  if r.VType <> vtPointer then
    ThrowVarRecError;
  Result := r.VPointer;
end;


procedure vrforget_data(var r:TVarRec);
begin
  if r.VType <> vtPointer then
    ThrowVarRecError;
  r.vType := vtInteger;
  r.vInteger := 0;
end;

procedure vrset_empty(var r:TVarRec);
begin
  try
    WITH r do
    CASE vtype of
      vtInteger:
        VInteger := 0;
      vtBoolean:
        VBoolean := FALSE;
      vtChar:
        VChar := #0 ;
      vtExtended:
        if (VExtended <> nil) then
        begin
          FreeMem(r.VExtended,sizeof(Extended));
          VExtended := nil;
        end;
{$IFNDEF WIN32}
      vtString:
        if (VString <> nil) then
        begin
          DisposeStr(VString);
           VString := nil;
        end;
{$ENDIF}
      vtPChar:
        if (VPChar <> nil) then
        begin
          StrDispose(VPChar);
          VPChar := nil;
        end;
      vtPointer:
        if (VPointer <> nil) then
        {$IFDEF WIN32}
          ReallocMem(VPointer,0);
        {$ELSE}
          ReallocMem16(VPointer,0);
        {$ENDIF}
      vtObject:
        if (VObject <> nil) then
        begin
          VObject.free;
          VObject := nil;
        end;
        {$IFDEF LONGSTRINGS}
      vtAnsiString:
        begin
          Finalize(string(VAnsiString));
        end;
        {$ENDIF}
    END;
  finally
    r.vtype := vtInteger;
  end;
end;

procedure vrset_integer(var r:TVarRec; v : Longint );
begin
  vrset_empty(r);
  r.VInteger := v;
  r.VType := vtInteger;
end;

procedure vrset_float(var r:TVarRec; v : Extended );
begin
  vrset_empty(r);
  with r do
  begin
    GetMem(VExtended,sizeof(Extended));
    VExtended^ := v;
    VType := vtExtended;
  end;
end;

procedure vrset_boolean(var r:TVarRec; v : boolean );
begin
  vrset_empty(r);
  r.VBoolean := v;
  r.VType := vtBoolean;
end;

procedure vrset_string(var r:TVarRec; const v : string );
begin
  vrset_empty(r);
{$IFDEF LONGSTRINGS}
  string(r.VAnsiString) := v;
  r.VType := vtAnsiString;
{$ELSE}
  AssignStr(r.VString,v);
  r.VType := vtString;
{$ENDIF}
end;

procedure vrset_cstr(var r:TVarRec; v : PChar );
begin
  vrset_empty(r);
  r.VPChar := StrNew(v);
  r.VType := vtPChar;
end;

procedure vrset_object(var r:TVarRec; obj : TObject);
begin
  vrset_empty(r);
  r.VObject := obj;
  r.VType := vtObject;
end;

procedure vrset_ptr(var r:TVarRec; p : Pointer);
begin
  vrset_empty(r);
  r.VPointer := p;
  r.VType := vtPointer;
end;

procedure vrset_data(var r:TVarRec; const buf; bsize:integer);
var
  p : Pointer;
begin
  vrset_empty(r);
  p := vr_createptr(bsize);
  Move(buf,p^,bsize);
  r.VPointer := p;
  r.VType := vtPointer;
end;

function vr_createptr(bufsize:integer) : Pointer;
begin
  Result := nil;
  {$IFDEF WIN32}
  ReallocMem(Result,bufsize);
  {$ELSE}
  ReallocMem16(Result,bufsize);
  {$ENDIF}
end;


function vr_readstring(const r:TVarRec) : string;
begin
  WITH r do
  begin
    CASE VType of
      vtInteger: Result :=  IntToStr(VInteger);
      vtExtended:  Result := FloatToStr(VExtended^);
      vtPChar: Result := StrPas(VPChar);
      vtString:  Result := VString^;
      {$IFDEF LONGSTRINGS}
      vtAnsiString: Result := string(VAnsiString);
      {$ENDIF}
    ELSE
       ThrowVarRecError;
    END;
  end;
end;



end.

